package main;

import java.io.IOException;
import java.io.PrintStream;
import java.util.Iterator;
import java.util.Properties;
import java.util.concurrent.Executors;
import java.util.logging.Level;
import java.util.logging.Logger;
import main.tools.Tools;
import messif.algorithms.Algorithm;
import messif.algorithms.AlgorithmMethodException;
import messif.objects.LocalAbstractObject;
import messif.objects.impl.ObjectFloatVectorNeuralNetworkL2;
import messif.objects.util.AbstractObjectList;
import messif.objects.util.StreamGenericAbstractObjectIterator;
import messif.operations.AnswerType;
import messif.operations.Approximate;
import messif.operations.data.BulkInsertOperation;
import messif.operations.query.ApproxKNNQueryOperation;
import sf.algorithm.SketchAlgorithmWithRefinement;
import sf.algorithm.SketchObjectMapperAlgorithm;
import sf.objects.sketchcreator.ObjectToSketchTransformator;
import sf.objects.sketchcreator.impl.GenHyperplane;

/**
 *
 * @author xmic
 */
public class KNNApproxEvaluation {


    public static final String storagePrefix = "c:\\Data\\2016\\2016_Unbalanced_sketches\\Decaf\\75_25\\";

    // args[0] file with descriptors
    // args[1] object class
    // args[2] objects to read
    public static void main(String[] args) throws Throwable {
        int finalLength = 205;
        float desiredCorrelation = -1/3f;
        System.setOut(new PrintStream(storagePrefix + "pivots\\PivotsCorrelatedTo" + desiredCorrelation + "_length_"+ finalLength + "_pairs.csv"));

        String algStorePath = storagePrefix + "SketchAlg.alg";
        String mapDBStorageFile = storagePrefix + "mapDBStorage.db";

        String pivotFile = "c:\\Data\\2015_Doktorske\\Desc\\profiset\\caffe\\pivots-random-512.data";

        String sampleSetFile = "c:\\Data\\2015_Doktorske\\Desc\\profiset\\caffe\\profi-neuralnet-100K.data.gz";

        String className = ObjectFloatVectorNeuralNetworkL2.class.getName();

        String csvFileWithPivotInfo = storagePrefix + "75_25_Pivots_512_oriented.csv";

        Properties props = new Properties();
        props.put("objectToSketchTranformatorClass", GenHyperplane.class.getName());
        props.put("finalSketchSize", Integer.toString(finalLength));
        props.put("paralelization", "8");
        props.put("candidatesK", "20000");
        props.put("mapDBStorageFile", mapDBStorageFile);
        props.put("pivotFile", pivotFile);
        props.put("csvFileWithPivotInfo", csvFileWithPivotInfo);
        props.put("objectClass", className);
        props.put("uniquePivots", "false");
        props.put("sampleSetFile", sampleSetFile);
        props.put("bitCandidatesCountForCorrelation", "16384");
        props.put("desiredCorrelation", Float.toString(desiredCorrelation));
        props.put("useFullMapper", "true");

        createNewAlgorithmStoreExit(props, -1, algStorePath, args[0]);
//        restoreAlgorithmAndEvaluateExample(props, algStorePath);
    }

    @SuppressWarnings("FinalizeCalledExplicitly")
    public static void createNewAlgorithmStoreExit(Properties props, int dataCount, String algStorePath, String dataFilePath) throws Throwable {
        SketchAlgorithmWithRefinement sketchAlgorithm = new SketchAlgorithmWithRefinement("SketchAlgorithm", props);
//        StreamGenericAbstractObjectIterator<LocalAbstractObject> dataIt = Tools.getIterator(props.getProperty("objectClass"), dataFilePath);
//        insertData(dataIt, dataCount, sketchAlgorithm);
//        sketchAlgorithm.printAlgorithmInfo();
//        sketchAlgorithm.storeToFile(algStorePath);
        sketchAlgorithm.finalize();
        System.exit(0);
    }

    @SuppressWarnings("FinalizeCalledExplicitly")
    public static void restoreAlgorithmAndEvaluateExample(Properties props, String algStorePath) throws Throwable {
        SketchAlgorithmWithRefinement sketchAlgorithm = (SketchAlgorithmWithRefinement) restoreAlgorithm(algStorePath);
        sketchAlgorithm.setOperationsThreadPool(Executors.newFixedThreadPool(32)); // paralelization
        AbstractObjectList<LocalAbstractObject> queries = Tools.getObjects(props.getProperty("objectClass"), storagePrefix + "profi-neuralnet-1000-query.data", 10);
        for (LocalAbstractObject query : queries) {
            ApproxKNNQueryOperation knn = new ApproxKNNQueryOperation(query, 1000, AnswerType.ORIGINAL_OBJECTS, 10, Approximate.LocalSearchType.ABS_OBJ_COUNT, -1);
            knn = sketchAlgorithm.executeOperation(knn);
            Tools.printAnswer(knn, System.out);
        }
        sketchAlgorithm.printAlgorithmInfo();
        sketchAlgorithm.finalize();
    }

    @SuppressWarnings("FinalizeCalledExplicitly")
    public static SketchObjectMapperAlgorithm createSketchObjectMapperAlgorithm(Iterator<LocalAbstractObject> data, long dataCount, LocalAbstractObject[] pivots, ObjectToSketchTransformator sketchCreator, String pathToSave) {
        try {
            if (dataCount < 0) {
                dataCount = Long.MAX_VALUE;
            }
            SketchObjectMapperAlgorithm ret = new SketchObjectMapperAlgorithm(sketchCreator);
            long batchCount = (long) Math.ceil(dataCount / 10000d);
            for (int i = 0; i < batchCount; i++) {
                System.gc();
                AbstractObjectList batchData = new AbstractObjectList(data, 6250);
                if (batchData.isEmpty()) {
                    break;
                }
                ret.executeOperation(new BulkInsertOperation(batchData));
            }
            ret.storeToFile(pathToSave);
            ret.finalize();
            return ret;
        } catch (Throwable ex) {
            Logger.getLogger(KNNApproxEvaluation.class.getName()).log(Level.SEVERE, null, ex);
        }
        return null;
    }

    public static Algorithm restoreAlgorithm(String path) {
        try {
            return Algorithm.restoreFromFile(path);
        } catch (IOException | NullPointerException | ClassNotFoundException | ClassCastException ex) {
            Logger.getLogger(KNNApproxEvaluation.class.getName()).log(Level.SEVERE, null, ex);
        }
        return null;
    }


    private static void insertData(StreamGenericAbstractObjectIterator<LocalAbstractObject> dataIt, int dataCount, SketchAlgorithmWithRefinement sketchAlgorithm) throws AlgorithmMethodException, NoSuchMethodException {
        int batchCount;
        if (dataCount > 0) {
            batchCount = dataCount / 10000;
        } else {
            batchCount = Integer.MAX_VALUE;
        }
        for (int i = 0; i < batchCount; i++) {
            AbstractObjectList batchList = new AbstractObjectList(dataIt, 10000);
            if (batchList.isEmpty()) {
                break;
            }
            System.gc();
            sketchAlgorithm.executeOperation(new BulkInsertOperation(batchList));
        }
    }
}

//        ByteArrayOutputStream baos = new ByteArrayOutputStream();
//        ObjectOutputStream oos = new ObjectOutputStream(baos);
//        int last = 0;
//        Random r = new Random();
//        for (int i = 0; i < 10; i++) {
//            Sketch sketch = new Sketch(64);
//            long l = r.nextLong();
//            sketch.tmpCreateFromLong(l);
//            oos.writeObject(sketch);
//            System.out.println(baos.size() - last);
//            last = baos.size();
//        }
//        oos.close();
//        List<Sketch> list = new ArrayList();
//        long heapSize = Runtime.getRuntime().totalMemory();
//        System.out.println(heapSize);
//        System.out.println(Runtime.getRuntime().maxMemory());
//        for (int i = 0; i < 1000000; i++) {
//            Sketch sketch = new Sketch(64);
//            long l = r.nextLong();
//            sketch.tmpCreateFromLong(l);
//            list.add(sketch);
//        }
//        System.out.println(Runtime.getRuntime().maxMemory());
//        System.out.println("Diff: " + (Runtime.getRuntime().totalMemory() - heapSize));
//        System.exit(0);
