/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package main.tools;

import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;
import messif.algorithms.Algorithm;
import messif.algorithms.impl.SequentialScan;
import messif.objects.LocalAbstractObject;
import messif.objects.util.RankedAbstractObject;
import messif.operations.data.BulkInsertOperation;
import messif.operations.query.KNNQueryOperation;
import sf.objects.impl.Sketch;

/**
 *
 * @author xmic
 */
public class EvaluateRecallOfSketchFiltering {

    private final List<KNNResultFileParser.EvaluatedQuery> groundTruth;
    private final List<Sketch> sketches;
    private final Map<String, LocalAbstractObject> queries;

    public EvaluateRecallOfSketchFiltering(String groundTruthPath, String sketchDataPath, String sketchQueriesPath) {
        KNNResultFileParser parser = new KNNResultFileParser(groundTruthPath);
        this.groundTruth = parser.getEvaluatedQueries();
        this.queries = Tools.getObjectsAsLocatorMap(Sketch.class.getName(), sketchQueriesPath, -1);
        this.sketches = Tools.getObjects(Sketch.class.getName(), sketchDataPath);
    }

    @SuppressWarnings("FinalizeCalledExplicitly")
    public float[] evaluateAverageRecall(int k, int[] candSetSizes) {
        float[] recallSums = new float[candSetSizes.length];
        try {
            Algorithm alg = new SequentialScan();
            BulkInsertOperation insert = new BulkInsertOperation(sketches);
            alg.executeOperation(insert);
            Iterator<KNNResultFileParser.EvaluatedQuery> it = groundTruth.iterator();
            while (it.hasNext()) {
                KNNResultFileParser.EvaluatedQuery groundTruthQuery = it.next();
                Sketch querySketch = (Sketch) queries.get(groundTruthQuery.getLocatorURI());
                KNNQueryOperation knn = new KNNQueryOperation(querySketch, candSetSizes[candSetSizes.length - 1]);
                knn = alg.executeOperation(knn);
                float[] evaluatedRecallsForQuery = evaluateRecallsForQuery(k, groundTruthQuery.getNearestNeighbours(), knn.getAnswer(), candSetSizes);
                for (int i = 0; i < recallSums.length; i++) {
                    recallSums[i] += evaluatedRecallsForQuery[i];
                }
            }
            alg.finalize();
            alg.destroy();
        } catch (Throwable ex) {
            Logger.getLogger(EvaluateRecallOfSketchFiltering.class.getName()).log(Level.SEVERE, null, ex);
        }
        for (int i = 0; i < recallSums.length; i++) {
            recallSums[i] = recallSums[i] / groundTruth.size();
        }
        return recallSums;
    }

    private float[] evaluateRecallsForQuery(int k, List<Integer> groundTruthNN, Iterator<RankedAbstractObject> candSet, int[] candSetSizes) {
        Set<Integer> candidates = new HashSet();
        float[] ret = new float[candSetSizes.length];
        int curPos = 0;
        int currCandSetSize = candSetSizes[curPos];
        while (candSet.hasNext()) {
            candidates.add(Integer.parseInt(candSet.next().getObject().getLocatorURI()));
            if (candidates.size() == currCandSetSize) {
                int presentObjects = 0;
                for (int i = 0; i < k; i++) {
                    if (candidates.contains(groundTruthNN.get(i))) {
                        presentObjects++;
                    }
                }
                ret[curPos] = presentObjects / (float) k;
                System.out.print(ret[curPos] + ";");
                curPos++;
                if (curPos < candSetSizes.length) {
                    currCandSetSize = candSetSizes[curPos];
                }
            }
        }
        System.out.println("");
        return ret;
    }
}
