package main.tools;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.PrintStream;
import java.util.Iterator;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;
import messif.objects.LocalAbstractObject;
import messif.objects.impl.ObjectFloatVectorNeuralNetworkL2;
import messif.objects.util.StreamGenericAbstractObjectIterator;
import sf.pivotAnalyzer.PivotSubsetSelector;

/**
 *
 * @author xmic
 */
// args[0] sample set file
// args[1] sample set object count
// args[2] class name
// args[3] pivots file
public class PrintHyperplanePivotsWellDividingSpace {

    public static void main(String[] args) {
        String className = ObjectFloatVectorNeuralNetworkL2.class.getName();
        String sampleSetFile = "c:\\Data\\2015_Doktorske\\Desc\\profiset\\caffe\\profi-neuralnet-100K.data.gz";
        String pivotFile = "c:\\Data\\2015_Doktorske\\Desc\\profiset\\caffe\\pivots-random-512.data";
        filterPivotsWellDividing(
                "c:\\Data\\2016\\2016_Unbalanced_sketches\\Decaf\\75_25\\75_25_Pivots_512_oriented.csv",
                "c:\\Data\\2016\\2016_Unbalanced_sketches\\Decaf\\75_25\\75_25_Pivots_512_oriented.err",
                className, sampleSetFile, -1, pivotFile);
//        countPivotDistancesToHyperplane("c:\\Data\\2015_Doktorske\\Desc\\pivots_filtered_hyperplane\\DistancesToHyperplane.err", args[1], args[0], storagePrefix + "pivots_filtered_hyperplane\\50_50.data");
    }

    private static void setErr(String path) {
        try {
            System.setErr(new PrintStream(path));
        } catch (FileNotFoundException ex) {
            Logger.getLogger(PrintHyperplanePivotsWellDividingSpace.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    private static void filterPivotsWellDividing(String output, String err, String className, String sampleSetFile, int sampleSetObjectCount, String pivotsFile) {
        if (err != null) {
            setErr(err);
        }
        Iterator<LocalAbstractObject> sampleSetIt = Tools.getObjects(className, sampleSetFile, sampleSetObjectCount).iterator();
        LocalAbstractObject[] pivots = (LocalAbstractObject[]) Tools.getObjects(className, pivotsFile, -1).toArray(new LocalAbstractObject[0]);
        PivotSubsetSelector selector = new PivotSubsetSelector(pivots);
        Map<LocalAbstractObject[], Float> selectedPivots = selector.getPivotsWithProperHyperplaneDivision(0.23f, 0.27f, sampleSetIt);
        if (output != null) {
            Tools.print(selectedPivots, new File(output));
        } else {
            Tools.print(selectedPivots, System.out);
        }
    }

    private static void countPivotDistancesToHyperplane(String err, String className, String sampleSetFile, String pivotsFile) {
        setErr(err);
        StreamGenericAbstractObjectIterator<LocalAbstractObject> sampleSetIt = Tools.getIterator(className, sampleSetFile);
        StreamGenericAbstractObjectIterator<LocalAbstractObject> pivotsIt = Tools.getIterator(className, pivotsFile);
        PivotSubsetSelector selector = new PivotSubsetSelector(pivotsIt);
        selector.getPivotsDistancesFromHyperplane(sampleSetIt, pivotsIt);
    }
}
