/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package sf.algorithm.HWTOP;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.TreeSet;
import messif.objects.util.RankedSortedCollection;
import messif.operations.AnswerType;
import sf.objects.impl.Sketch;

/**
 *
 * @author xmic
 */
public class Node implements Serializable, Comparable<Node> {

    private static final long serialVersionUID = 159756320894L;
    private static int nodeCounter = 0;
    private final int id;

    private final List<Sketch> sketches;
    private TreeSet<Node> children;
    private final int depth;
    private final int splittingInChildren;

    final int[] nodeAssignment;
    int lowerBoundOnSketchesInSubstree;

    private String lastQueryUri = null;

    public Node(int depth, int maxDepth, int[] nodeAssignment) {
        this.depth = depth;
        this.nodeAssignment = nodeAssignment;
        if (nodeAssignment == null) {
            splittingInChildren = 1;
        } else {
            splittingInChildren = nodeAssignment.length * 2;
        }
        lowerBoundOnSketchesInSubstree = Integer.MAX_VALUE;
        if (depth == maxDepth) {
            sketches = new ArrayList<>();
            children = null;
        } else {
            sketches = null;
            children = new TreeSet<>();
        }
        id = nodeCounter;
        nodeCounter++;
    }

    public void insertSketch(Sketch[] sketchesPerLevels, int sketchLength) {
        if (sketches != null) {
            sketches.add(sketchesPerLevels[depth - 1]);
            return;
        }
        int[] childKey = sketchesPerLevels[depth].getCardinalitiesInBlocks(splittingInChildren, sketchLength);
        Node nodeForSketch = new Node(depth + 1, sketchesPerLevels.length, childKey);
        if (children.contains(nodeForSketch)) {
            nodeForSketch = children.floor(nodeForSketch);
        } else {
            children.add(nodeForSketch);
        }
        nodeForSketch.insertSketch(sketchesPerLevels, sketchLength);
    }

    public List<Integer> getNumberOfObjectsInLeaves() {
        List<Integer> ret = new ArrayList<>();
        if (sketches != null) {
            ret.add(sketches.size());
            return ret;
        }
        Iterator<Node> it = children.iterator();
        while (it.hasNext()) {
            ret.addAll(it.next().getNumberOfObjectsInLeaves());
        }
        return ret;
    }

    public float knnSearch(String queryLocator, Sketch[] queries, int sketchLength, RankedSortedCollection answer, AnswerType type, float answerDistance, int lowerBoundToSearch) {
        if (sketches != null) {
            Main.t2 -= System.currentTimeMillis();
            Iterator<Sketch> it = sketches.iterator();
            while (it.hasNext()) {
                Sketch sketch = it.next();
                float distance = queries[depth - 1].getDistance(sketch);
                answer.add(type, sketch, distance, null);
                HWTOPAlgorithm.distCounts++;
            }
            Main.t2 += System.currentTimeMillis();
            lowerBoundOnSketchesInSubstree = sketchLength;
            return answer.isFull() ? answer.getLastDistance() : answerDistance;
        }

        Sketch querySketchForNextLevel = queries[depth];
        if (lastQueryUri == null || !lastQueryUri.equals(queryLocator)) {
            adjustChildren(querySketchForNextLevel, sketchLength);
            lastQueryUri = queryLocator;
        }

        Main.t4 -= System.currentTimeMillis();
        lowerBoundOnSketchesInSubstree = children.first().lowerBoundOnSketchesInSubstree;
        Main.t4 += System.currentTimeMillis();
        while (lowerBoundOnSketchesInSubstree < answerDistance && lowerBoundOnSketchesInSubstree <= lowerBoundToSearch) {
            Main.t4 -= System.currentTimeMillis();
            Node nodeToProcess = children.first();
            children.remove(nodeToProcess);
            Main.t4 += System.currentTimeMillis();
            answerDistance = nodeToProcess.knnSearch(queryLocator, queries, sketchLength, answer, type, answerDistance, lowerBoundToSearch);
            lowerBoundToSearch = (int) answerDistance;
            Main.t4 -= System.currentTimeMillis();
            children.add(nodeToProcess);
//            for (Node child : children) {
//                System.out.println("After processing, child: " + child.lowerBoundOnSketchesInSubstree);
//            }
//            System.out.println("\n\n");
//
            lowerBoundOnSketchesInSubstree = children.first().lowerBoundOnSketchesInSubstree;
            Main.t4 += System.currentTimeMillis();
        }
        return answerDistance;
    }

    public void getStructure(Map<Integer, Integer> map) {
        if (children != null) {
            int count = 0;
            if (map.containsKey(splittingInChildren)) {
                count = map.get(splittingInChildren);
            }
            map.put(splittingInChildren, count + children.size());
            for (Node child : children) {
                child.getStructure(map);
            }
        }
    }

    @Override
    public int hashCode() {
        return Arrays.hashCode(this.nodeAssignment);
    }

    @Override
    public boolean equals(Object obj) {
        if (obj == null) {
            return false;
        }
        if (getClass() != obj.getClass()) {
            return false;
        }
        final Node other = (Node) obj;
        return Arrays.equals(this.nodeAssignment, other.nodeAssignment);
    }

    private void adjustChildren(Sketch querySketchForNextLevel, int sketchLength) {
        Main.t1 -= System.currentTimeMillis();
        int[] queryKeyInNextLevel = querySketchForNextLevel.getCardinalitiesInBlocks(splittingInChildren, sketchLength);
        Main.t1 += System.currentTimeMillis();
        Main.t3 -= System.currentTimeMillis();
        TreeSet<Node> newSortedChildren = new TreeSet<>();
        for (Node child : children) {
            int newLowerBound = 0;
            for (int i = 0; i < queryKeyInNextLevel.length; i++) {
                newLowerBound += Math.abs(queryKeyInNextLevel[i] - child.nodeAssignment[i]);
            }
            child.lowerBoundOnSketchesInSubstree = newLowerBound;
            newSortedChildren.add(child);
        }
        children = newSortedChildren;

//        for (Node child : children) {
//            System.out.println("Adjust children, child: " + child.lowerBoundOnSketchesInSubstree);
//        }
//        System.out.println("\n\n\n\n");

        Main.t3 += System.currentTimeMillis();
    }

    @Override
    public int compareTo(Node o) {
        if (lowerBoundOnSketchesInSubstree == Integer.MAX_VALUE || o.lowerBoundOnSketchesInSubstree == Integer.MAX_VALUE) {
            for (int i = 0; i < nodeAssignment.length; i++) {
                int ret = Integer.compare(nodeAssignment[i], o.nodeAssignment[i]);
                if (ret != 0) {
                    return ret;
                }
            }
            return 0;
        }
        if (lowerBoundOnSketchesInSubstree != o.lowerBoundOnSketchesInSubstree) {
            return (lowerBoundOnSketchesInSubstree < o.lowerBoundOnSketchesInSubstree) ? -1 : 1;
        }
        return (id < o.id) ? -1 : ((id == o.id) ? 0 : 1);
    }

}
