/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package sf.algorithm.HWTOP_nefunckni;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.BitSet;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.SortedMap;
import java.util.TreeMap;
import main.tools.Tools;
import messif.objects.util.RankedSortedCollection;
import messif.operations.AnswerType;
import sf.objects.impl.Sketch;

/**
 *
 * @author xmic
 */
public class Node implements Serializable {

    private static final long serialVersionUID = 159756321894L;

    private final List<Sketch> sketches;
    private final Map<int[], Node> children;
    private final int depth;
    private final int splittingInChildren;
    private static int[][] bitsPermutationsForTreeLevels;
    private static BitSet[] bitsToOrientInTreeLevels;
    private static int sketchLength;

    public Node(int depth) {
        this.depth = depth;
        splittingInChildren = (int) Math.pow(2, depth);
        if (depth == bitsPermutationsForTreeLevels.length) {
            sketches = new ArrayList<>();
            children = null;
        } else {
            sketches = null;
            children = new TreeMap<>(new Tools.IntArraySameLengthsComparator());
        }
    }

    public static void setBitsPermutationsForTreeLevels(int[][] bitsPermutationsForTreeLevels) {
        Node.bitsPermutationsForTreeLevels = bitsPermutationsForTreeLevels;
    }

    public static void setBitsToOrientInTreeLevels(BitSet[] bitsToOrientInTreeLevels) {
        Node.bitsToOrientInTreeLevels = bitsToOrientInTreeLevels;
    }

    public static void setSketchLength(int sketchLength) {
        Node.sketchLength = sketchLength;
    }

    public void insertSketch(Sketch sketch) {
        if (sketches != null) {
            // TO DO: splitting?
            Sketch modifiedSketch = Tools.orientSketch(sketch, bitsToOrientInTreeLevels[depth - 1], bitsPermutationsForTreeLevels[depth - 1]);
            sketches.add(modifiedSketch);
            return;
        }
        Sketch modifiedSketch = Tools.orientSketch(sketch, bitsToOrientInTreeLevels[depth], bitsPermutationsForTreeLevels[depth]);
        int[] childKey = modifiedSketch.getCardinalitiesInBlocks(splittingInChildren, sketchLength);
        Tools.printArray(childKey, ",");;
        System.out.println("");
        if (!children.containsKey(childKey)) {
            children.put(childKey, new Node(depth + 1));
        }
        children.get(childKey).insertSketch(sketch);
    }

    public List<Integer> getNumberOfObjectsInLeaves() {
        List<Integer> ret = new ArrayList<>();
        if (sketches != null) {
            ret.add(sketches.size());
            return ret;
        }
        Iterator<Map.Entry<int[], Node>> it = children.entrySet().iterator();
        while (it.hasNext()) {
            ret.addAll(it.next().getValue().getNumberOfObjectsInLeaves());
        }
        return ret;
    }

    protected SortedMap<Integer, List<int[]>> childrenSortedForKnnQuery = null;
    private String lastQueryUri = null;

    public Float knnSearch(Sketch queryOrig, Sketch queryForThisLevel, RankedSortedCollection answer, AnswerType type, Float answerDistance, int lowerBoundToSearch) {
        if (sketches != null) {
            Main.t2 -= System.currentTimeMillis();
            Iterator<Sketch> it = sketches.iterator();
            while (it.hasNext()) {
                Sketch sketch = it.next();
                float distance = queryForThisLevel.getDistance(sketch);
                answer.add(type, sketch, distance, null);
                RootNode.distCounts++;
            }
            Main.t2 += System.currentTimeMillis();
            return answer.getLastDistance();
//            return 259f;
        }

        Sketch querySketchForNextLevel = null;
        if (childrenSortedForKnnQuery == null || lastQueryUri == null || !lastQueryUri.equals(queryOrig.getLocatorURI())) {
            querySketchForNextLevel = Tools.orientSketch(queryOrig, bitsToOrientInTreeLevels[depth], bitsPermutationsForTreeLevels[depth]);
            Main.t1 -= System.currentTimeMillis();
            int[] queryKeyInNextLevel = querySketchForNextLevel.getCardinalitiesInBlocks(splittingInChildren, sketchLength);
            Main.t1 += System.currentTimeMillis();
//            System.out.print("Query: ");;
//            Tools.printArray(queryKeyInNextLevel, ",");
//            System.out.println("");
            childrenSortedForKnnQuery = Tools.sortObjectKeys(children, queryKeyInNextLevel);
            lastQueryUri = queryOrig.getLocatorURI();
        }
        Iterator<Integer> itOverLowerBounds = childrenSortedForKnnQuery.keySet().iterator();

        Set<Integer> lowerBoundsProcessed = new HashSet<>();
        while (itOverLowerBounds.hasNext()) {
            int lowerBound = itOverLowerBounds.next();
            if (lowerBound > lowerBoundToSearch) {
                return -1f;
            }
            List<int[]> keys = childrenSortedForKnnQuery.get(lowerBound);
            Set<int[]> keysToRemove = new HashSet<>();
            int tmp = 0;;
            if (keys != null) {
                for (int[] key : keys) {
                    tmp = key.length;
//                    System.out.print("Check: ");;
//                    Tools.printArray(key, ",");
//                    System.out.println(" (dist " + lowerBound + ")");

                    if (lowerBoundToSearch < answerDistance) {
                        if (querySketchForNextLevel == null) {
                            querySketchForNextLevel = Tools.orientSketch(queryOrig, bitsToOrientInTreeLevels[depth], bitsPermutationsForTreeLevels[depth]);
                        }
                        int v = Main.numberOfVisitedNodes.get(key.length);;
                        Main.numberOfVisitedNodes.put(key.length, v + 1);

                        float ansPartial = children.get(key).knnSearch(queryOrig, querySketchForNextLevel, answer, type, answerDistance, lowerBoundToSearch);
                        if (ansPartial >= 0) {
                            answerDistance = ansPartial;;
                            keysToRemove.add(key);
                            int t = Main.numberOfCompletelyProcessedNodes.get(tmp);;
                            Main.numberOfCompletelyProcessedNodes.put(tmp, t + 1);
                        }
                        if (lowerBoundToSearch >= answerDistance) {
                            break;
                        }
                    }
                }
                if (!keysToRemove.isEmpty()) {
                    keys.removeAll(keysToRemove);
                }
                if (keys.isEmpty()) {
                    lowerBoundsProcessed.add(lowerBound);
                }
            }
        }
        for (Integer lowerBoundProcessed : lowerBoundsProcessed) {
            childrenSortedForKnnQuery.remove(lowerBoundProcessed);
        }
        return answerDistance;
    }

    public void getStructure(Map<Integer, Integer> map) {
        if (children != null) {
            int length = children.keySet().iterator().next().length;
            int count = 0;
            if (map.containsKey(length)) {
                count = map.get(length);
            }
            map.put(length, count + children.size());
            for (Node child : children.values()) {
                child.getStructure(map);
            }
        }
    }
}
