/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package sf.algorithm.HWTOP_nefunckni;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.BitSet;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;
import java.util.logging.Level;
import java.util.logging.Logger;
import main.tools.Tools;
import messif.algorithms.Algorithm;
import messif.objects.LocalAbstractObject;
import messif.objects.util.RankedAbstractObject;
import messif.objects.util.RankedSortedCollection;
import messif.operations.data.BulkInsertOperation;
import messif.operations.query.KNNQueryOperation;
import sf.algorithm.myHWT.OrientAndShuffleBitsForHWT;
import sf.objects.impl.Sketch;

/**
 *
 * @author xmic
 */
public class RootNode extends Algorithm implements Serializable {

    private static final long serialVersionUID = 159756321895L;

    public static Integer distCounts = 0;

    public static final Integer IMPLICIT_MAX_NUMBER_OF_OBJECTS_PER_LEAF = 200;
    private final Map<Integer, Node> children;
    private final BitSet[] bitsToOrientInTreeLevels;
    private final int[][] bitsPermutationsForTreeLevels;

    private int sketchLength;

    /**
     *
     * @param sampleOfSketches
     * @param maxDepth max depth of tree, this root is level -1, i.e. the
     * sketches are split into 2^n parts in level n
     * @param maxNumberOfObjectsPerLeaf
     * @param sketchLength
     */
    public RootNode(List<Sketch> sampleOfSketches, int maxDepth, int maxNumberOfObjectsPerLeaf, int sketchLength) {
        super("HWT with oriented and permutated bits");
        Logger.getLogger(RootNode.class.getName()).log(Level.INFO, "Creating tree that splits sketches to {0} parts", Math.pow(2, maxDepth - 1));
        bitsToOrientInTreeLevels = new BitSet[maxDepth];
        bitsPermutationsForTreeLevels = new int[maxDepth][sketchLength];
        children = new HashMap<>();
        this.sketchLength = sketchLength;
        learnOrientationAndPermutation(sampleOfSketches, maxDepth, sketchLength);
    }

    public RootNode(List<Sketch> sampleOfSketches, int maxDepth, int sketchLength) {
        this(sampleOfSketches, maxDepth, IMPLICIT_MAX_NUMBER_OF_OBJECTS_PER_LEAF, sketchLength);
    }

    private void learnOrientationAndPermutation(List<Sketch> sampleOfSketches, int maxDepth, int sketchLength) {
        Logger.getLogger(RootNode.class.getName()).log(Level.INFO, "Starting learning the bit orientations and permutations");
        float[][] corrMatrix = Tools.evaluateCorrelationMatrix(sampleOfSketches);
        for (int i = 0; i < maxDepth; i++) {
            OrientAndShuffleBitsForHWT o = new OrientAndShuffleBitsForHWT(Tools.copyMatrix(corrMatrix), (int) Math.pow(2, i), sketchLength);
            bitsToOrientInTreeLevels[i] = o.evaluateIndexesOfBitsToSwitch();
            bitsPermutationsForTreeLevels[i] = o.getNewBitsPermutation();
//            bitsPermutationsForTreeLevels[i] = o.getOldBitsPermutation();
        }
        Node.setBitsPermutationsForTreeLevels(bitsPermutationsForTreeLevels);
        Node.setBitsToOrientInTreeLevels(bitsToOrientInTreeLevels);
        Node.setSketchLength(sketchLength);
//        Node.setMaxNumberOfObjectsPerLeaf(maxNumberOfObjectsPerLeaf);
        Logger.getLogger(RootNode.class.getName()).log(Level.INFO, "Learning the bit orientations and permutations finished");
    }

    private void readObject(ObjectInputStream in) throws IOException, ClassNotFoundException {
        in.defaultReadObject();
        Node.setBitsPermutationsForTreeLevels(bitsPermutationsForTreeLevels);
        Node.setBitsToOrientInTreeLevels(bitsToOrientInTreeLevels);
        Node.setSketchLength(sketchLength);
    }

    public void insertSketches(BulkInsertOperation op) {
        Iterator<? extends LocalAbstractObject> it = op.getInsertedObjects().iterator();
        int counter = 0;
        while (it.hasNext()) {
            Sketch sketchOrig = (Sketch) it.next();
            Sketch modifiedSketch = Tools.orientSketch(sketchOrig, bitsToOrientInTreeLevels[0], null);
            System.out.println("\n" + sketchOrig.getLocatorURI());;
            int card = modifiedSketch.getCardinality();
            if (!children.containsKey(card)) {
                children.put(card, new Node(1));
            }
            children.get(card).insertSketch(sketchOrig);
            counter++;
            if (counter % 100000 == 0) {
                Logger.getLogger(RootNode.class.getName()).log(Level.INFO, "Inserted {0} objects", counter);
            }
        }
        op.endOperation();
    }

    public List<Integer> getNumberOfObjectsInLeaves() {
        List<Integer> ret = new ArrayList<>();
        Iterator<Map.Entry<Integer, Node>> it = children.entrySet().iterator();
        while (it.hasNext()) {
            ret.addAll(it.next().getValue().getNumberOfObjectsInLeaves());
        }
        return ret;
    }

    /**
     * Returns map where key is the number of parts in split sketches and value
     * is number of nodes with key with the given number of partes
     *
     * @return
     */
    public Map<Integer, Integer> getTreeStructureDescription() {
        Map<Integer, Integer> ret = new TreeMap<>();
        ret.put(1, children.size());
        for (Map.Entry<Integer, Node> child : children.entrySet()) {
            child.getValue().getStructure(ret);
        }
        return ret;
    }

    public KNNQueryOperation knnSearch(KNNQueryOperation op) {
//        System.out.println("\n\n\nStart:");
        Sketch queryOrig = (Sketch) op.getQueryObject();
        Sketch queryForNextLevel = Tools.orientSketch(queryOrig, bitsToOrientInTreeLevels[0], null);
        int cardQuery = queryForNextLevel.getCardinality();
        int k = op.getK();
        RankedSortedCollection answer = new RankedSortedCollection(k, k);

        Float answerDistance = Sketch.MAX_DISTANCE;
        Set<Integer> childrenToProcess = new HashSet<>(children.keySet());
        int range = 0;

        while (range < answerDistance && !childrenToProcess.isEmpty()) {
            int min = Math.max(0, cardQuery - range);
            int max = cardQuery + range;
            for (int childCard = min; childCard <= max; childCard++) {
                if (range < answerDistance && min >= 0 && childrenToProcess.contains(childCard)) {
                    Node child = children.get(childCard);
//                    System.out.println("\nCheck: " + childCard + " (range " + range + ")");
                    int v = Main.numberOfVisitedNodes.get(1);
                    Main.numberOfVisitedNodes.put(1, v + 1);
                    float ansCand = child.knnSearch(queryOrig, queryForNextLevel, answer, op.getAnswerType(), answerDistance, range);
                    if (ansCand >= 0) {
                        answerDistance = ansCand;
                    }
                    if (child.childrenSortedForKnnQuery.isEmpty()) {
                        childrenToProcess.remove(childCard);
                        int t = Main.numberOfCompletelyProcessedNodes.get(1);
                        Main.numberOfCompletelyProcessedNodes.put(1, t + 1);
                    }
                }
            }
            range++;
        }
        Main.t4 -= System.currentTimeMillis();
        Iterator<RankedAbstractObject> it = answer.iterator();
        while (it.hasNext()) {
            RankedAbstractObject obj = it.next();
            op.addToAnswer(obj.getObject(), obj.getDistance(), null);
        }
        Main.t4 += System.currentTimeMillis();
        op.endOperation();

//        for (Map.Entry<Integer, Node> child : children.entrySet()) {
//            int size = child.getValue().childrenSortedForKnnQuery.size();
//            if (size != 0) {
//                String s = "achtung";
//            }
//        }
//        System.out.println(distCounts + "Konec\n\n\n\n\n\n\n\n\n\n\n\n");
        return op;
    }

}
