/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package sf.algorithm.myHWT;

import java.util.BitSet;
import java.util.Iterator;
import java.util.List;
import java.util.SortedSet;
import java.util.TreeSet;
import sf.objects.impl.Sketch;

/**
 *
 * @author xmic
 */
public class OrientAndShuffleBitsForHWT extends OrientBitsForHWTSumOfCorrOverParts {

    protected final int[] newBitsPermutation;

    public OrientAndShuffleBitsForHWT(float[][] correlationMatrix, int numberOfparts, int sketchLength) {
        super(correlationMatrix, numberOfparts, sketchLength);
        newBitsPermutation = new int[sketchLength];
    }

    public OrientAndShuffleBitsForHWT(List<Sketch> sketches, int numberOfparts, int sketchLength) {
        super(sketches, numberOfparts, sketchLength);
        newBitsPermutation = new int[sketchLength];
    }

    @Override
    public BitSet evaluateIndexesOfBitsToSwitch() {
        int blockSize = sketchLength / numberOfparts;
        SortedSet<Integer> indexesToProcess = new TreeSet<>();
        for (int i = 0; i < sketchLength; i++) {
            indexesToProcess.add(i);
        }
        BitSet ret = new BitSet();
        for (int blockCounter = 0; !indexesToProcess.isEmpty(); blockCounter++) {
            int[] indexesOfBitsForCurrBlock = new int[blockSize];
            System.arraycopy(findMaxCorr(indexesToProcess), 0, indexesOfBitsForCurrBlock, 0, 2);
//            indexesOfBitsForCurrBlock[0] = indexesToProcess.first();
            if (indexesOfBitsForCurrBlock[1] < 0 && indexesOfBitsForCurrBlock[1] != -Integer.MAX_VALUE) {
                setIndexToSwitch(indexesOfBitsForCurrBlock[1], ret);
                indexesOfBitsForCurrBlock[1] = Math.abs(indexesOfBitsForCurrBlock[1]);
            }
            indexesToProcess.remove(indexesOfBitsForCurrBlock[0]);
            indexesToProcess.remove(indexesOfBitsForCurrBlock[1]);
            for (int i = 2; i < indexesOfBitsForCurrBlock.length; i++) {
                indexesOfBitsForCurrBlock[i] = -1;
                Object[] indexOfBitWithMaxCorr = findIndexWithMaxCorr(indexesOfBitsForCurrBlock, indexesToProcess);
                indexesOfBitsForCurrBlock[i] = (int) indexOfBitWithMaxCorr[0];
                if ((boolean) indexOfBitWithMaxCorr[1]) {
                    setIndexToSwitch((Integer) indexOfBitWithMaxCorr[0], ret);
                }
            }
//            Arrays.sort(indexesOfBitsForCurrBlock);
            int lengthToCopy = Math.min(indexesOfBitsForCurrBlock.length, sketchLength - blockCounter * blockSize);
            System.arraycopy(indexesOfBitsForCurrBlock, 0, newBitsPermutation, blockCounter * blockSize, lengthToCopy);
        }
        return ret;
    }

    private Object[] findIndexWithMaxCorr(int[] indexesOfBitsForCurrBlock, SortedSet<Integer> indexesToProcess) {
        Iterator<Integer> it = indexesToProcess.iterator();
        float maxSumCorrFound = -Float.MAX_VALUE;
        int idxToRet = -1;
        boolean flipBit = false;
        while (it.hasNext()) {
            int cand = it.next();
            float sumForCand = 0;
            for (int i = 0; i < indexesOfBitsForCurrBlock.length && indexesOfBitsForCurrBlock[i] != -1; i++) {
                sumForCand += bestFoundCorrelationMatrix[cand][indexesOfBitsForCurrBlock[i]];
            }
            if (maxSumCorrFound < Math.abs(sumForCand)) {
                idxToRet = cand;
                maxSumCorrFound = Math.abs(sumForCand);
                flipBit = sumForCand < 0;
            }
        }
        indexesToProcess.remove(idxToRet);
        return new Object[]{idxToRet, flipBit};
    }

    /**
     * ret[i] is position in orig sketch which should be placed to ith in new
     * sketch
     *
     * @return
     */
    public int[] getNewBitsPermutation() {
        return newBitsPermutation;
    }

    /**
     * i is considered as the position in old sketch and ret[i] is the position
     * in new sketch which should be placed value in ith old bit
     *
     * @return
     */
    public int[] getOldBitsPermutation() {
        int[] ret = new int[newBitsPermutation.length];
        for (int i = 0; i < newBitsPermutation.length; i++) {
            for (int pos = 0; pos < newBitsPermutation.length; pos++) {
                if (newBitsPermutation[pos] == i) {
                    ret[i] = pos;
                    break;
                }
            }
        }
        return ret;
    }

    private int[] findMaxCorr(SortedSet<Integer> indexesToProcess) {
        if (indexesToProcess.size() == 1) {
            return new int[]{indexesToProcess.first(), -Integer.MAX_VALUE};
        }
        float maxCorr = -1;
        int[] ret = new int[2];
        for (int i : indexesToProcess) {
            for (int j : indexesToProcess) {
                if (i != j && Math.abs(bestFoundCorrelationMatrix[i][j]) > maxCorr) {
                    maxCorr = Math.abs(bestFoundCorrelationMatrix[i][j]);
                    if (bestFoundCorrelationMatrix[i][j] < 0) {
                        ret[1] = -j;
                    } else {
                        ret[1] = j;
                    }
                    ret[0] = i;
                }
            }
        }
        return ret;
    }

    private void setIndexToSwitch(int i, BitSet bitset) {
        int abs = Math.abs(i);
        bitset.set(abs);
        bestFoundCorrelationMatrix = changeBitOrientation(abs, bestFoundCorrelationMatrix);
    }

}
