/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package sf.algorithm.myHWT;

import java.util.ArrayList;
import java.util.BitSet;
import java.util.Iterator;
import java.util.List;
import java.util.SortedSet;
import java.util.TreeSet;
import java.util.logging.Level;
import java.util.logging.Logger;
import sf.objects.impl.Sketch;

/**
 *
 * @author xmic
 */
public class OrientAndShuffleBitsForHWTDynamicLength extends OrientAndShuffleBitsForHWT {

    private final int[] clusterSizes;

    public OrientAndShuffleBitsForHWTDynamicLength(float[][] correlationMatrix, int sketchLength) {
        super(correlationMatrix, 0, sketchLength);
        clusterSizes = new int[sketchLength];
    }

    public OrientAndShuffleBitsForHWTDynamicLength(List<Sketch> sketches, int sketchLength) {
        super(sketches, 0, sketchLength);
        clusterSizes = new int[sketchLength];
    }

    @Override
    public BitSet evaluateIndexesOfBitsToSwitch() {
        int clusterId = 0;
        int totalSize = 0;
        SortedSet<Integer> indexesToProcess = new TreeSet<>();
        for (int i = 0; i < sketchLength; i++) {
            indexesToProcess.add(i);
        }
        BitSet ret = new BitSet();
        while (!indexesToProcess.isEmpty()) {
            List<Integer> indexesOfBitsForCurrBlock = new ArrayList<>();
            int[] maxCorrIndexes = findMaxCorr(indexesToProcess);
            if (maxCorrIndexes[1] < 0 && maxCorrIndexes[1] != -Integer.MAX_VALUE) {
                setIndexToSwitch(maxCorrIndexes[1], ret);
                maxCorrIndexes[1] = Math.abs(maxCorrIndexes[1]);
            }
            indexesOfBitsForCurrBlock.add(maxCorrIndexes[0]);
            if (maxCorrIndexes[1] != -Integer.MAX_VALUE) {
                indexesOfBitsForCurrBlock.add(maxCorrIndexes[1]);
            }
            indexesToProcess.remove(maxCorrIndexes[0]);
            indexesToProcess.remove(maxCorrIndexes[1]);
            while (true) {
                Object[] indexOfBitWithMaxCorr = findIndexWithMaxCorrAndSameSigns(indexesOfBitsForCurrBlock, indexesToProcess);
                if ((int) indexOfBitWithMaxCorr[0] == -1) {
                    // start new cluster
                    clusterSizes[clusterId] = indexesOfBitsForCurrBlock.size();
                    totalSize += clusterSizes[clusterId];
                    clusterId++;
                    break;
                } else {
                    indexesOfBitsForCurrBlock.add((int) indexOfBitWithMaxCorr[0]);
                    if ((boolean) indexOfBitWithMaxCorr[1]) {
                        setIndexToSwitch((Integer) indexOfBitWithMaxCorr[0], ret);
                    }
                }
            }
            int start = totalSize - indexesOfBitsForCurrBlock.size();
            for (int i = 0; i < indexesOfBitsForCurrBlock.size(); i++) {
                int idx = indexesOfBitsForCurrBlock.get(i);
                if (idx != -Integer.MAX_VALUE) {
                    newBitsPermutation[start + i] = idx;
                }
            }
        }
        Logger.getLogger(OrientAndShuffleBitsForHWTDynamicLength.class.getName()).log(Level.INFO, "Identified {0} clusters of bit positions having all pairwise correlations positive", clusterId);
        return ret;
    }

    public int[] getClusterSizes() {
        return clusterSizes;
    }

    private Object[] findIndexWithMaxCorrAndSameSigns(List<Integer> indexesOfBitsForCurrBlock, SortedSet<Integer> indexesToProcess) {
        Iterator<Integer> it = indexesToProcess.iterator();
        float maxSumCorrFound = -Float.MAX_VALUE;
        int idxToRet = -1;
        boolean flipBit = false;
        while (it.hasNext()) {
            int cand = it.next();
            float sumForCand = 0;
            boolean searchingPositive = bestFoundCorrelationMatrix[cand][indexesOfBitsForCurrBlock.get(0)] > 0;
            for (int i = 0; i < indexesOfBitsForCurrBlock.size(); i++) {
                float corr = bestFoundCorrelationMatrix[cand][indexesOfBitsForCurrBlock.get(i)];
                if ((searchingPositive && corr < 0) || !searchingPositive && corr > 0) {
                    sumForCand = -Float.MAX_VALUE;
                    break;
                }
                sumForCand += corr;
            }
            if (maxSumCorrFound < Math.abs(sumForCand) && sumForCand != -Float.MAX_VALUE) {
                idxToRet = cand;
                maxSumCorrFound = Math.abs(sumForCand);
                flipBit = sumForCand < 0;
            }
        }
        if (idxToRet != -1) {
            indexesToProcess.remove(idxToRet);
        }
        return new Object[]{idxToRet, flipBit};
    }

    /**
     * ret[i] is position in orig sketch which should be placed to ith in new
     * sketch
     *
     * @return
     */
    @Override
    public int[] getNewBitsPermutation() {
        return newBitsPermutation;
    }

    /**
     * i is considered as the position in old sketch and ret[i] is the position
     * in new sketch which should be placed value in ith old bit
     *
     * @return
     */
    @Override
    public int[] getOldBitsPermutation() {
        int[] ret = new int[newBitsPermutation.length];
        for (int i = 0; i < newBitsPermutation.length; i++) {
            for (int pos = 0; pos < newBitsPermutation.length; pos++) {
                if (newBitsPermutation[pos] == i) {
                    ret[i] = pos;
                    break;
                }
            }
        }
        return ret;
    }

    private int[] findMaxCorr(SortedSet<Integer> indexesToProcess) {
        if (indexesToProcess.size() == 1) {
            return new int[]{indexesToProcess.first(), -Integer.MAX_VALUE};
        }
        float maxCorr = -1;
        int[] ret = new int[2];
        for (int i : indexesToProcess) {
            for (int j : indexesToProcess) {
                if (i != j && Math.abs(bestFoundCorrelationMatrix[i][j]) > maxCorr) {
                    maxCorr = Math.abs(bestFoundCorrelationMatrix[i][j]);
                    if (bestFoundCorrelationMatrix[i][j] < 0) {
                        ret[1] = -j;
                    } else {
                        ret[1] = j;
                    }
                    ret[0] = i;
                }
            }
        }
        return ret;
    }

    private void setIndexToSwitch(int i, BitSet bitset) {
        int abs = Math.abs(i);
        bitset.set(abs);
        bestFoundCorrelationMatrix = changeBitOrientation(abs, bestFoundCorrelationMatrix);
    }

}
