/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package sf.algorithm.myHWT;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;
import java.util.logging.Level;
import java.util.logging.Logger;
import main.tools.Tools;
import messif.algorithms.Algorithm;
import messif.algorithms.AlgorithmMethodException;
import messif.buckets.CapacityFullException;
import messif.objects.util.RankedAbstractObject;
import messif.operations.data.BulkInsertOperation;
import messif.operations.query.KNNQueryOperation;
import sf.objects.impl.Sketch;

/**
 *
 * @author xmic
 */
public class SketchApproxAlgorithm extends Algorithm {

    private final Map<int[], List<Sketch>> sketches = new TreeMap<>(new Tools.IntArraySameLengthsComparator());
    private final int numberOfBlocks;
    private final int sketchLength;
    public static final int THRESHOLD_NO_CHANGE = 2;

    public SketchApproxAlgorithm(String algorithmName, int numberOfBlocks, int sketchLength) throws IllegalArgumentException {
        super(algorithmName);
        this.numberOfBlocks = numberOfBlocks;
        this.sketchLength = sketchLength;
        Logger.getLogger(SketchApproxAlgorithm.class.getName()).log(Level.INFO, "Max number of buckets is approximately {0}", Math.pow(sketchLength / numberOfBlocks + 1, numberOfBlocks));
    }

    public static void main(String[] args) throws CapacityFullException, InstantiationException {
        try {
            int sketchLength = 205;
            Algorithm alg = new SketchApproxAlgorithm("MyAlgorithm", 1, sketchLength);
//            Algorithm alg = new SequentialScan();
            Iterator dataIt = Tools.getIterator(Sketch.class, "c:\\Data\\2016_orientedSketches\\DeCAF_Profiset\\1M_" + sketchLength + "_bits.txt");
            List<Sketch> queries = Tools.getObjects(Sketch.class.getName(), "c:\\Data\\2016_orientedSketches\\DeCAF_Profiset\\Queries-1000_" + sketchLength + "_bits.txt", -1);
            BulkInsertOperation insert = new BulkInsertOperation(dataIt);
            alg.executeOperation(insert);

            int k = 10;
//            int range = 36;
            for (Sketch query : queries) {
                KNNQueryOperation op = new KNNQueryOperation(query, k);
//                RangeQueryOperation op = new RangeQueryOperation(query, range);
                alg.executeOperation(op);
//                Iterator<RankedAbstractObject> answer = op.getAnswer();
//                while (answer.hasNext()) {
//                    RankedAbstractObject next = answer.next();
//                    System.out.println(next.getDistance() + ":" + next.getObject().getLocatorURI());
//                }
            }
            System.out.println("Saved: ");
            for (Float saved : savedDistanceComputations) {
                System.out.println(saved);
            }
            System.out.println("\n" + sumAnswerDists);
        } catch (AlgorithmMethodException | NoSuchMethodException ex) {
            Logger.getLogger(SketchApproxAlgorithm.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    public void insertObjects(BulkInsertOperation op) {
        Iterator<Sketch> insertedObjects = (Iterator<Sketch>) op.getInsertedObjects().iterator();
        while (insertedObjects.hasNext()) {
            Sketch sketch = insertedObjects.next();
            int[] cardinalitiesInBlocks = sketch.getCardinalitiesInBlocks(numberOfBlocks, sketchLength);
            if (!sketches.containsKey(cardinalitiesInBlocks)) {
                List<Sketch> list = new ArrayList<>();
                sketches.put(cardinalitiesInBlocks, list);
            }
            sketches.get(cardinalitiesInBlocks).add(sketch);
        }
        Logger.getLogger(SketchApproxAlgorithm.class.getName()).log(Level.INFO, "Objects inserted. Currently are occupied {0} buckets", sketches.size());
        op.endOperation();
    }

    private static final List<Float> savedDistanceComputations = new ArrayList<>();
    private static int sumAnswerDists = 0;

    public KNNQueryOperation knnSearch(KNNQueryOperation op) {
        int counter = 0;
        Sketch querySketch = (Sketch) op.getQueryObject();
        int[] queryCards = querySketch.getCardinalitiesInBlocks(numberOfBlocks, sketchLength);
        // for sorting variant
        Map<Integer, List<int[]>> lowerBoundedKeys = sortObjectKeys(queryCards);
        int minimalDistanceOfCurList = 0;
        float answerDistance = Sketch.MAX_DISTANCE;
        int lastSumOfAnswers = Integer.MAX_VALUE;
        int noChange = 0;
//        long t2 = 0;
        while (minimalDistanceOfCurList < answerDistance) {
//            List<int[]> keys = getAllKeysInDistance(minimalDistanceOfCurList, queryCards);
//            keys = filterOnlyPresent(keys, sketches.keySet());
            List<int[]> keys = lowerBoundedKeys.get(minimalDistanceOfCurList);
            if (keys != null && !keys.isEmpty()) {
                for (int[] key : keys) {
                    if (minimalDistanceOfCurList < answerDistance && sketches.containsKey(key)) {
                        List<Sketch> sublist = sketches.get(key);
//                        t2 -= System.currentTimeMillis();
                        for (Sketch dataObject : sublist) {
                            op.addToAnswer(dataObject);
                            counter++;
                        }
//                        t2 += System.currentTimeMillis();
                    }
                    if (op.isAnswerFull()) {
                        answerDistance = op.getAnswerDistance();
                    }
                }
                if (op.isAnswerFull()) {
                    Iterator<RankedAbstractObject> answer = op.getAnswer();
                    int sum = 0;
                    while (answer.hasNext()) {
                        sum += (int) answer.next().getDistance();
                    }
                    if (sum == lastSumOfAnswers) {
                        noChange++;
                        if (noChange == THRESHOLD_NO_CHANGE) {
                            break;
                        }
                    } else {
                        noChange = 0;
                    }
                    lastSumOfAnswers = sum;
//                    System.out.println(lastDistInAnswer);
                }
            }
            minimalDistanceOfCurList++;
        }
//        System.out.println("Adding to answer     : " + t2 + "; " + counter + " objects;  saved: " + (100 - counter / 10000f) + " %, kNN dist: " + minimalDistanceOfCurList);
        op.endOperation();
        savedDistanceComputations.add(100 - counter / 10000f);
        sumAnswerDists += lastSumOfAnswers;
//        System.err.println("Pocet: " + counter + ", uspora: " + (100 - counter / 10000f) + " %, posledni vzdalenost: " + minimalDistanceOfCurList);
        return op;
    }
//    public RangeQueryOperation rangeSearch(RangeQueryOperation op) {
//        int counter = 0;
//        int radius = (int) op.getRadius();
//        Sketch querySketch = (Sketch) op.getQueryObject();
//        int queryCard = querySketch.getCardinality();
//        int minimalDistanceOfCurList = 0;
//        while (minimalDistanceOfCurList < radius) {
//            Integer key = queryCard + minimalDistanceOfCurList;
//            if (/*key > addAllUpToLimit && key < addAllFromLimit && */sketches.containsKey(key)) {
//                List<Sketch> sublist = sketches.get(key);
//                for (Sketch dataObject : sublist) {
//                    op.addToAnswer(dataObject, radius);
//                    counter++;
//                }
//            }
//            key = queryCard - minimalDistanceOfCurList;
//            if (minimalDistanceOfCurList != 0 && minimalDistanceOfCurList <= radius && /*key > addAllUpToLimit && key < addAllFromLimit && */ sketches.containsKey(key)) {
//                List<Sketch> sublist = sketches.get(key);
//                for (Sketch dataObject : sublist) {
//                    op.addToAnswer(dataObject, radius);
//                    counter++;
//                }
//            }
//            minimalDistanceOfCurList++;
//        }
//        op.endOperation();
//        savedDistanceComputations.add(100 - counter / 10000f);
//        return op;
//    }
//

    private Map<Integer, List<int[]>> sortObjectKeys(int[] queryCards) {
//        long l = System.currentTimeMillis();
        Map<Integer, List<int[]>> ret = new TreeMap<>();
        Set<int[]> keySet = sketches.keySet();
        for (int[] key : keySet) {
            int sum = 0;
            for (int i = 0; i < key.length; i++) {
                sum += Math.abs(key[i] - queryCards[i]);
            }
            if (!ret.containsKey(sum)) {
                ret.put(sum, new ArrayList<>());
            }
            ret.get(sum).add(key);
        }
//        System.out.println("Sorting grid to query: " + (System.currentTimeMillis() - l));
        return ret;
    }

    @Deprecated
    private List<int[]> getKeysInDistanceBySeqScan(int distOfList, int[] queryCards) {
        List ret = new ArrayList<>();
        Set<int[]> keySet = sketches.keySet();
        for (int[] key : keySet) {
            int sum = 0;
            for (int i = 0; i < key.length; i++) {
                sum += Math.abs(key[i] - queryCards[i]);
            }
            if (sum == distOfList) {
                ret.add(key);
            }
        }
        return ret;
    }

//    //Chyba!
//    @Deprecated
//    private List<int[]> getAllKeysInDistance(int dist, int[] queryCards) {
//        if(dist >= 4){
//            String s = "";;
//        }
//        List<int[]> ret = new ArrayList<>();
//        int[] startVector = new int[queryCards.length];
//        startVector[0] = queryCards[0] - dist;
//        System.arraycopy(queryCards, 1, startVector, 1, startVector.length - 1);
//        Collection<int[]> batch = new HashSet<>();
//        ret.add(startVector);
//        batch.add(startVector);
//        Tools.printCollection(Collections.singletonList(startVector));;
//        System.out.println("");;
//        int curZeroAxis = startVector[0];
//        while (true) {
//            batch = iterate(curZeroAxis, batch, queryCards[0], 1, queryCards);
//            if (batch.isEmpty()) {
//                break;
//            }
//            Tools.printCollection(batch);;
//            System.out.println("");;
//            ret.addAll(batch);
//            curZeroAxis++;
//        }
//        if (dist == 0) {
//            return ret;
//        }
//        startVector = new int[queryCards.length];
//        startVector[0] = queryCards[0] + dist;
//        System.arraycopy(queryCards, 1, startVector, 1, startVector.length - 1);
//        ret.add(startVector);
//        batch.clear();
//        batch.add(startVector);
//        Tools.printCollection(Collections.singletonList(startVector));
//        System.out.println("");;
//        curZeroAxis = startVector[0];
//        while (true) {
//            batch = iterate(curZeroAxis, batch, queryCards[0], -1, queryCards);
//            if (batch.isEmpty()) {
//                break;
//            }
//            Tools.printCollection(batch);;
//            System.out.println("");;
//            ret.addAll(batch);
//            curZeroAxis--;
//        }
////        System.out.println(dist + ";" + ret.size());;
//        return ret;
//    }
//
//    private Collection<int[]> iterate(int curZeroAxis, Collection<int[]> batch, int limitForZeroAxis, int incr, int[] queryCards) {
//        Collection<int[]> ret = new ArrayList<>();
//        curZeroAxis += incr;
//        if ((curZeroAxis > limitForZeroAxis && incr == 1) || (curZeroAxis <= limitForZeroAxis && incr == -1)) {
//            return ret;
//        }
//        for (int[] point : batch) {
//            for (int i = 1; i < point.length; i++) {
//                if (point[i] >= queryCards[i]) {
//                    int[] add1 = getUpdatedPoint(point, curZeroAxis, i, 1);
//                    ret.add(add1);
//                }
//                if (point[i] <= queryCards[i]) {
//                    int[] add2 = getUpdatedPoint(point, curZeroAxis, i, -1);
//                    ret.add(add2);
//                }
//            }
//        }
//        return ret;
//    }
//
//    private int[] getUpdatedPoint(int[] point, int curZeroAxis, int idx, int inc) {
//        int[] ret = new int[point.length];
//        System.arraycopy(point, 0, ret, 0, point.length);
//        ret[0] = curZeroAxis;
//        ret[idx] += inc;
//        return ret;
//    }
//
//    private List<int[]> filterOnlyPresent(List<int[]> all, Set<int[]> present) {
//        List<int[]> ret = new ArrayList<>();
//        for (int[] tuple : all) {
//            if (present.contains(tuple)) {
//                ret.add(tuple);
//            }
//        }
//        return ret;
//    }

}
