package sf.deprecated.vykryvaniProstoru;

import java.io.FileOutputStream;
import java.io.IOException;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;
import messif.objects.LocalAbstractObject;
import sf.objects.impl.PivotsDistanceList;
import sf.objects.impl.Sketch;

/**
 *
 * @author xmic
 */
public class Evaluator {

    private final List<LocalAbstractObject> queries;
    private final PivotsDistanceList pivotDistanceList;
    private final Map<LocalAbstractObject, Sketch> sketches;
    private final String outputFileName;

    public Evaluator(List<LocalAbstractObject> queries, PivotsDistanceList pivotDistanceList, Map<LocalAbstractObject, Sketch> sketches, String outputFileName) {
        this.queries = queries;
        this.pivotDistanceList = pivotDistanceList;
        this.sketches = sketches;
        this.outputFileName = outputFileName;
    }

    public void evaluateExperiment(Experiment exp) {
        FileOutputStream os = null;
        try {
            os = new FileOutputStream(outputFileName);
            for (LocalAbstractObject query : queries) {
                Iterator<ResultHolder> it = exp.evaluateQuery(query, sketches, pivotDistanceList).iterator();
                while (it.hasNext()) {
                    ResultHolder rs = it.next();
                    rs.write(os);
                }
                os.write('\n');
                Logger.getLogger(Evaluator.class.getName()).log(Level.INFO, "Evaluated query.");
            }
        } catch (IOException ex) {
            Logger.getLogger(Evaluator.class.getName()).log(Level.SEVERE, null, ex);
        } finally {
            try {
                os.flush();
                os.close();
            } catch (IOException ex) {
                Logger.getLogger(Evaluator.class.getName()).log(Level.SEVERE, null, ex);
            }
        }
    }
}
