package sf.hyperplaneDistanceDistribution.impl;

import java.io.IOException;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.TreeMap;
import java.util.logging.Level;
import java.util.logging.Logger;
import messif.objects.LocalAbstractObject;
import sf.hyperplaneDistanceDistribution.DistanceDistibutionToDividingLineToCsv;
import sf.hyperplaneDistanceDistribution.impl.DistanceDistributionToFarPivotsCouples.FloatDescendingComparator;
import static sf.hyperplaneDistanceDistribution.impl.DistanceDistributionToRandomPivotsCouplesWithGivenPartSizes.dividesToPartOfSize;

/**
 *
 * @author xmic
 */
public class DistanceDistributionToRandomPivotsCouplesWithGivenPartSizesBestValues extends DistanceDistibutionToDividingLineToCsv {

    public static final Float SIZE_OF_SMALLER_PART = 0.45f;
    public static final Integer NUMBER_OF_CAND_PIVOT_PAIRS = 512;

    public DistanceDistributionToRandomPivotsCouplesWithGivenPartSizesBestValues(List<LocalAbstractObject> pivots, List<LocalAbstractObject> objects) {
        super(pivots, objects);
    }

    @Override
    public void writeDistanceDistribution(OutputStream os) throws IOException {
        List<LocalAbstractObject[]> pivotPairs = getWellDividingPivotsCandidates(NUMBER_OF_CAND_PIVOT_PAIRS);
        float[] distsToHypplaneSum = new float[pivotPairs.size()];
        Arrays.fill(distsToHypplaneSum, 0f);
        for (int j = 0; j < objects.size(); j++) {
            LocalAbstractObject obj = objects.get(j);
            for (int i = 0; i < pivotPairs.size(); i++) {
                LocalAbstractObject piv0 = pivotPairs.get(i)[0];
                LocalAbstractObject piv1 = pivotPairs.get(i)[1];
                float d0 = obj.getDistance(piv0);
                float d1 = obj.getDistance(piv1);
                float pivDist = piv0.getDistance(piv1);
                float distToHyperplane = (d0 * d0 - d1 * d1) / (2 * pivDist);
                distToHyperplane = Math.abs(distToHyperplane);
                distsToHypplaneSum[i] += distToHyperplane;
            }
        }
        int resultSize = pivots.size() / 2;
        TreeMap<Float, LocalAbstractObject[]> pivotPairsMap = new TreeMap(new FloatDescendingComparator());
        for (int i = 0; i < pivotPairs.size(); i++) {
            pivotPairsMap.put(distsToHypplaneSum[i], pivotPairs.get(i));
        }
        while (pivotPairsMap.size() > resultSize) {
            pivotPairsMap.remove(pivotPairsMap.lastKey());
        }
        pivotPairs = new ArrayList(pivotPairsMap.values());
        super.writeDistanceDistribution(os, pivotPairs);
    }

    public List<LocalAbstractObject[]> getWellDividingPivotsCandidates(int pairsNumber) {
        List<LocalAbstractObject[]> ret = new ArrayList();
        for (int i = 0; i < pivots.size() - 1; i++) {
            for (int j = i + 1; j < pivots.size(); j++) {
                if (ret.size() == pairsNumber) {
                    return ret;
                }
                LocalAbstractObject[] pair = new LocalAbstractObject[2];
                pair[0] = pivots.get(i);
                pair[1] = pivots.get(j);
                if (dividesToPartOfSize(SIZE_OF_SMALLER_PART, pair, objects)) {
                    ret.add(pair);
                }
            }
        }
        if (ret.size() != pairsNumber) {
            Logger.getLogger(DistanceDistributionToRandomPivotsCouplesWithGivenPartSizesBestValues.class.getName()).log(Level.WARNING, "Not possible to find enought pivot pair meeting condition.");
        }
        return ret;
    }

}
