package sf.objects.impl;

import java.io.IOException;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;
import static main.tools.oldNotUsed.DistanceDistributionToHyperplanesExperiment.floatToByteArrayForPrint;
import messif.objects.LocalAbstractObject;

/**
 *
 * @author xmic
 */
public class PivotsDistanceList {

    private final List<LocalAbstractObject> pivots;
    private final float[][] distances;
    private final float[] ranges;

    public PivotsDistanceList(List<LocalAbstractObject> pivots, float range) {
        this.pivots = pivots;
        int size = pivots.size();
        this.ranges = new float[size];
        Arrays.fill(ranges, range);
        distances = new float[size][size];
        for (int i = 0; i < size; i++) {
            distances[i][i] = 0f;
        }
        for (int i = 0; i < size - 1; i++) {
            for (int j = i + 1; j < size; j++) {
                float dist = pivots.get(i).getDistance(pivots.get(j));
                distances[i][j] = dist;
                distances[j][i] = dist;
            }
        }
    }

    public PivotsDistanceList(List<LocalAbstractObject> pivots, float[] ranges) {
        this.pivots = pivots;
        int size = pivots.size();
        this.ranges = ranges;
        distances = new float[size][size];
        for (int i = 0; i < size; i++) {
            distances[i][i] = 0f;
        }
        for (int i = 0; i < size - 1; i++) {
            for (int j = i + 1; j < size; j++) {
                float dist = pivots.get(i).getDistance(pivots.get(j));
                distances[i][j] = dist;
                distances[j][i] = dist;
            }
        }
    }

    public PivotsDistanceList(LocalAbstractObject[] pivots, float[] ranges) {
        this(Arrays.asList(pivots), ranges);
    }

    public float[] getRanges() {
        return ranges;
    }

    public float getDistance(int i1, int i2) {
        return distances[i1][i2];
    }

    public LocalAbstractObject getPivot(int index) {
        return pivots.get(index);
    }

    public List<LocalAbstractObject> getPivots() {
        return Collections.unmodifiableList(pivots);
    }

    public void writeDownDistances(OutputStream os) {
        List<Float> dists = new ArrayList();
        for (int i = 0; i < pivots.size() - 1; i++) {
            for (int j = i + 1; j < pivots.size(); j++) {
                dists.add(getDistance(i, j));
            }
        }
        Collections.sort(dists);
        try {
            for (int i = 0; i < dists.size(); i++) {
                byte[] bytes = floatToByteArrayForPrint(dists.get(i));
                os.write(bytes);
                os.write('\n');
            }
        } catch (IOException ex) {
            Logger.getLogger(PivotsDistanceList.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    public void writeDownPivots(OutputStream os) {
        try {
            for (int i = 0; i < pivots.size(); i++) {
                pivots.get(i).write(os);
            }
        } catch (IOException ex) {
            Logger.getLogger(PivotsDistanceList.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

}
