package sf.objects.sketchcreator.impl;

import java.util.List;
import messif.objects.LocalAbstractObject;
import sf.objects.impl.Sketch;
import sf.objects.sketchcreator.ObjectToSketchTransformator;
import sf.pivotAnalyzer.CsvPivotInfoFileProcessor;
import sf.pivotAnalyzer.csvPivotInfoFileProcessorImpl.CsvFileBallRanges;

/**
 *
 * @author xmic
 */
public class BallPartitioning extends ObjectToSketchTransformator {

    private float[] ranges;

    public BallPartitioning(String pivotFile, String objectClass, int pivotCount, String csvFilePivotRadiusInfo, boolean uniquePivots) {
        super(pivotFile, objectClass, -1);
        ranges = new float[pivotCount];
        CsvPivotInfoFileProcessor pivotCreator = new CsvFileBallRanges(csvFilePivotRadiusInfo, pivots, ranges);
        pivots = pivotCreator.getPivotsAccordingCsvLocators(pivotCount);
    }

    public BallPartitioning(LocalAbstractObject[] pivots, float[] ranges) {
        super(pivots);
        this.ranges = ranges;
    }

    @Override
    public Sketch transform(LocalAbstractObject objectToTransform) {
        if (objectToTransform == null) {
            return null;
        }
        Sketch ret = new Sketch(objectToTransform.getLocatorURI(), pivots.length);
        for (int i = 0; i < pivots.length; i++) {
            float dist = objectToTransform.getDistance(pivots[i]);
            if (dist <= ranges[i]) {
                ret.setBitValuePositive(i);
            }
        }
        return ret;
    }

    @Override
    public void transformToInvertedFile(LocalAbstractObject sampleObj, List<Sketch> invertedSketches, int objectNumber) {
        for (int i = 0; i < pivots.length; i++) {
            float dist = sampleObj.getDistance(pivots[i]);
            if (dist <= ranges[i]) {
                invertedSketches.get(i).setBitValuePositive(objectNumber);
            }
        }
    }

    @Override
    public void removeCorrelatedBits(int[] uncorrelatedBits) {
        LocalAbstractObject[] newPivots = new LocalAbstractObject[uncorrelatedBits.length];
        float[] newRanges = new float[uncorrelatedBits.length];
        for (int i = 0; i < uncorrelatedBits.length; i++) {
            int sketchIndex = uncorrelatedBits[i];
            newPivots[i] = pivots[sketchIndex];
            newRanges[i] = ranges[sketchIndex];
        }
        pivots = newPivots;
        ranges = newRanges;
        for (int i = 0; i < pivots.length; i++) {
            System.out.println(pivots[i].getLocatorURI() + ";" + ranges[i]);
        }
    }

    @Override
    public int getInvertedSketchesCount() {
        return pivots.length;
    }

}
