package sf.objects.sketchcreator.impl;

import java.io.IOException;
import java.io.Serializable;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;
import main.tools.Tools;
import messif.objects.LocalAbstractObject;
import sf.objects.impl.Sketch;
import sf.objects.sketchcreator.ObjectToSketchTransformator;
import sf.pivotAnalyzer.CsvPivotInfoFileProcessor;
import sf.pivotAnalyzer.PivotSubsetSelector;
import sf.pivotAnalyzer.csvPivotInfoFileProcessorImpl.CsvFileHyperplanePairs;

/**
 *
 * @author xmic
 */
public class GenHyperplane extends ObjectToSketchTransformator implements Serializable {

    @SuppressWarnings("FieldNameHidesFieldInSuperclass")
    public static final long serialVersionUID = 736012244806410477l;

    public GenHyperplane(String pivotFile, String objectClass, int sketchSize, String csvFileWithPivotsPairs, boolean uniquePivots) {
        super(pivotFile, objectClass, 2560);
        if (csvFileWithPivotsPairs != null) {
            CsvPivotInfoFileProcessor pivotCreator = new CsvFileHyperplanePairs(csvFileWithPivotsPairs, uniquePivots, pivots);
            pivots = pivotCreator.getPivotsAccordingCsvLocators(2 * sketchSize);
        }
    }

    public GenHyperplane(String pivotFile, String objectClass, int sketchSizeOrHalfOfInvestigatedPivots, boolean makeAllPairs) {
        super(pivotFile, objectClass, 2 * sketchSizeOrHalfOfInvestigatedPivots);
        if (makeAllPairs) {
            int length = pivots.length;
            LocalAbstractObject[] newPivots = new LocalAbstractObject[length * (length - 1)];
            int counter = 0;
            for (int i = 0; i < length - 1; i++) {
                for (int j = i + 1; j < length; j++) {
                    newPivots[counter] = pivots[i];
                    counter++;
                    newPivots[counter] = pivots[j];
                    counter++;
                }
            }
            pivots = newPivots;
        }
    }

    public GenHyperplane(String pivotFile, String objectClass, boolean makeAllPairs) {
        this(pivotFile, objectClass, -1, makeAllPairs);
    }

    public GenHyperplane(LocalAbstractObject[] pivots) {
        super(pivots);
    }

    private void readObject(java.io.ObjectInputStream in) throws IOException, ClassNotFoundException {
        in.defaultReadObject();
    }

    @Override
    public Sketch transform(LocalAbstractObject objectToTransform) {
        if (objectToTransform == null) {
            return null;
        }
        Sketch ret = new Sketch(objectToTransform.getLocatorURI(), pivots.length / 2);
        for (int i = 0; i < pivots.length - 1; i = i + 2) {
            float d0 = objectToTransform.getDistance(pivots[i]);
            float d1 = objectToTransform.getDistance(pivots[i + 1]);
            if (d0 < d1) {
                ret.setBitValuePositive(i / 2);
            }
        }
        return ret;
    }

    @Override
    public void transformToInvertedFile(LocalAbstractObject sampleObj, List<Sketch> invertedSketches, int objectNumber) {
        for (int i = 0; i < pivots.length - 1; i = i + 2) {
            float d0 = sampleObj.getDistance(pivots[i]);
            float d1 = sampleObj.getDistance(pivots[i + 1]);
            if (d0 < d1) {
                Sketch s = invertedSketches.get(i / 2);
                synchronized (s) {
                    s.setBitValuePositive(objectNumber);
                }
            }
        }
    }

    @Override
    public void removeCorrelatedBits(int[] uncorrelatedBits) {
        LocalAbstractObject[] newPivots = new LocalAbstractObject[2 * uncorrelatedBits.length];
        for (int i = 0; i < uncorrelatedBits.length; i++) {
            int sketchIndex = uncorrelatedBits[i];
            newPivots[2 * i] = pivots[2 * sketchIndex];
            newPivots[2 * i + 1] = pivots[2 * sketchIndex + 1];
        }
        pivots = newPivots;
        Tools.printAsPairs(pivots, System.out);
    }

    @Override
    public int getInvertedSketchesCount() {
        return pivots.length / 2;
    }

    public void reorderPivotsByOutlier(LocalAbstractObject outlier) {
        for (int i = 0; i < pivots.length; i = i + 2) {
            float d0 = pivots[i].getDistance(outlier);
            float d1 = pivots[i + 1].getDistance(outlier);
            if (d0 < d1) {
                LocalAbstractObject switched = pivots[i];
                pivots[i] = pivots[i + 1];
                pivots[i + 1] = switched;
            }
        }
        Sketch verify = transform(outlier);
        try {
            System.out.println("If tranformed outlier to sketch should be zero: ");
            verify.write(System.out);
        } catch (IOException ex) {
            Logger.getLogger(GenHyperplane.class.getName()).log(Level.SEVERE, null, ex);
        }

    }

    public void reorderPivotsToSwitchBits(int... switchBits) {
        if (switchBits == null) {
            return;
        }
        for (int switchBit : switchBits) {
            int i = 2 * switchBit;
            LocalAbstractObject switched = pivots[i];
            pivots[i] = pivots[i + 1];
            pivots[i + 1] = switched;
        }
    }
}
