/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package sf.objects.sketchcreator.impl;

import java.io.IOException;
import java.io.Serializable;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import messif.objects.LocalAbstractObject;
import sf.objects.impl.Sketch;
import sf.objects.sketchcreator.ObjectToSketchTransformator;
import sf.pivotAnalyzer.csvPivotInfoFileProcessorImpl.CsvFileVoronoiPivots;

/**
 *
 * @author xmic
 */
public class VoronoiSketchTransformator extends ObjectToSketchTransformator implements Serializable {

    @SuppressWarnings("FieldNameHidesFieldInSuperclass")
    public static final long serialVersionUID = 736012244806410499l;

    private Map<String, LocalAbstractObject> pivotsMapping;
    private int sketchLength;

    public VoronoiSketchTransformator(String pivotFile, String objectClass, int sketchSize, String csvFileWithPivotsPairs, boolean uniquePivots) {
        super(pivotFile, objectClass, -1);
        if (csvFileWithPivotsPairs != null) {
            CsvFileVoronoiPivots csvFile = new CsvFileVoronoiPivots(csvFileWithPivotsPairs, uniquePivots, pivots);
            pivotsMapping = csvFile.getPivotsMapping();
            sketchLength = getSketchLengthRemoveLong();
//            consilidateAsHyperplanesToZero("", pivotsMapping);
//            try {
//                System.setOut(new PrintStream("c:\\Data\\2017\\2017_Voronoi_Sketches\\Pivots_100000obj_512p.txt"));
//            } catch (FileNotFoundException ex) {
//                Logger.getLogger(VoronoiSketchTransformator.class.getName()).log(Level.SEVERE, null, ex);
//            }
//            PivotsForVoronoiSketchesSelector.printPivots(pivotsMapping);
        }
    }

    public VoronoiSketchTransformator(LocalAbstractObject[] pivots, String csvFileWithPivotsPairs) {
        super(pivots);
        if (csvFileWithPivotsPairs != null) {
            CsvFileVoronoiPivots csvFile = new CsvFileVoronoiPivots(csvFileWithPivotsPairs, false, pivots);
            pivotsMapping = csvFile.getPivotsMapping();
            sketchLength = getSketchLengthRemoveLong();
        }
    }

    private void readObject(java.io.ObjectInputStream in) throws IOException, ClassNotFoundException {
        in.defaultReadObject();
    }

    @Override
    public Sketch transform(LocalAbstractObject objectToTransform) {
        if (objectToTransform == null) {
            return null;
        }
        Sketch ret = new Sketch(objectToTransform.getLocatorURI(), sketchLength);
        String key = "";
        while (key.length() != sketchLength) {
            LocalAbstractObject p0 = pivotsMapping.get(key + "0");
            LocalAbstractObject p1 = pivotsMapping.get(key + "1");
            float d0 = objectToTransform.getDistance(p0);
            float d1 = objectToTransform.getDistance(p1);
            if (d0 < d1) {
                key += "0";
            } else {
                ret.setBitValuePositive(key.length());
                key += "1";
            }
        }
        return ret;
    }

    protected String transformToWord(LocalAbstractObject objectToTransform) {
        if (objectToTransform == null) {
            return null;
        }
        String ret = "";
        while (ret.length() != sketchLength) {
            LocalAbstractObject p0 = pivotsMapping.get(ret + "0");
            LocalAbstractObject p1 = pivotsMapping.get(ret + "1");
            float d0 = objectToTransform.getDistance(p0);
            float d1 = objectToTransform.getDistance(p1);
            if (d0 < d1) {
                ret += "0";
            } else {
                ret += "1";
            }
        }
        return ret;
    }

    @Override
    public void transformToInvertedFile(LocalAbstractObject sampleObj, List<Sketch> invertedSketches, int objectNumber) {
        throw new UnsupportedOperationException("Not supported yet."); //To change body of generated methods, choose Tools | Templates.
    }

    @Override
    public void removeCorrelatedBits(int[] uncorrelatedBits) {
        throw new UnsupportedOperationException("Not supported yet."); //To change body of generated methods, choose Tools | Templates.
    }

    @Override
    public int getInvertedSketchesCount() {
        return sketchLength;
    }

    public void reorderPivotsByOutlier(LocalAbstractObject outlier) {
        throw new UnsupportedOperationException("Not supported yet."); //To change body of generated methods, choose Tools | Templates.
    }

    public void reorderPivotsToSwitchBits(int[] switchBits) {
        throw new UnsupportedOperationException("Not supported yet."); //To change body of generated methods, choose Tools | Templates.
    }

    private int getSketchLengthRemoveLong() {
        Map<Integer, Integer> lengths = new HashMap<>();
        for (Map.Entry<String, LocalAbstractObject> pivot : pivotsMapping.entrySet()) {
            int key = pivot.getKey().length();
            if (!lengths.containsKey(key)) {
                lengths.put(key, 0);
            }
            lengths.put(key, lengths.get(key) + 1);
        }
        int i;
        for (i = 1; i < lengths.size(); i++) {
            Integer numberOfPivots = lengths.get(i);
            if (numberOfPivots != Math.pow(2, i)) {
                break;
            }
        }
        int ret = i - 1;
        Iterator<String> it = new HashSet<>(pivotsMapping.keySet()).iterator();
        while (it.hasNext()) {
            String key = it.next();
            if (key.length() > ret) {
                pivotsMapping.remove(key);
            }
        }
        return ret;
    }

//    // shift pivots to better preserve distances
//    private void consilidateAsHyperplanesToZero(String base, Map<String, LocalAbstractObject> pivotsMapping) {
//        LocalAbstractObject p0 = pivotsMapping.get(base + "0");
//        LocalAbstractObject p1 = pivotsMapping.get(base + "1");
//        if (p0 == null || p1 == null) {
//            return;
//        }
//        consilidateAsHyperplanesToZeroBit(base + "0", p0, p1);
//        consilidateAsHyperplanesToZeroBit(base + "1", p0, p1);
//        consilidateAsHyperplanesToZero(base + "0", pivotsMapping);
//        consilidateAsHyperplanesToZero(base + "1", pivotsMapping);
//    }
//
//    private void consilidateAsHyperplanesToZeroBit(String base, LocalAbstractObject p0, LocalAbstractObject p1) {
//        LocalAbstractObject a = pivotsMapping.get(base + "0");
//        LocalAbstractObject b = pivotsMapping.get(base + "1");
//        if (a == null || b == null) {
//            return;
//        }
//        //05
//        float hypB = b.getDistance(p0) + b.getDistance(p1);
//        float hypA = a.getDistance(p0) + a.getDistance(p1);
//        //01
//        // pivot ten, aby to byl rodic - tady nejde poznat
////        float hypA = a.getDistance(p0) - b.getDistance(p0);
////        if (hypA < 0) {
////            pivotsMapping = switchBits(base + "1", pivotsMapping);
////        }
//        //03
//        //        float hypB = Math.abs(b.getDistance(p0) - b.getDistance(p1));
//        //        float hypA = Math.abs(a.getDistance(p0) - a.getDistance(p1));
//        //04
//        //        float hypB = b.getDistance(p0) - b.getDistance(p1);
//        //        float hypA = a.getDistance(p0) - a.getDistance(p1);
//        System.err.println("\n" + hypB + ", " + hypA);
//        if (hypB < hypA) {
//            pivotsMapping = switchBits(base, pivotsMapping);
//        }
//
//    }
//
//    private Map<String, LocalAbstractObject> switchBits(String base, Map<String, LocalAbstractObject> pivotsMapping) {
//        Logger.getLogger(VoronoiSketchTransformator.class.getName()).log(Level.INFO, "Switching pivots {0}0 and {0}1", new Object[]{base});
//        Map<String, LocalAbstractObject> ret = new HashMap<>();
//        Iterator<String> it = pivotsMapping.keySet().iterator();
//        while (it.hasNext()) {
//            String key = it.next();
//            if (!key.startsWith(base) || key.length() <= base.length()) {
//                ret.put(key, pivotsMapping.get(key));
//                continue;
//            }
//            // prohodit nasledujici cislici v klici
//            int pos = base.length();
//            String newKey = key;
//            char newChar;
//            if (newKey.charAt(pos) == '0') {
//                newChar = '1';
//            } else {
//                newChar = '0';
//            }
//            newKey = newKey.substring(0, pos) + newChar + newKey.substring(pos + 1);
//            LocalAbstractObject value = pivotsMapping.get(key);
//            ret.put(newKey, value);
//        }
//        return ret;
//    }

}
