/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package sf.pivotAnalyzer;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;
import messif.objects.LocalAbstractObject;

/**
 *
 * @author xmic
 */
public class PivotsForVoronoiSketchesSelector {

    private final float balanceToFind = 0.5f;
    private Map<String[], Float> evaluatedDistances;

    private final LocalAbstractObject[] objects;
    private final LocalAbstractObject[] pivots;

    private final Random r = new Random();

    public PivotsForVoronoiSketchesSelector(List<LocalAbstractObject> objects, List<LocalAbstractObject> setOfPivots) {
        this(objects.toArray(new LocalAbstractObject[objects.size()]),
                setOfPivots.toArray(new LocalAbstractObject[setOfPivots.size()]));
    }

    public PivotsForVoronoiSketchesSelector(LocalAbstractObject[] objects, LocalAbstractObject[] setOfPivots) {
        this.evaluatedDistances = new HashMap();
        this.objects = objects;
        this.pivots = setOfPivots;
    }

    /**
     *
     * @return Map of pivots: "bit prefix",pivot
     */
    public Map<String, LocalAbstractObject> analysePivots() {
        Map<String, LocalAbstractObject> ret = new HashMap<>();
        Map<LocalAbstractObject, LocalAbstractObject[]> pivotsWithObjects = new HashMap<>();
        splitObjects(objects, pivotsWithObjects, ret, "");
        return ret;
    }

    private void splitObjects(LocalAbstractObject[] objects, Map<LocalAbstractObject, LocalAbstractObject[]> pivotsWithObjects, Map<String, LocalAbstractObject> pivotsWithPrefix, String curPreffix) {
        if (objects.length <= 1) {
            return;
        }
        LocalAbstractObject pivot1Best = null;
        LocalAbstractObject pivot2Best = null;
        float bestDev = 1;
        boolean isBreaked = false;
        Set<Integer> checkedI = new HashSet<>();
        while (checkedI.size() != pivots.length - 1) {
            int i = selectRandomIndex(checkedI, 0, pivots.length - 1);
            if (bestDev < 0.002 || bestDev * objects.length < 1) {
                break;
            }
            System.err.println("\nPicked new i: " + i);
            LocalAbstractObject pivot1 = pivots[i];
            Set<Integer> checkedJ = new HashSet<>();
            while (checkedJ.size() < pivots.length - i - 1) {
                int j = selectRandomIndex(checkedJ, i + 1, pivots.length);
//                System.out.println("j:" + j + ", i: " + i + ", banned size:" + checkedJ.size() + ", limit: " + (pivots.length - i - 1));
                LocalAbstractObject pivot2 = pivots[j];
                float splitRatio = evaluateRatioOfGHP(objects, pivot1, pivot2);
                float found = Math.abs(splitRatio - balanceToFind);
                if (found < bestDev) {
                    Logger.getLogger(PivotsForVoronoiSketchesSelector.class.getName())
                            .log(Level.INFO, "Found balance {0}, pivots: {1}, i: {2}, j: {3}", new Object[]{Float.toString(found), pivots.length, i, j});
                    bestDev = found;
                    pivot1Best = pivot1;
                    pivot2Best = pivot2;
                    if (bestDev < 0.002 || bestDev * objects.length < 1f) {
                        isBreaked = true;
                        break;
                    }
                }
            }
        }

        if (!isBreaked) {
            Logger.getLogger(PivotsForVoronoiSketchesSelector.class.getName()).log(Level.WARNING, "Proper balancing not found.");
        }
//        // varianta 01
        LocalAbstractObject prevPivot = pivotsWithPrefix.get(curPreffix);
        if (prevPivot != null) {
            float d1 = prevPivot.getDistance(pivot1Best);
            float d2 = prevPivot.getDistance(pivot2Best);
            if (d2 < d1) {
                LocalAbstractObject tmp = pivot1Best;
                pivot1Best = pivot2Best;
                pivot2Best = tmp;
            }
        }
        List<LocalAbstractObject>[] splittedObjects = ghp(objects, pivot1Best, pivot2Best);
        LocalAbstractObject[] toArray1 = splittedObjects[0].toArray(new LocalAbstractObject[0]);
        LocalAbstractObject[] toArray2 = splittedObjects[1].toArray(new LocalAbstractObject[0]);
        Logger.getLogger(PivotsForVoronoiSketchesSelector.class.getName()).log(Level.INFO, "Found pivots: {0}0, {0}1. Lists sizes: {1} and {2}", new Object[]{curPreffix, toArray1.length, toArray2.length});
        addResultsAndCont(objects.length, toArray1, pivot1Best, pivotsWithObjects, pivotsWithPrefix, curPreffix + "0");
        addResultsAndCont(objects.length, toArray2, pivot2Best, pivotsWithObjects, pivotsWithPrefix, curPreffix + "1");
        Logger.getLogger(PivotsForVoronoiSketchesSelector.class.getName()).log(Level.INFO, "Finished splitting lists of sizes {1} and {2}, preffix: {0}", new Object[]{curPreffix, toArray1.length, toArray2.length});
    }

    private float evaluateRatioOfGHP(LocalAbstractObject[] objects, LocalAbstractObject pivot1, LocalAbstractObject pivot2) {
        int sum = 0;
        for (LocalAbstractObject object : objects) {
            float d1 = getDistance(object, pivot1);
            float d2 = getDistance(object, pivot2);
            if (d1 < d2) {
                sum++;
            }
        }
        return (float) sum / objects.length;
    }

    private List<LocalAbstractObject>[] ghp(LocalAbstractObject[] objects, LocalAbstractObject pivot1, LocalAbstractObject pivot2) {
        List<LocalAbstractObject> p1 = new ArrayList<>();
        List<LocalAbstractObject> p2 = new ArrayList<>();
        for (LocalAbstractObject object : objects) {
            float d1 = getDistance(object, pivot1);
            float d2 = getDistance(object, pivot2);
            if (d1 < d2) {
                p1.add(object);
            } else {
                p2.add(object);
            }
        }
        return new List[]{p1, p2};
    }

    private float getDistance(LocalAbstractObject o1, LocalAbstractObject o2) {
        String[] key = new String[]{o1.getLocatorURI(), o2.getLocatorURI()};
        if (evaluatedDistances.containsKey(key)) {
            return evaluatedDistances.get(key);
        }
        float dist = o1.getDistance(o2);
        if (evaluatedDistances.size() < 100000 * 400) {
            evaluatedDistances.put(key, dist);
        }
        return dist;
    }

    public static void printPivots(Map<String, LocalAbstractObject> map) {
        Iterator<String> it = map.keySet().iterator();
        while (it.hasNext()) {
            String key = it.next();
            LocalAbstractObject pivot = map.get(key);
            System.out.println(key + ";" + pivot.getLocatorURI());
        }
    }

    private void addResultsAndCont(int lengthOrig, LocalAbstractObject[] foundArray, LocalAbstractObject pivot, Map<LocalAbstractObject, LocalAbstractObject[]> pivotsWithObjects, Map<String, LocalAbstractObject> pivotsWithPrefix, String curKey) {
        if (foundArray.length < lengthOrig) {
            pivotsWithObjects.put(pivot, foundArray);
            pivotsWithPrefix.put(curKey, pivot);
            splitObjects(foundArray, pivotsWithObjects, pivotsWithPrefix, curKey);
        } else {
            Logger.getLogger(PivotsForVoronoiSketchesSelector.class.getName()).log(Level.WARNING, "Not possible to split array of length {0}", lengthOrig);
        }
    }

    private int selectRandomIndex(Set<Integer> banned, int lower, int upper) {
        int dif = upper - lower;
        while (true) {
            Integer ret = lower + r.nextInt(dif);
            if (!banned.contains(ret)) {
                banned.add(ret);
                return ret;
            }
        }
    }
}
