package sf.pivotAnalyzer.threads;

import java.util.Comparator;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;
import messif.objects.LocalAbstractObject;

/**
 *
 * @author xmic
 */
public abstract class PivotCoupleAnalyzer extends Thread {

    protected final LocalAbstractObject[] pivots;
    protected final List<LocalAbstractObject> sampleObjects;
    protected final Map<int[], Float> result = new TreeMap<>(new IntArraySameLengthComparator());

    public PivotCoupleAnalyzer(LocalAbstractObject[] pivots, List<LocalAbstractObject> sampleObjects) {
        this.pivots = pivots;
        this.sampleObjects = sampleObjects;
    }

    public Map<int[], Float> getResult() {        
        return result;
    }

    public float getPivotDistance(int pivotIdx0, int pivotIdx1) {
        if (pivotIdx0 == pivotIdx1) {
            return 0f;
        }
        if (pivotIdx0 > pivots.length || pivotIdx1 > pivots.length) {
            return -1f;
        }
        return pivots[pivotIdx0].getDistance(pivots[pivotIdx1]);
    }

    public static class IntArraySameLengthComparator implements Comparator<int[]> {

        public IntArraySameLengthComparator() {
        }

        @Override
        public int compare(int[] o1, int[] o2) {
            for (int i = 0; i < o1.length; i++) {
                int ret = Integer.compare(o1[i], o2[i]);
                if (ret != 0) {
                    return ret;
                }
            }
            return 0;
        }
    }

    protected float[] initDistances(LocalAbstractObject sampleObject) {
        float[] ret = new float[pivots.length];
        for (int i = 0; i < pivots.length; i++) {
            ret[i] = pivots[i].getDistance(sampleObject);
        }
        return ret;
    }

    public float getDistanceToHyperplane(float pivot0Dist, float pivot1Dist, float pivotDist) {
        float ret = (pivot0Dist * pivot0Dist - pivot1Dist * pivot1Dist) / (2 * pivotDist);
        return Math.abs(ret);
    }

}
