package sf.pivotAnalyzer.uncorrelatedSelectorImpl;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;
import sf.objects.impl.Sketch;
import sf.pivotAnalyzer.uncorrelatedSelector.InvertedSketchesCorrelationCounter;

/**
 *
 * @author xmic
 */
public class SmallDistancesSteps extends InvertedSketchesCorrelationCounter {

    public SmallDistancesSteps(List<Sketch> invertedSketches, int lengthOfInvertedSketch) {
        super(invertedSketches, lengthOfInvertedSketch, 0);
    }

    @Override
    /**
     * Ingores tuplesLength parameter!
     */
    protected int[] selectLowCorrelatedSubset(int tuplesLength, float[][] sketchesCorrelations, int sketchSize) {
        Integer[] retIntegers = null;
        int correlationStep = Math.max(sketchesCorrelations.length * (sketchesCorrelations.length - 1) / 2 / 100, sketchSize * sketchSize);
//        int curPositionLimit = 183295;;
        int curPositionLimit = 260000;;
        List<Float> correlations = new ArrayList<>();
        for (int i = 0; i < sketchesCorrelations.length - 1; i++) {
            for (int j = i + 1; j < sketchesCorrelations.length; j++) {
                correlations.add(sketchesCorrelations[i][j]);
            }
        }
        Collections.sort(correlations);
        while (retIntegers == null) {
            if (curPositionLimit != 0) {
                Logger.getLogger(InvertedSketchesCorrelationCounter.class.getName()).log(Level.INFO, "No clique found using {0} smallest correlations. (Up to {1})", new Object[]{curPositionLimit, correlations.get(curPositionLimit)});
            }
            curPositionLimit += correlationStep;
            float curCorrelationLimit = correlations.get(curPositionLimit);
            int[] cands = new int[sketchesCorrelations.length];
            Set<Integer>[] adjacentMatrix = new Set[sketchesCorrelations.length];
            for (int i = 0; i < sketchesCorrelations.length; i++) {
                adjacentMatrix[i] = new HashSet<>();
            }
            Arrays.fill(cands, 0);
            for (int i = 0; i < sketchesCorrelations.length - 1; i++) {
                for (int j = i + 1; j < sketchesCorrelations.length; j++) {
                    if (sketchesCorrelations[i][j] < curCorrelationLimit) {
                        cands[i]++;
                        cands[j]++;
                        adjacentMatrix[i].add(j);
                        adjacentMatrix[j].add(i);
                    }
                }
            }
            boolean removedOk = false;
            while (!removedOk) {
                removedOk = true;
                for (int i = 0; i < sketchesCorrelations.length; i++) {
                    if (adjacentMatrix[i].size() < sketchSize) {
                        removedOk = false;
                        adjacentMatrix[i].clear();
                        for (int j = 0; j < sketchesCorrelations.length; j++) {
                            adjacentMatrix[j].remove(i);
                        }
                    }
                }
            }
            List<int[]> indexesCounts = new ArrayList<>();
            for (int i = 0; i < sketchesCorrelations.length; i++) {
                indexesCounts.add(new int[]{i, cands[i]});
            }
            retIntegers = searchClique(indexesCounts, adjacentMatrix, sketchSize);
        }
        int[] ret = new int[retIntegers.length];
        for (int i = 0; i < retIntegers.length; i++) {
            ret[i] = retIntegers[i];
        }
        float highestCorrelation = getHighestCorrelation(retIntegers, sketchesCorrelations);
        Logger.getLogger(InvertedSketchesCorrelationCounter.class.getName()).log(Level.INFO, "Found subset with worst correlation {0}.", highestCorrelation);
        return ret;
    }

    private Integer[] searchClique(List<int[]> indexesCounts, Set<Integer>[] adjacentMatrix, int sketchSize) {
        Collections.sort(indexesCounts, new SmallDistancesSteps.IntArrayComparatorBySecondValue());
        if (indexesCounts.get(indexesCounts.size() - sketchSize + 1)[1] < sketchSize) {
            return null;
        }
        List<Integer> candidates = new ArrayList<>();
        Integer[] ret = searchCliqueRecursive(indexesCounts, candidates, adjacentMatrix, sketchSize, 0, new ArrayList<Integer>()).toArray(new Integer[0]);
        if (ret.length < sketchSize) {
            return null;
        }
        return ret;
    }

    private long lastLog = 0;

    private List<Integer> searchCliqueRecursive(List<int[]> indexesCounts, List<Integer> candidates, Set<Integer>[] adjacentMatrix, int sketchSize, int start, List<Integer> logged) {
        if (candidates.size() == sketchSize) {
            return candidates;
        }
        for (int i = start; i < adjacentMatrix.length - sketchSize + candidates.size() + 1; i++) {
            if (candidates.isEmpty() && System.currentTimeMillis() - lastLog > 10000) {
                Logger.getLogger(InvertedSketchesCorrelationCounter.class.getName()).log(Level.INFO, "New start from index {0}.", i);
                lastLog = System.currentTimeMillis();
            }
            Integer index = indexesCounts.get(i)[0];
            boolean canBeAdd = true;
            for (Integer cand : candidates) {
                if (!adjacentMatrix[cand].contains(index)) {
                    canBeAdd = false;
                    break;
                }
            }
            if (canBeAdd) {
                candidates.add(index);
                if (!logged.contains(candidates.size())) {
                    logged.add(candidates.size());
                    Logger.getLogger(InvertedSketchesCorrelationCounter.class.getName()).log(Level.INFO, "Found uncorrelated {0} bits.", candidates.size());
                }
                if (candidates.size() == sketchSize) {
                    return candidates;
                }
                List<Integer> ret = searchCliqueRecursive(indexesCounts, candidates, adjacentMatrix, sketchSize, i + 1, logged);
                if (ret.equals(candidates) && candidates.size() != sketchSize) {
                    candidates.remove(index);
                } else if (candidates.size() == sketchSize) {
                    return candidates;
                }
            }
        }
        return candidates;
    }

    private static class IntArrayComparatorBySecondValue implements Comparator<int[]> {

        @Override
        public int compare(int[] o1, int[] o2) {
            int ret = Integer.compare(o1[1], o2[1]);
            if (ret != 0) {
                return ret;
            }
            return Integer.compare(o1[0], o2[0]);
        }
    }

}
