/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package sf.recallEstimation.hamThresholdEstimation;

import org.apache.commons.math4.analysis.UnivariateFunction;
import sf.recallEstimation.UnivariateFunctionImpl.PxbSmallBinAnalogue;

/**
 *
 * @author xmic
 */
public class MapperDistDToHammingThresholdT {

    private final double maxDistance;
    private final Double sketchesIDim;
    private final UnivariateFunction pxProbSkBitsDif;
    private final int skLength;
    private final double thresholdPcum;

    public static final Float ERROR_THRESHOLD_P_CUM = 0.0000001f;

    public MapperDistDToHammingThresholdT(double maxDistance, double sketchesIDim, int skLength, UnivariateFunction pxProbSkBitsDif, double thresholdPcum) {
        this.maxDistance = maxDistance;
        this.sketchesIDim = sketchesIDim;
        this.pxProbSkBitsDif = pxProbSkBitsDif;
        this.skLength = skLength;
        this.thresholdPcum = thresholdPcum;
    }

    public float findXMappedToT(int t) {
        if (t == 0) {
            return 0;
        }
        double curX = 0f;
        double curErr = Float.MAX_VALUE;
        double step = maxDistance / 2;
        while (curErr > ERROR_THRESHOLD_P_CUM) {
            double curProb = estimateProbabilityXUpToB(curX, t);
            curErr = curProb - thresholdPcum;
            if (curErr > 0) {
                curX += step;
            } else {
                curX -= step;
            }
            step = step / 2;
            if (step == 0) {
                break;
            }
            curErr = Math.abs(curErr);
        }
        return (float) curX;
    }

    private double estimateProbabilityXUpToB(double x, int t) {
        double[] px0b = new double[skLength + 1];
        double[] px0bCumulative = new double[skLength + 1];

        double pxProbSkBitsDifX = pxProbSkBitsDif.value(x);
        PxbSmallBinAnalogue pxb = new PxbSmallBinAnalogue(skLength, sketchesIDim, pxProbSkBitsDifX);

        double sum = 0;
        for (int i = 0; i <= skLength; i++) {
            px0b[i] = pxb.value(i);
            sum += px0b[i];
            px0bCumulative[i] = sum;
        }
        return px0bCumulative[t];
    }

}
