/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package sf.recallEstimation.integrating;

import java.util.logging.Level;
import java.util.logging.Logger;
import org.apache.commons.math4.analysis.integration.BaseAbstractUnivariateIntegrator;
import org.apache.commons.math4.analysis.integration.TrapezoidIntegrator;
import org.apache.commons.math4.exception.MaxCountExceededException;
import org.apache.commons.math4.exception.TooManyEvaluationsException;

/**
 *
 * @author xmic
 */
public class OnePeakPositiveFunctionIntegrator extends BaseAbstractUnivariateIntegrator {

    public static final Double INNER_STEP_PRECISION = 1E-3;
    public static final Double INNER_PRECISION = 1E-15;
    public static final Double INNER_PRECISION_Y = 1E-100;
    public static final Double RELATIVE_PRECISION = 1E-10;

    protected Double meanX;

    public OnePeakPositiveFunctionIntegrator(double minX, double maxX) {
        super(1, Integer.MAX_VALUE);
        meanX = null;
    }

    public OnePeakPositiveFunctionIntegrator(double mean) {
        super(1, Integer.MAX_VALUE);
        this.meanX = mean;
    }

    @Override
    protected double doIntegrate() throws TooManyEvaluationsException, MaxCountExceededException {
        double minX = getMin();
        double maxX = getMax();
        if (meanX == null) {
            meanX = findApproxMean(minX, maxX, null);
        }
        double minXNonInft = minX == Double.NEGATIVE_INFINITY ? 0 : minX;
        double firstStep = (maxX - minXNonInft) / 65536;
        minX = findPeakStart(minX, firstStep, meanX == null ? maxX : Math.min(meanX, maxX));
        maxX = findPeakEnd(maxX, firstStep, meanX == null ? minXNonInft : Math.max(minXNonInft, meanX));
        Logger.getLogger(OnePeakPositiveFunctionIntegrator.class.getName()).log(Level.INFO, "Integral new bounds are from {0} to {1}", new Object[]{minX, maxX});
        if (minX == maxX) {
            return 0;
        }
        BaseAbstractUnivariateIntegrator integrator = new TrapezoidIntegrator(RELATIVE_PRECISION, TrapezoidIntegrator.DEFAULT_ABSOLUTE_ACCURACY, 3, TrapezoidIntegrator.TRAPEZOID_MAX_ITERATIONS_COUNT);
        double ret = integrator.integrate(Integer.MAX_VALUE, function, minX, maxX);
        Logger.getLogger(OnePeakPositiveFunctionIntegrator.class.getName()).log(Level.INFO, "Integral evaluated by {0} using {1} function evaluations.", new Object[]{integrator.getClass().getSimpleName(), integrator.getEvaluations()});
        return ret;
    }

    private double findPeakStart(double minX, double step, double limit) {
        if (minX == Double.NEGATIVE_INFINITY) {
            return minX;
        }
        double ret = minX;
        if (step < INNER_STEP_PRECISION) {
            return ret;
        }
        if (ret >= limit) {
            return limit;
        }
        double val = computeObjectiveValue(minX);
        while (val < INNER_PRECISION_Y && ret < limit) {
            ret = minX;
            minX += step;
            if (ret == minX) {// step is too small
                break;
            }
            val = computeObjectiveValue(minX);
        }
        return findPeakStart(ret, step / 2, limit);
    }

    private double findPeakEnd(double maxX, double step, double limit) {
        double ret = maxX;
        if (step < INNER_STEP_PRECISION || ret == limit) {
            return ret;
        }
        double val = computeObjectiveValue(maxX);
        while (val < INNER_PRECISION_Y && ret > limit) {
            ret = maxX;
            maxX -= step;
            if (ret == maxX) {// step is too small
                break;
            }
            val = computeObjectiveValue(maxX);
        }
        return findPeakEnd(ret, step / 2, limit);
    }

    @SuppressWarnings("InfiniteRecursion")
    private Double findApproxMean(double minX, double maxX, Double step) {
        if (minX == Double.NEGATIVE_INFINITY) {
            minX = 0;
        }
        if (step == null) {
            step = (maxX - minX) / 2;
        }
        if (step < INNER_STEP_PRECISION) {
            return null;
        }
        double x = minX;
        while (x < maxX) {
            double y = computeObjectiveValue(x);
            if (y > INNER_PRECISION) {
                return x;
            }
            if (x == x + step) {
                break;
            }
            x += step;
        }
        return findApproxMean(minX, maxX, step / 2);
    }

}
