/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package sf.recallEstimation.pxEvaluator;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.Set;
import java.util.SortedMap;
import java.util.TreeMap;
import java.util.logging.Level;
import java.util.logging.Logger;
import main.tools.Tools;
import messif.objects.LocalAbstractObject;
import org.apache.commons.math4.stat.descriptive.moment.Mean;
import org.apache.commons.math4.stat.descriptive.moment.Variance;
import sf.objects.impl.Sketch;

/**
 *
 * @author xmic
 */
public class PxEvaluatorOnCreatedSketches {

    public final int MIN_NUMBER_OF_EXAMPLES_FOR_BIT_TO_COUNT = 10;

    private final Map<String[], Float> distancesOnData;
    private final String sketchPath;
    private final int sketchLength;
    private final double distInterval;

    public PxEvaluatorOnCreatedSketches(Map<String[], Float> distancesOnData, String sketchPath, int sketchLength, double distInterval) {
        this.distancesOnData = distancesOnData;
        this.sketchPath = sketchPath;
        this.sketchLength = sketchLength;
        this.distInterval = distInterval;
    }

    private float iDim = 0;

    public float getiDim() {
        return iDim;
    }

    /**
     * Returns pairs [x,y] for plot
     *
     * @return x .. orig distance, y .. probability
     */
    public SortedMap<Float, Float> evaluateProbabilities() {
        Map<String, LocalAbstractObject> sketches = Tools.getObjectsAsLocatorMap(Sketch.class.getName(), sketchPath, -1);
        iDim = evaluateIDim(sketches);
        List<SortedMap<Float, Float>> bitProbabilities = evalBitProbsAndHamDistDistrs(sketches);
        SortedMap<Float, Float> ret = PxEvaluator.createAverages(bitProbabilities, sketchLength);
        float x = 0;
        Float lastKey = ret.lastKey();
        while (x < lastKey) {
            if (!ret.containsKey(x)) {
                ret.put(x, 0f);
            }
            x += distInterval;
        }
//        if (!ret.containsKey(0f)) {
//            ret.put(0f, 0f);
//        }
        return ret;
    }

    private List<SortedMap<Float, Float>> evalBitProbsAndHamDistDistrs(Map<String, LocalAbstractObject> sketches) {
        // probs evaluation
        // for each bit of sketches contains map to create plot [x,y],
        // where x is original distance and y is a probability that coresponding bit differs
        List<SortedMap<Float, Float>> ret = new ArrayList<>();
        for (int i = 0; i < sketchLength; i++) {
            ret.add(new TreeMap<>());
        }
        Map<Float, Integer> denominators = new HashMap<>();
        Iterator<Map.Entry<String[], Float>> it = distancesOnData.entrySet().iterator();
        int counter = 0;
        while (it.hasNext()) {
            counter++;
            Map.Entry<String[], Float> value = it.next();
            String[] locators = value.getKey();
            float dist = value.getValue();
            if (counter % (distancesOnData.size() / 20) == 0) {
                Logger.getLogger(TestPxEvaluator.class.getName()).log(Level.INFO, "Processed {0} percent", counter / (distancesOnData.size() / 100));
            }
            // probs evaluation
            Sketch sk1 = (Sketch) sketches.get(locators[0]);
            Sketch sk2 = (Sketch) sketches.get(locators[1]);
            if (sk1 == null || sk2 == null) {
                throw new RuntimeException("Error: missing sketch. " + locators[0] + ": " + sk1 + "; " + locators[1] + ": " + sk2);
            }
            PxEvaluator.updateProbabilities(dist, sk1, sk2, ret, denominators, distInterval, sketchLength);
        }
        for (int i = 0; i < sketchLength; i++) {
            SortedMap<Float, Float> map = ret.get(i);
            Set<Float> keySet = map.keySet();
            for (Float key : keySet) {
                Float value = map.get(key);
                int denominator = denominators.get(key);
                if (MIN_NUMBER_OF_EXAMPLES_FOR_BIT_TO_COUNT < denominator) {
                    map.put(key, value / denominator);
                } else {
                    map.put(key, 0f);
                }
            }
        }
        return ret;
    }

    public static Map<String[], Float> evaluateDistances(List<LocalAbstractObject> data) {
        Map<String[], Float> ret = new HashMap<>();
        int counter = 0;
        int total = (data.size() - 1) * data.size() / 2;
        for (int i = 0; i < data.size() - 1; i++) {
            LocalAbstractObject obj1 = data.get(i);
            for (int j = i + 1; j < data.size(); j++) {
                LocalAbstractObject obj2 = data.get(j);
                String[] key = new String[]{obj1.getLocatorURI(), obj2.getLocatorURI()};
                ret.put(key, obj1.getDistance(obj2));
                if (counter % 10000 == 0) {
                    Logger.getLogger(PxEvaluatorOnCreatedSketches.class.getName()).log(Level.INFO, "Evaluating distances x in the metric space. Processed {0} dists, i.e. {1} %", new Object[]{counter, 100f * counter / total});
                }
                counter++;
            }
        }
//        Random r = new Random();
//        int total = 2000000;
//        for (int counter = 0; counter < total; counter++) {
//            int i = r.nextInt(data.size());
//            int j = r.nextInt(data.size());
//            while(j == i){
//                j = r.nextInt(data.size());
//            }
//            LocalAbstractObject obj1 = data.get(i);
//            LocalAbstractObject obj2 = data.get(j);
//            String[] key = new String[]{obj1.getLocatorURI(), obj2.getLocatorURI()};
//            ret.put(key, obj1.getDistance(obj2));
//            if (counter % 100000 == 0) {
//                Logger.getLogger(PxEvaluatorOnCreatedSketches.class.getName()).log(Level.INFO, "Evaluating distances x in the metric space. Processed {0} dists, i.e. {1} %", new Object[]{counter, 100f * counter / total});
//            }
//        }
        return ret;
    }

    private float evaluateIDim(Map<String, LocalAbstractObject> sketchesMap) {
        System.err.println("Starting iDim evaluation");
        Mean m = new Mean();
        Variance v = new Variance();
        List<LocalAbstractObject> sketches = new ArrayList<>(sketchesMap.values());
        double[] dists = new double[1000000];
        Random r = new Random();
        for (int i = 0; i < dists.length; i++) {
            int i1 = r.nextInt(sketches.size());
            int i2 = r.nextInt(sketches.size());
            dists[i] = sketches.get(i1).getDistance(sketches.get(i2));
        }
//        int size = sketches.size() * (sketches.size() - 1) / 2;
//        if (size <= 0 || size > 10000000) {
//            size = 10000000;
//        }
//        double[] dists = new double[size];
//        int c = 0;
//        for (int i = 0; i < sketches.size() - 1; i++) {
//            LocalAbstractObject sk1 = sketches.get(i);
//            for (int j = i + 1; j < sketches.size(); j++) {
//                if (dists.length == c) {
//                    break;
//                }
//                LocalAbstractObject sk2 = sketches.get(j);
//                dists[c] = sk1.getDistance(sk2);
//                c++;
//            }
//        }
        double mean = m.evaluate(dists);
        double var = v.evaluate(dists);
        float ret = (float) (mean * mean / 2 / var);
        System.err.println("Finish iDim evaluation: " + ret);
        return ret;
    }

}
