/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package sf.separability;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;
import main.tools.Tools;
import org.apache.commons.math4.analysis.BivariateFunction;
import sf.separabilityAnalytique.BivariateIntegrator;

/**
 *
 * @author xmic
 */
public abstract class QualityOfSketchingTechnique {

    protected final double maxDistance;

    protected QualityOfSketchingTechnique(double maxDistance) {
        this.maxDistance = maxDistance;
    }

    public double evaluateNormalized() {
        return evaluateNormalized(-1d);
    }

    public double evaluateNormalized(double maxX1, double stepIntegrating) {
        BivariateFunction sepx1x2 = createInstanceSepDistsx1x2InHammSpace();
        BivariateIntegrator integrator;
        if (maxX1 >= 0) {
            integrator = new BivariateIntegrator(sepx1x2, 0, maxDistance, maxX1, stepIntegrating);
        } else {
            integrator = new BivariateIntegrator(sepx1x2, 0, maxDistance, maxDistance, stepIntegrating);
        }
        double quality = integrator.evaluate();
        return quality / maxDistance / maxDistance;
    }

    public double evaluateNormalized(double maxX1) {
        BivariateFunction sepx1x2 = createInstanceSepDistsx1x2InHammSpace();
        BivariateIntegrator integrator;
        if (maxX1 >= 0) {
            integrator = new BivariateIntegrator(sepx1x2, 0, maxDistance, maxX1);
        } else {
            integrator = new BivariateIntegrator(sepx1x2, 0, maxDistance);
        }
        double quality = integrator.evaluate();;;
//        return quality / maxDistance / maxDistance;
        return quality;
    }

    protected abstract BivariateFunction createInstanceSepDistsx1x2InHammSpace();

    public static float getStepSize(float maxDistance, int countOfIntervals) {
        return maxDistance / countOfIntervals;
    }

    public static Map<String, Double> getMedianRecallsSIFT() {
        Map<String, Double> ret = new HashMap<>();
        ret.put("BP_50_064", 0.44);
        ret.put("BP_50_128", 0.61);
        ret.put("BP_50_192", 0.71);
        ret.put("BP_50_256", 0.74);
        ret.put("GHP_50_064", 0.51);
        ret.put("GHP_50_128", 0.73);
        ret.put("GHP_50_192", 0.82);
        ret.put("GHP_50_256", 0.87);
        ret.put("GHP_80_064", 0.3);
        ret.put("GHP_80_128", 0.54);
        ret.put("GHP_80_192", 0.66);
        ret.put("GHP_80_256", 0.76);
        ret.put("NSP_50_064", 0.53);
        ret.put("NSP_50_128", 0.79);
        ret.put("THRR_50_064", 0.57);
        ret.put("THRR_50_128", 0.79);
        return ret;
    }

    public static Map<String, Double> getMedianRecallsDeCAF() {
        Map<String, Double> ret = new HashMap<>();
        ret.put("BP_50_256", 0.11);
        ret.put("BP_50_192", 0.1);
        ret.put("BP_50_064", 0.06);
        ret.put("BP_50_128", 0.09);
        ret.put("GHP_80_064", 0.34);
        ret.put("NSP_50_064", 0.4);
        ret.put("GHP_50_064", 0.42);
        ret.put("THRR_50_064", 0.56);
        ret.put("GHP_80_128", 0.57);
        ret.put("GHP_80_192", 0.68);
        ret.put("GHP_50_128", 0.62);
        ret.put("NSP_50_128", 0.63);
        ret.put("GHP_50_192", 0.71);
        ret.put("GHP_50_256", 0.77);
        ret.put("GHP_80_256", 0.78);
        ret.put("NSP_50_192", 0.74);
        ret.put("THRR_50_128", 0.75);
        ret.put("NSP_50_256", 0.8);
        ret.put("THRR_50_192", 0.84);
        ret.put("THRR_50_256", 0.88);
        return ret;
    }

    public static Map<String, Double> getMedianRecallsDeCAF100NoReranking() {
        Map<String, Double> ret = new HashMap<>();
        ret.put("BP_50_064", 0.01);
        ret.put("BP_50_128", 0.02);
        ret.put("BP_50_192", 0.03);
        ret.put("BP_50_256", 0.03);
        ret.put("GHP_50_064", 0.08);
        ret.put("GHP_50_128", 0.14);
        ret.put("GHP_50_192", 0.18);
        ret.put("GHP_50_256", 0.2);
        ret.put("GHP_80_064", 0.06);
        ret.put("GHP_80_128", 0.13);
        ret.put("GHP_80_192", 0.16);
        ret.put("GHP_80_256", 0.21);
        ret.put("NSP_50_064", 0.07);
        ret.put("NSP_50_128", 0.15);
        ret.put("NSP_50_192", 0.2);
        ret.put("NSP_50_256", 0.23);
        ret.put("THRR_50_064", 0.12);
        ret.put("THRR_50_128", 0.19);
        ret.put("THRR_50_192", 0.24);
        ret.put("THRR_50_256", 0.27);
        return ret;
    }

    public static double[] getCorrelations(Map<String, Double>[] qualityEstimations, Map<String, Double> medianRecalls) {
        double[] recalls = new double[medianRecalls.size()];
        Set<String> sketchFiles = medianRecalls.keySet();
        double[] ret = new double[qualityEstimations.length];
        for (int i = 0; i < qualityEstimations.length; i++) {
            Map<String, Double> qualityEstimation = qualityEstimations[i];
            if (qualityEstimation.size() != medianRecalls.size()) {
                Logger.getLogger(QualityOfSketchingTechnique.class.getName()).log(Level.WARNING, "Different number of estimations and recall values");
            }
            double[] estimations = new double[qualityEstimation.size()];
            Iterator<String> it = sketchFiles.iterator();
            int c = 0;
            while (it.hasNext()) {
                String key = it.next();
                recalls[c] = medianRecalls.get(key);
                estimations[c] = qualityEstimation.get(key);
                c++;
            }
            ret[i] = Tools.pearsonCorrelationCoefficient(recalls, estimations);
        }
        return ret;
    }

}
