/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package sf.separability.measuredGeneral;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.SortedMap;
import java.util.TreeMap;
import java.util.logging.Level;
import java.util.logging.Logger;

import main.tools.Tools;
import messif.objects.LocalAbstractObject;

import org.apache.commons.math4.analysis.BivariateFunction;

import sf.objects.impl.Sketch;
import sf.separability.QualityOfSketchingTechnique;
import sf.separabilityMeasured.QualityOfSketchingTechniqueMeasured;

/**
 *
 * @author xmic
 */
public class QualityOfSketchingTechGenlMeas extends QualityOfSketchingTechnique {

    public static final Float IMPLICIT_T = 1f;
    private final double thT;
    private final BivariateFunction probOrderPreserve;
    private final BivariateFunction probOrderSwap;
    private final HellingerDistImpl hellingerDist;
    private final Map<Integer, double[][]> conditionalDistribsForXIntervals;
    private static int skippedBuckets;

    public QualityOfSketchingTechGenlMeas(Map<Integer, Double> distanceXDensity, Map<String[], Float> distsOnData, String sketchesPath, int sketchLength, double cumulativeThresholdT, int countOfIntervals, float stepSize) {
        super(stepSize * countOfIntervals);
        if (cumulativeThresholdT <= 0) {
            thT = IMPLICIT_T;
        } else {
            thT = cumulativeThresholdT;
        }
        skippedBuckets = 0;
        conditionalDistribsForXIntervals = conditionalDistributionOfHamDist(distsOnData, sketchesPath, countOfIntervals, sketchLength, stepSize);
        Logger.getLogger(QualityOfSketchingTechGenlMeas.class.getName()).log(Level.INFO, "Skipped {0} buckets due to a low number of objects in it", skippedBuckets);
        probOrderPreserve = new ProbOrderPreserveImpl(distanceXDensity, conditionalDistribsForXIntervals, stepSize);
        probOrderSwap = new ProbOrderSwapImpl(distanceXDensity, conditionalDistribsForXIntervals, stepSize);
        hellingerDist = new HellingerDistImpl(distanceXDensity, conditionalDistribsForXIntervals, stepSize);
    }

    private static Map<Integer, double[][]> conditionalDistributionOfHamDist(Map<String[], Float> distsOnData, String sketchesPath, int countOfIntervals, int sketchLength, float stepSize) {
        Iterator<Map.Entry<String[], Float>> distsXIt = distsOnData.entrySet().iterator();
        SortedMap<Integer, List<Float>> distancesXAndHammings = initMapIntListFloat(countOfIntervals);
        Map<String, LocalAbstractObject> sketches = Tools.getObjectsAsLocatorMap(Sketch.class.getName(), sketchesPath, -1);
        while (distsXIt.hasNext()) {
            Map.Entry<String[], Float> next = distsXIt.next();
            float dist = next.getValue();
            if (dist >= stepSize * countOfIntervals) {
                continue;
            }
            int key = (int) (dist / stepSize);
            String[] locators = next.getKey();
            float hammDist = sketches.get(locators[0]).getDistance(sketches.get(locators[1]));
            distancesXAndHammings.get(key).add(hammDist);
        }
        Map<Integer, double[][]> ret = new HashMap<>();
        for (Iterator<Map.Entry<Integer, List<Float>>> itX = distancesXAndHammings.entrySet().iterator(); itX.hasNext();) {
            Map.Entry<Integer, List<Float>> next = itX.next();
            int idx = next.getKey();
            List<Float> hammingDistances = next.getValue();
            if (hammingDistances.size() < QualityOfSketchingTechniqueMeasured.MIN_NUMBER_OF_DISTANCES) {
                if (!hammingDistances.isEmpty()) {
                    skippedBuckets++;
                }
                hammingDistances.clear();
            }
            double[][] probBiCumBi = evaluateDiscreteDensities(hammingDistances, sketchLength);
            if (!hammingDistances.isEmpty() && Math.abs(probBiCumBi[1][sketchLength] - 1) > 1E-4) {
                throw new RuntimeException("Wrong cumulative probability for bucket with index " + idx + " (" + probBiCumBi[1][sketchLength] + ")");
            }
            ret.put(idx, probBiCumBi);
        }
        return ret;
    }

    private static Map<Integer, Double> initMapIntFloat(int countOfIntervals) {
        Map<Integer, Double> ret = new HashMap<>();
        for (int i = 0; i < countOfIntervals; i++) {
            ret.put(i, 0d);
        }
        return ret;
    }

    private static SortedMap<Integer, List<Float>> initMapIntListFloat(int countOfIntervals) {
        SortedMap<Integer, List<Float>> ret = new TreeMap<>();
        for (int i = 0; i < countOfIntervals; i++) {
            ret.put(i, new ArrayList<>());
        }
        return ret;
    }

    private static int distsCount = 0;

    public static Map<Integer, Double> evaluateDensityOfDistX(Map<String[], Float> distsOnData, int countOfIntervals, float stepSize) {
        Map<Integer, Double> ret = initMapIntFloat(countOfIntervals);
        for (Iterator<Map.Entry<String[], Float>> it = distsOnData.entrySet().iterator(); it.hasNext();) {
            Map.Entry<String[], Float> entry = it.next();
            float distX = entry.getValue();
            if (distX >= countOfIntervals * stepSize) {
                continue;
            }
            distsCount++;
            int idx = (int) (distX / stepSize);
            double value = ret.get(idx) + 1;
            ret.put(idx, value);
        }
        for (Integer key : ret.keySet()) {
            double newValue = ret.get(key) / distsCount;
            ret.put(key, newValue);
        }
//        float sum = 0;
//        System.out.println("AAAAA");
//        for (Integer key : ret.keySet()) {
//            sum += ret.get(key);
//            System.out.println(ret.get(key) + ";" + sum);
//        }
        return ret;
    }

    private static double[][] evaluateDiscreteDensities(List<Float> hammingDistances, int sketchLength) {
        double[][] ret = new double[2][sketchLength + 1];
        if (hammingDistances.isEmpty()) {
            return ret;
        }
        for (Float hamDist : hammingDistances) {
            int hamDistI = hamDist.intValue();
            ret[0][hamDistI]++;
        }
        for (int i = 0; i <= sketchLength; i++) {
            ret[0][i] = ret[0][i] / hammingDistances.size();
        }
        ret[1][0] = ret[0][0];
        for (int i = 1; i <= sketchLength; i++) {
            ret[1][i] = ret[1][i - 1] + ret[0][i];
        }
        return ret;
    }

    @Override
    protected BivariateFunction createInstanceSepDistsx1x2InHammSpace() {
        return probOrderPreserve;
    }

    public double evaluateNormalizedPxPy(Map<Integer, Double> distanceXDensity, double stepSize) {
        double quality = 0;
        int numberOfBuckets = (int) (maxDistance / stepSize);

        double area = 0;
        double cumulative = 0;
//        for (int i2 = 1; i2 < numberOfBuckets; i2++) {
        for (int i2 = 0; i2 < numberOfBuckets; i2++) {
            if (cumulative >= thT) {	//area
                Logger.getLogger(QualityOfSketchingTechGenlMeas.class.getName()).log(Level.INFO, "*** t={0}*** area={1}***", new Object[]{i2 * stepSize, area});
                break;
            }
            Double dx2 = distanceXDensity.get(i2);
            if (dx2 == null || dx2 == 0) {
                continue;
            }
            double qualityTmp = 0;
            double areaTmp = 0;
//            for (int i1 = 0; i1 < i2; i1++) {
            for (int i1 = i2 + 1; i1 < numberOfBuckets; i1++) {
                Double dx1 = distanceXDensity.get(i1);
                if (dx1 == null || dx1 == 0) {
                    continue;
                }
                qualityTmp += probOrderSwap.value(i2, i1) * dx1;
//                qualityTmp += probOrderSwap.value(i1, i2) * dx1;
                areaTmp += dx1;
            }

            cumulative += dx2;
            quality += qualityTmp * dx2;
            area += areaTmp * dx2;

        }
        return 1 - quality / area;
    }

    public double evaluateNormalized(Map<Integer, Double> distanceXDensity, double stepSize) {

        double quality = 0;
        int numberOfBuckets = (int) (maxDistance / stepSize);

        double area = 0;
        double cumulative = 0;
//        for (int i2 = 1; i2 < numberOfBuckets; i2++) {
        for (int i2 = 0; i2 < numberOfBuckets; i2++) {
            if (cumulative >= thT) {	//area
                Logger.getLogger(QualityOfSketchingTechGenlMeas.class.getName()).log(Level.INFO, "*** t=" + i2 * stepSize + "*** area=" + area + "***");
                break;
            }
            Double dx2 = distanceXDensity.get(i2);
            if (dx2 == null || dx2 == 0) {
                continue;
            }
            double qualityTmp = 0;
            double areaTmp = 0;
//            for (int i1 = 0; i1 < i2; i1++) {
            for (int i1 = i2 + 1; i1 < numberOfBuckets; i1++) {
                Double dx1 = distanceXDensity.get(i1);
                if (dx1 == null || dx1 == 0) {
                    continue;
                }
                dx1 = 1.0;
//                qualityTmp += probOrderSwap.value(i1, i2) * dx1;
                qualityTmp += probOrderSwap.value(i2, i1) * dx1;
                areaTmp += dx1;
            }

            cumulative += dx2;
            dx2 = 1.0;
            quality += qualityTmp * dx2;
            area += areaTmp * dx2;

        }
        return 1 - quality / area;
    }

    public double evaluateHellinger(Map<Integer, Double> distanceXDensity, double stepSize) {

        double bdist = 0;
        int numberOfBuckets = (int) (maxDistance / stepSize);

        double area = 0;
        double cumulative = 0;
//        for (int i2 = 1; i2 < numberOfBuckets; i2++) {
        for (int i2 = 0; i2 < numberOfBuckets; i2++) {
            if (cumulative >= thT) {	//area
                Logger.getLogger(QualityOfSketchingTechGenlMeas.class.getName()).log(Level.INFO, "*** t=" + i2 * stepSize + "*** area=" + area + "***");
                break;
            }
            Double dx2 = distanceXDensity.get(i2);
            if (dx2 == null || dx2 == 0) {
                continue;
            }
            double bDistTmp = 0;
            double areaTmp = 0;
//            for (int i1 = 0; i1 < i2; i1++) {
            for (int i1 = i2 + 1; i1 < numberOfBuckets; i1++) {
                Double dx1 = distanceXDensity.get(i1);
                if (dx1 == null || dx1 == 0) {
                    continue;
                }
                dx1 = 1.0;
//                bDistTmp += hellingerDist.value(i1, i2) * dx1;
                bDistTmp += hellingerDist.value(i2, i1) * dx1;
                areaTmp += dx1;
            }

            cumulative += dx2;
            dx2 = 1.0;
            bdist += bDistTmp * dx2;
            area += areaTmp * dx2;
        }
        return bdist / area;
    }
// public double evaluateMYBhattacharyya ( Map<Integer, Double> distanceXDensity, double stepSize) {
//	 	
//     double overlapProb = 0;
//     int numberOfBuckets = (int) (maxDistance / stepSize);
//   
//     double area=0;
//     double cumulative=0;
//     for (int i2 = 1; i2 < numberOfBuckets; i2++) {
//     	if(cumulative>=thT) {	//area
//     		Logger.getLogger(QualityOfSketchingTechGenlMeas.class.getName()).log(Level.INFO,"*** t=" +i2*stepSize+"*** area="+ area+"***");
//     		break;
//     	}
//     	Double dx2 = distanceXDensity.get(i2);
//     	if (dx2 == null  || dx2 == 0 ) { 
//     		continue;
//     	}	
//     	double overlapProbTmp=0;
//     	double areaTmp=0;
//     	for (int i1 = 0; i1 <i2 ; i1++) {  
//     		Double dx1 = distanceXDensity.get(i1);
//     		if (dx1== null  || dx1 == 0 )
//     			continue;
//     		dx1=1.0;
//     		overlapProbTmp += hellingerDist.value2(i1, i2)*dx1;
//     		areaTmp+=dx1;
//     	}
//     	
//     	cumulative +=dx2;
//     	dx2=1.0;
//     	overlapProb +=overlapProbTmp*dx2;
//     	area +=areaTmp*dx2;
//     	
//
//
//     };
//         
//     return  overlapProb/area;
// }

}
