/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package sf.separabilityMeasured;

import java.awt.geom.Point2D;
import java.io.FileNotFoundException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.SortedMap;
import java.util.TreeMap;
import main.tools.Tools;
import messif.objects.LocalAbstractObject;
import org.apache.commons.math4.stat.descriptive.moment.Mean;
import org.apache.commons.math4.stat.descriptive.moment.Variance;
import sf.objects.impl.Sketch;
import sf.separability.QualityOfSketchingTechnique;
import sf.separability.SepDistsx1x2InHammSpace;

/**
 *
 * @author xmic
 */
public class QualityOfSketchingTechniqueMeasured extends QualityOfSketchingTechnique {

    public static int MIN_NUMBER_OF_DISTANCES = 500;

    private final SepDistsx1x2InHammSpace separability;

    public QualityOfSketchingTechniqueMeasured(double maxDistance, List<Point2D.Float> meansOfPxbForX, List<Point2D.Float> variancesOfPxbForX) {
        this(maxDistance, meansOfPxbForX, variancesOfPxbForX, null);
    }

    public QualityOfSketchingTechniqueMeasured(double maxDistance, List<Point2D.Float> meansOfPxbForX, List<Point2D.Float> variancesOfPxbForX, List<Point2D.Float> sketchesPairsCounts) {
        super(maxDistance);
        separability = new SepDistsx1x2InHammSpaceMeasuredImpl(meansOfPxbForX, variancesOfPxbForX, sketchesPairsCounts);
    }

    @Override
    public SepDistsx1x2InHammSpace createInstanceSepDistsx1x2InHammSpace() {
        return separability;
    }

    public static void evaluateQualityOfSketches(Map<String[], Float> distsOnData, String sketchesPath, String output, double maxDistance, double maxX1, int distInterval) throws FileNotFoundException {
        List<Point2D.Float>[] pxbStats = evaluateMeansAndVariancesOfpxb(distsOnData, sketchesPath, distInterval, maxDistance);
        QualityOfSketchingTechnique evaluator = new QualityOfSketchingTechniqueMeasured(maxDistance, pxbStats[0], pxbStats[1]);
        double quality = evaluator.evaluateNormalized(maxX1);
        System.out.print("Normalized quality of sketching technique with max dist " + maxDistance + " is;");
        System.out.println(quality);
    }

    private static List<Point2D.Float>[] evaluateMeansAndVariancesOfpxb(Map<String[], Float> distsOnData, String sketchesPath, int distInterval, double maxDistance) {
        Iterator<Map.Entry<String[], Float>> distsXIt = distsOnData.entrySet().iterator();
        SortedMap<Integer, List<Float>> distancesXAndHammings = new TreeMap<>();
        Map<String, LocalAbstractObject> sketches = Tools.getObjectsAsLocatorMap(Sketch.class.getName(), sketchesPath, -1);
        while (distsXIt.hasNext()) {
            Map.Entry<String[], Float> next = distsXIt.next();
            int key = (int) (next.getValue() / distInterval) * distInterval;
            if (!distancesXAndHammings.containsKey(key)) {
                distancesXAndHammings.put(key, new ArrayList());
            }
            String[] locators = next.getKey();
            float hammDist = sketches.get(locators[0]).getDistance(sketches.get(locators[1]));
            distancesXAndHammings.get(key).add(hammDist);
        }
        Mean m = new Mean();
        Variance v = new Variance();
        List<Point2D.Float>[] ret = new List[2];
        ret[0] = new ArrayList<>();
        ret[1] = new ArrayList<>();
//        ret[2] = new ArrayList<>();
        ret[0].add(new Point2D.Float(0, 0));
        ret[1].add(new Point2D.Float(0, Float.MIN_VALUE));
//        ret[2].add(new Point2D.Float(-1, distsOnData.size()));
//        System.out.println("x;Numbers of distances;mean;variance");
        for (Iterator<Map.Entry<Integer, List<Float>>> itX = distancesXAndHammings.entrySet().iterator(); itX.hasNext();) {
            Map.Entry<Integer, List<Float>> next = itX.next();
            float x = next.getKey();
            double[] hamDists = Tools.floatsToDoubles(next.getValue());
            float mean = (float) m.evaluate(hamDists);
            float variance = (float) v.evaluate(hamDists);
//            ret[2].add(new Point2D.Float(x, hamDists.length));
            if (hamDists.length >= MIN_NUMBER_OF_DISTANCES && x <= maxDistance && x > 0) {
//                System.out.println(x + ";" + hamDists.length + ";" + mean + ";" + variance + ";");
                ret[0].add(new Point2D.Float(x, mean));
                ret[1].add(new Point2D.Float(x, variance));
            }
        }
        return ret;
    }

}
