/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package sf.separability.measuredGeneral;

import java.io.FileNotFoundException;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.SortedMap;
import java.util.TreeMap;
import main.tools.Tools;
import messif.objects.LocalAbstractObject;
import org.apache.commons.math4.analysis.BivariateFunction;
import sf.objects.impl.Sketch;
import sf.separability.QualityOfSketchingTechnique;
import sf.separabilityAnalytique.BivariateIntegratorLucia;
import sf.separabilityMeasured.QualityOfSketchingTechniqueMeasured;

/**
 *
 * @author xmic
 */
public class QualityOfSketchingTechGenlMeas extends QualityOfSketchingTechnique {

    private final BivariateFunction probOrderPreserve;
    private final Map<Integer, float[][]> conditionalDistribsForXIntervals;

    public QualityOfSketchingTechGenlMeas(Map<Integer, Float> distanceXDensity, Map<String[], Float> distsOnData, String sketchesPath, int sketchLength, String output, double maxDistance, double maxX1, int distInterval) {
        super(maxDistance);
        conditionalDistribsForXIntervals = conditionalDistributionOfHamDist(distsOnData, sketchesPath, distInterval, maxDistance, sketchLength);
        probOrderPreserve = new ProbOrderPreserveImpl(distanceXDensity, conditionalDistribsForXIntervals, distInterval);

//        System.out.println(probOrderPreserve.value(0, 0));
//        for (int start = 0; start < 8; start++) {
//            System.out.println(probOrderPreserve.value(start * 100, 700));
//            System.out.println(probOrderPreserve.value(start * 100, 350));
//        }
    }

    public void evaluateAndPrint(String output, double maxX1, int distInterval) throws FileNotFoundException {
        double quality = evaluateNormalized(maxX1, distInterval);
        if (output != null) {
            System.setOut(new PrintStream(output));
        }
        System.out.print("Normalized quality of sketching technique with max dist " + maxDistance + " on path is;");
        System.out.println(quality);
        System.out.println("\n\n\n\n");
    }

    private static Map<Integer, float[][]> conditionalDistributionOfHamDist(Map<String[], Float> distsOnData, String sketchesPath, int distInterval, double maxDistance, int sketchLength) {
        Iterator<Map.Entry<String[], Float>> distsXIt = distsOnData.entrySet().iterator();
        SortedMap<Integer, List<Float>> distancesXAndHammings = initMapIntListFloat(maxDistance, distInterval);
        Map<String, LocalAbstractObject> sketches = Tools.getObjectsAsLocatorMap(Sketch.class.getName(), sketchesPath, -1);
        while (distsXIt.hasNext()) {
            Map.Entry<String[], Float> next = distsXIt.next();
            float dist = next.getValue();
            if (dist > maxDistance) {
                continue;
            }
            int key = (int) (dist / distInterval);
            String[] locators = next.getKey();
            float hammDist = sketches.get(locators[0]).getDistance(sketches.get(locators[1]));
            distancesXAndHammings.get(key).add(hammDist);
        }
        Map<Integer, float[][]> ret = new HashMap<>();
        for (Iterator<Map.Entry<Integer, List<Float>>> itX = distancesXAndHammings.entrySet().iterator(); itX.hasNext();) {
            Map.Entry<Integer, List<Float>> next = itX.next();
            int idx = next.getKey();
            List<Float> hammingDistances = next.getValue();
            if (hammingDistances.size() < QualityOfSketchingTechniqueMeasured.MIN_NUMBER_OF_DISTANCES) {
                hammingDistances.clear();
            }
            float[][] probBiCumBi = evaluateDiscreteDensities(hammingDistances, sketchLength);
            ret.put(idx, probBiCumBi);
        }
        return ret;
    }

    private static Map<Integer, Float> initMapIntFloat(double maxDistance, int distInterval) {
        Map<Integer, Float> ret = new HashMap<>();
        for (int i = 0; i * distInterval <= maxDistance; i++) {
            ret.put(i, 0f);
        }
        return ret;
    }

    private static SortedMap<Integer, List<Float>> initMapIntListFloat(double maxDistance, int distInterval) {
        SortedMap<Integer, List<Float>> ret = new TreeMap<>();
        for (int i = 0; i * distInterval <= maxDistance; i++) {
            ret.put(i, new ArrayList<>());
        }
        return ret;
    }

    private static int distsCount = 0;

    public static Map<Integer, Float> evaluateDensityOfDistX(Map<String[], Float> distsOnData, double maxDistance, int distInterval) {
        Map<Integer, Float> ret = initMapIntFloat(maxDistance, distInterval);
        for (Iterator<Map.Entry<String[], Float>> it = distsOnData.entrySet().iterator(); it.hasNext();) {
            Map.Entry<String[], Float> entry = it.next();
            float distX = entry.getValue();
            if (distX > maxDistance) {
                continue;
            }
            distsCount++;
            int idx = (int) (distX / distInterval);
            float value = ret.get(idx) + 1;
            ret.put(idx, value);
        }
        for (Integer key : ret.keySet()) {
            float newValue = ret.get(key) / distsCount;
            ret.put(key, newValue);
        }
//        float sum = 0;
//        System.out.println("AAAAA");
//        for (Integer key : ret.keySet()) {
//            sum += ret.get(key);
//            System.out.println(ret.get(key) + ";" + sum);
//        }
        return ret;
    }

    private static float[][] evaluateDiscreteDensities(List<Float> hammingDistances, int sketchLength) {
        float[][] ret = new float[2][sketchLength + 1];
        if (hammingDistances.isEmpty()) {
            return ret;
        }
        for (Float hamDist : hammingDistances) {
            int hamDistI = hamDist.intValue();
            ret[0][hamDistI]++;
        }
        for (int i = 0; i <= sketchLength; i++) {
            ret[0][i] = ret[0][i] / hammingDistances.size();
        }
        ret[1][0] = ret[0][0];
        for (int i = 1; i <= sketchLength; i++) {
            ret[1][i] = ret[1][i - 1] + ret[0][i];
        }
        return ret;
    }

    @Override
    protected BivariateFunction createInstanceSepDistsx1x2InHammSpace() {
        return probOrderPreserve;
    }

    @Override
    public double evaluateNormalized(double maxX1, double stepIntegrating) {
        BivariateIntegratorLucia integrator;
        if (maxX1 >= 0) {
            integrator = new BivariateIntegratorLucia(probOrderPreserve, 0d, maxDistance, maxX1, stepIntegrating);
        } else {
            integrator = new BivariateIntegratorLucia(probOrderPreserve, 0d, maxDistance, maxDistance, stepIntegrating);
//            maxX1 = maxDistance;
        }
        double quality = integrator.evaluate();
//        return quality / maxDistance / maxDistance;
        return 1 - 2 * quality;
//        return quality;
    }

}
