/**
 * *****************************************************************************
 * Copyright (c) 2013, Lucia Vadicamo (NeMIS Lab., ISTI-CNR, Italy) All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer. 2. Redistributions in
 * binary form must reproduce the above copyright notice, this list of
 * conditions and the following disclaimer in the documentation and/or other
 * materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 * ****************************************************************************
 */
package sf.separabilityAnalytique;

import java.util.logging.Level;
import java.util.logging.Logger;
import org.apache.commons.math4.analysis.BivariateFunction;

/**
 * @author Lucia
 *
 */
public class BivariateIntegratorLucia {

    private final BivariateFunction func;
    private final double a;
    private final double b;
    private final double t;

    public static final int IMPLICIT_INTERVAL = 700;
    public final int n_interval;

//    public BivariateIntegratorLucia(BivariateFunction func, double a, double b) {
//        this(func, a, b, b);
//    }
//
//    public BivariateIntegratorLucia(BivariateFunction func, double a, double b, double t) {
//        this(func, a, b, t, IMPLICIT_INTERVAL);
//    }

    public BivariateIntegratorLucia(BivariateFunction func, double a, double b, double t, double step) {

        this.func = func;
        this.a = a;
        this.b = b;
        if (t <= 0) {
            this.t = b;
        } else {
            this.t = t;
        }
        if (step <= 0) {
            this.n_interval = IMPLICIT_INTERVAL;
        } else {
            this.n_interval = (int) ((b - a) / step);
        }
    }

    public double evaluate() {
        checkParams();
        return evaluateWithStep(n_interval);
    }

    private void checkParams() {
        boolean ok = b > a;
        ok = ok && func != null;
        if (!ok) {
            throw new IllegalArgumentException("Bad limits or function func: " + a + ", " + b + "; " + func);
        }
    }

    private double evaluateWithStep(int n) {
        Logger.getLogger(BivariateIntegratorLucia.class.getName()).log(Level.INFO, "Trying to evaluate with granularity {0}", n);

        double x0 = a;
        double xN = b; //TODO t
        double delta = (xN - x0) / n;
        System.err.println("Step: " + delta);

        if (t < n) {
            int maxi = (int) ((t - x0) / delta);
            maxi++;
            n = maxi + 1;
            xN = x0 + n * delta;
        }

        double ret = 0;

        //xi= a+ i*delta
        //XN= a+ i*n
        double lastsum = 0;
        for (int h = 1; h < n; h++) {
            double xh = x0 + h * delta;
            lastsum += func.value(xh, xN);
            ret += 2 * func.value(x0, xh);

            double volume = 0;
            for (int j = 1; j < h; j++) {
                double xj = x0 + j * delta;
                volume += func.value(xj, xh);
            }
            ret += 4 * volume;
        }

        ret += func.value(x0, xN);
//        double volume=0;
//    	for(int j=1; j<n; j++) {
//    		double xj= x0+j*delta;
//    		volume+=func.value(xj, xN);
//    	}
//    	ret+=2*volume;
        ret += 2 * lastsum;

        ret = ret * delta * delta / 4;

        return ret;
    }

}
