/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package sf.separabilityAnalytique;

import java.util.logging.Level;
import java.util.logging.Logger;
import org.apache.commons.math4.analysis.BivariateFunction;

/**
 *
 * @author xmic
 */
public class BivariateIntegratorLuciaOld {

    private final BivariateFunction func;
    private final double minDistance;
    private final double maxDistance;
    private final double maxX1;

    public static final Double IMPLICIT_STEP = 0.1d;
    public final double step;

    public BivariateIntegratorLuciaOld(BivariateFunction func, double minDistance, double maxDistance) {
        this(func, minDistance, maxDistance, maxDistance);
    }

    public BivariateIntegratorLuciaOld(BivariateFunction func, double minDistance, double maxDistance, double maxX1) {
        this(func, minDistance, maxDistance, maxX1, IMPLICIT_STEP);
    }

    public BivariateIntegratorLuciaOld(BivariateFunction func, double minDistance, double maxDistance, double maxX1, double step) {
        this.func = func;
        this.minDistance = minDistance;
        this.maxDistance = maxDistance;
        if (maxX1 <= 0) {
            this.maxX1 = maxDistance;
        } else {
            this.maxX1 = maxX1;
        }
        if (step <= 0) {
            this.step = IMPLICIT_STEP;
        } else {
            this.step = step;
        }
    }

    public double evaluate() {
        checkParams();
        return evaluateWithStep(step);
    }

    private void checkParams() {
        boolean ok = maxDistance > minDistance;
        ok = ok && func != null;
        if (!ok) {
            throw new IllegalArgumentException("Bad limits or function func: " + minDistance + ", " + maxDistance + "; " + func);
        }
    }

    private double evaluateWithStep(double step) {
        Logger.getLogger(BivariateIntegratorLucia.class.getName()).log(Level.INFO, "Trying to evaluate with granularity {0}", step);
        System.err.println("Step: " + step);
        double ret = 0;

        int lastPercentage = 0;
        double lastRet = 0;
        double x2 = minDistance;
        while (x2 <= maxX1) {
//            double percentageDone = (x1 - minDistance) / (maxX1 - minDistance);
//            percentageDone = percentageDone * 100;
//            if (percentageDone > lastPercentage + 5) {
//                Logger.getLogger(BivariateIntegrator.class.getName()).log(Level.INFO, "Done {0} percent. Current result: {1}. Diff: {2}", new Object[]{percentageDone, ret, (ret - lastRet)});
//                lastPercentage = (int) percentageDone;
//                lastRet = ret;
//            }
            double x1 = minDistance;
            while (x1 < x2) {
                double a = func.value(x1, x2);
                double b = func.value(x1 + step, x2);
                double c = func.value(x1, x2 + step);
                double d = func.value(x1 + step, x2 + step);
                x1 += step;
                if (a == -Double.MAX_VALUE || b == -Double.MAX_VALUE || c == -Double.MAX_VALUE || d == -Double.MAX_VALUE) {
                    continue;
                }
                double volume = step * step * (a + b + c + d) / 4;
//                System.err.println(volume);
//                if (a > 0 && b > 0 && c > 0 && d > 0) {
//                    String s = "";
//                }
                ret += volume;
            }
            x2 += step;
        }
        return ret;
    }

}
