/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package sf.algorithm.HWTOP;

import java.io.Serializable;
import java.util.BitSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;
import java.util.logging.Level;
import java.util.logging.Logger;
import main.tools.Tools;
import messif.algorithms.Algorithm;
import messif.objects.LocalAbstractObject;
import messif.objects.util.RankedAbstractObject;
import messif.objects.util.RankedSortedCollection;
import messif.operations.data.BulkInsertOperation;
import messif.operations.query.KNNQueryOperation;
import sf.algorithm.myHWT.OrientAndShuffleBitsForHWT;
import sf.objects.impl.Sketch;

/**
 *
 * @author xmic
 */
public class HWTOPAlgorithm extends Algorithm implements Serializable {

    private static final long serialVersionUID = 159756321895L;

    public static Integer distCounts = 0;
    private final BitSet[] bitsToOrientInTreeLevels;
    private final int[][] bitsPermutationsForTreeLevels;
    private final Node root;

    private final int sketchLength;

    /**
     *
     * @param sampleOfSketches
     * @param maxDepth max depth of tree, this root is level -1, i.e. the
     * sketches are split into 2^n parts in level n
     * @param sketchLength
     */
    public HWTOPAlgorithm(List<Sketch> sampleOfSketches, int maxDepth, int sketchLength) {
        super("HWT with oriented and permutated bits");
        Logger.getLogger(HWTOPAlgorithm.class.getName()).log(Level.INFO, "Creating tree that splits sketches to {0} parts", Math.pow(2, maxDepth - 1));
        bitsToOrientInTreeLevels = new BitSet[maxDepth];
        bitsPermutationsForTreeLevels = new int[maxDepth][sketchLength];
        this.sketchLength = sketchLength;
        root = new Node(0, maxDepth, null);
        learnOrientationAndPermutation(sampleOfSketches, maxDepth, sketchLength);
    }

    private void learnOrientationAndPermutation(List<Sketch> sampleOfSketches, int maxDepth, int sketchLength) {
        Logger.getLogger(HWTOPAlgorithm.class.getName()).log(Level.INFO, "Starting learning the bit orientations and permutations");
        float[][] corrMatrix = Tools.evaluateCorrelationMatrix(sampleOfSketches);
        for (int i = 0; i < maxDepth; i++) {
            OrientAndShuffleBitsForHWT o = new OrientAndShuffleBitsForHWT(Tools.copyMatrix(corrMatrix), (int) Math.pow(2, i), sketchLength);
            bitsToOrientInTreeLevels[i] = o.evaluateIndexesOfBitsToSwitch();
            if (i != 0) {
                bitsPermutationsForTreeLevels[i] = o.getOldBitsPermutation();
            } else {
                bitsPermutationsForTreeLevels[i] = null;
            }
        }
        Logger.getLogger(HWTOPAlgorithm.class.getName()).log(Level.INFO, "Learning the bit orientations and permutations finished");
    }

    public void insertSketches(BulkInsertOperation op) {
        Iterator<? extends LocalAbstractObject> it = op.getInsertedObjects().iterator();
        int counter = 0;
        while (it.hasNext()) {
            Sketch sketchOrig = (Sketch) it.next();
            Sketch[] sketchesPerLevels = getSketchPerLevels(sketchOrig);
            root.insertSketch(sketchesPerLevels, sketchLength);
            counter++;
            if (counter % 100000 == 0) {
                Logger.getLogger(HWTOPAlgorithm.class.getName()).log(Level.INFO, "Inserted {0} objects", counter);
            }
        }
        op.endOperation();
    }

    public List<Integer> getNumberOfObjectsInLeaves() {
        return root.getNumberOfObjectsInLeaves();
    }

    /**
     * Returns map where key is the number of parts in split sketches and value
     * is number of nodes with key with the given number of partes
     *
     * @return
     */
    public Map<Integer, Integer> getTreeStructureDescription() {
        Map<Integer, Integer> ret = new TreeMap<>();
        root.getStructure(ret);
        return ret;
    }

    public KNNQueryOperation knnSearch(KNNQueryOperation op) {
//        System.out.println("\n\n\nStart:");
        Sketch queryOrig = (Sketch) op.getQueryObject();
        String locatorURI = queryOrig.getLocatorURI();
        Sketch[] queriesPerLevels = getSketchPerLevels(queryOrig);

        int k = op.getK();
        RankedSortedCollection answer = new RankedSortedCollection(k, k);

        float currAnswer = Sketch.MAX_DISTANCE;
        int currRange = 0;

//        while (currRange < currAnswer) {
        root.knnSearch(locatorURI, queriesPerLevels, sketchLength, answer, op.getAnswerType(), currAnswer, currRange);
//            currAnswer = root.knnSearch(locatorURI, queriesPerLevels, sketchLength, answer, op.getAnswerType(), currAnswer, currRange);
//            currRange = root.lowerBoundOnSketchesInSubstree;
//        }
        Iterator<RankedAbstractObject> it = answer.iterator();
        while (it.hasNext()) {
            RankedAbstractObject obj = it.next();
            op.addToAnswer(obj.getObject(), obj.getDistance(), null);
        }
        op.endOperation();
//        System.out.println(distCounts + "Konec\n\n\n\n\n\n\n\n\n\n\n\n");
        return op;
    }

    private Sketch[] getSketchPerLevels(Sketch sketch) {
        Sketch[] ret = new Sketch[bitsToOrientInTreeLevels.length];
        Main.t6 -= System.currentTimeMillis();
        for (int i = 0; i < ret.length; i++) {
            ret[i] = Tools.orientSketch(sketch, bitsToOrientInTreeLevels[i], bitsPermutationsForTreeLevels[i]);
        }
        Main.t6 += System.currentTimeMillis();
        return ret;
    }

}
