/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package sf.algorithm.myHWT;

import java.util.BitSet;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;
import main.tools.Tools;
import sf.objects.impl.Sketch;

/**
 *
 * @author xmic
 */
public class OrientBitsForHWTSumOfCorrOverParts extends OrientBitsForHWT {

    protected final int numberOfparts;
    protected final int sketchLength;

    public OrientBitsForHWTSumOfCorrOverParts(float[][] correlationMatrix, int numberOfparts, int sketchLength) {
        super(correlationMatrix);
        this.numberOfparts = numberOfparts;
        this.sketchLength = sketchLength;
    }

    public OrientBitsForHWTSumOfCorrOverParts(List<Sketch> sketches, int numberOfparts, int sketchLength) {
        super(sketches);
        this.numberOfparts = numberOfparts;
        this.sketchLength = sketchLength;
    }

    @Override
    public BitSet evaluateIndexesOfBitsToSwitch() {
        BitSet ret = new BitSet();
        int blockSize = sketchLength / numberOfparts;
        for (int blockStart = 0; blockStart < sketchLength; blockStart += blockSize) {
            float[][] currSubmatrix = new float[blockSize][blockSize];
            for (int i = 0; i < blockSize; i++) {
                int column = blockStart + i;
                if (column >= bestFoundCorrelationMatrix.length) {
                    break;
                }
                for (int j = 0; j < blockSize; j++) {
                    int row = blockStart + j;
                    if (row >= bestFoundCorrelationMatrix.length) {
                        break;
                    }
                    currSubmatrix[i][j] = bestFoundCorrelationMatrix[row][column];
                }
            }
            Set<Integer> indexesOnSubmatrix = evaluateIndexesOfBitsToSwitchBruteForce(currSubmatrix);
            for (Integer i : indexesOnSubmatrix) {
                ret.set(i + blockStart);
            }
        }
        return ret;
    }

    private Set<Integer> evaluateIndexesOfBitsToSwitchBruteForce(float[][] matrix) {
        float[] sumAndCount = evaluateMatrixSumAndCountOfNegative(matrix);
        float greatestSum = sumAndCount[0];
        Logger.getLogger(OrientBitsForHWTGlobalSum.class.getName()).log(Level.INFO, "Matrix sum: {0}, number of negative values: {1}", new Object[]{sumAndCount[0], sumAndCount[1]});
        Set<Integer> ret = new HashSet<>();
        for (int poss = 0; poss < Math.pow(2, matrix.length); poss++) {
            Set<Integer> currSwitches = binaryStringToSet(Integer.toBinaryString(poss));
            if (currSwitches.size() > matrix.length / 2) {
                continue;
            }
            float[][] currMatrix = Tools.copyMatrix(matrix);
            for (int idx : currSwitches) {
                currMatrix = changeBitOrientation(idx, currMatrix);
            }
            sumAndCount = evaluateMatrixSumAndCountOfNegative(currMatrix);
            if (sumAndCount[0] > greatestSum) {
                greatestSum = sumAndCount[0];
                ret = new HashSet<>(currSwitches);
                Logger.getLogger(OrientBitsForHWTGlobalSum.class.getName()).log(Level.INFO, "Matrix sum: {0}, number of negative values: {1}", new Object[]{sumAndCount[0], sumAndCount[1]});
            }
        }
        return ret;
    }

    private Set<Integer> binaryStringToSet(String s) {
        Set<Integer> ret = new HashSet<>();
        for (int i = 0; i < s.length(); i++) {
            if (s.charAt(i) == '1') {
                ret.add(i);
            }
        }
        return ret;
    }
}
