/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package sf.algorithm.myHWT;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;
import main.tools.Tools;
import messif.algorithms.Algorithm;
import messif.algorithms.AlgorithmMethodException;
import messif.objects.LocalAbstractObject;
import messif.objects.impl.ObjectFloatVectorL2;
import messif.operations.data.BulkInsertOperation;
import messif.operations.query.KNNQueryOperation;
import messif.operations.query.RangeQueryOperation;
import sf.objects.impl.Sketch;
import sf.objects.sketchcreator.impl.GenHyperplane;

/**
 *
 * @author xmic
 */
public class SketchPreciseAlgorithm extends Algorithm {

    private final Map<Integer, List<Sketch>> sketches = new HashMap<>();

    public static void main(String[] args) {
        try {
            int sketchLength = 256;
            SketchPreciseAlgorithm alg = new SketchPreciseAlgorithm("MyAlgorithm");
            Iterator dataIt = Tools.getIterator(Sketch.class, "c:\\Data\\2016_orientedSketches\\DeCAF_Profiset\\1M_" + sketchLength + "_bits.txt");
            List<Sketch> queries = Tools.getObjects(Sketch.class.getName(), "c:\\Data\\2016_orientedSketches\\DeCAF_Profiset\\Queries-1000_" + sketchLength + "_bits.txt", -1);
            BulkInsertOperation insert = new BulkInsertOperation(dataIt);
            alg.executeOperation(insert);

            int k = 10;
//            int range = 36;
            for (Sketch query : queries) {
                KNNQueryOperation op = new KNNQueryOperation(query, k);
//                RangeQueryOperation op = new RangeQueryOperation(query, range);
                op.setParameter("sketchLength", sketchLength);
                alg.executeOperation(op);
//                Iterator<RankedAbstractObject> answer = knn.getAnswer();
//                while (answer.hasNext()) {
//                    RankedAbstractObject next = answer.next();
//                    System.out.println(next.getDistance() + ":" + next.getObject().getLocatorURI());
//                }
            }
            System.out.println("Saved: ");
            for (Float saved : savedDistanceComputations) {
                System.out.println(saved);
            }
        } catch (AlgorithmMethodException | NoSuchMethodException ex) {
            Logger.getLogger(SketchPreciseAlgorithm.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    public static List<Sketch> getQueries(boolean oriented, int sketchLength) {
        String clazz = ObjectFloatVectorL2.class.getName();
        String pivots = "c:\\Data\\2015_Doktorske\\Desc\\profiset\\caffe\\pivots-random-2560.data";
        String csv = "c:\\Data\\2015_Doktorske\\Desc\\profiset\\caffe\\pivots_filtered_hyperplane\\2560_random\\UncorrelatedPivots_033_256_pairs.csv";
        GenHyperplane tr = new GenHyperplane(pivots, clazz, sketchLength, csv, false);

        if (oriented) {
            String dataForOutlier = "c:\\Data\\2015_Doktorske\\Desc\\profiset\\caffe\\profi-neuralnet-1M.data.gz";
            LocalAbstractObject outlier = Tools.getObject(dataForOutlier, "0000028271", clazz);
            tr.reorderPivotsByOutlier(outlier);
        }
        Iterator<LocalAbstractObject> it = Tools.getIterator(clazz, "c:\\Data\\2015_Doktorske\\Desc\\profiset\\caffe\\profi-neuralnet-1000-query.data");
        List<Sketch> ret = new ArrayList<>();
        while (it.hasNext()) {
            LocalAbstractObject next = it.next();
            ret.add(tr.transform(next));
        }
        return ret;
    }

    public SketchPreciseAlgorithm(String algorithmName) throws IllegalArgumentException {
        super(algorithmName);
    }

    public void insertObjects(BulkInsertOperation op) {
        Iterator<Sketch> insertedObjects = (Iterator<Sketch>) op.getInsertedObjects().iterator();
        while (insertedObjects.hasNext()) {
            Sketch sketch = insertedObjects.next();
            Integer card = sketch.getCardinality();
            if (!sketches.containsKey(card)) {
                List<Sketch> list = new ArrayList<>();
                sketches.put(card, list);
            }
            sketches.get(card).add(sketch);
        }
        op.endOperation();
    }

    private static final List<Float> savedDistanceComputations = new ArrayList<>();

    public RangeQueryOperation rangeSearch(RangeQueryOperation op) {
        int counter = 0;
        int radius = (int) op.getRadius();
        Sketch querySketch = (Sketch) op.getQueryObject();
        int queryCard = querySketch.getCardinality();
//        //
//        int sketchLength = op.getParameter("sketchLength", Integer.class);
//        int addAllUpToLimit = radius - queryCard;
//        int addAllFromLimit = 2 * sketchLength - radius - queryCard;
//        for (int i = 0; i <= addAllUpToLimit; i++) {
//            List<Sketch> sublist = sketches.get(i);
//            if (sublist == null) {
//                continue;
//            }
//            for (Sketch dataObject : sublist) {
//                op.addToAnswer(dataObject, radius);
//            }
//        }
//        for (int i = addAllFromLimit; i <= sketchLength; i++) {
//            List<Sketch> sublist = sketches.get(i);
//            if (sublist == null) {
//                continue;
//            }
//            for (Sketch dataObject : sublist) {
//                op.addToAnswer(dataObject, radius);
//            }
//        }
//        //
        int minimalDistanceOfCurList = 0;
        while (minimalDistanceOfCurList < radius) {
            Integer key = queryCard + minimalDistanceOfCurList;
            if (/*key > addAllUpToLimit && key < addAllFromLimit && */sketches.containsKey(key)) {
                List<Sketch> sublist = sketches.get(key);
                for (Sketch dataObject : sublist) {
                    op.addToAnswer(dataObject, radius);
                    counter++;
                }
            }
            key = queryCard - minimalDistanceOfCurList;
            if (minimalDistanceOfCurList != 0 && minimalDistanceOfCurList <= radius && /*key > addAllUpToLimit && key < addAllFromLimit && */ sketches.containsKey(key)) {
                List<Sketch> sublist = sketches.get(key);
                for (Sketch dataObject : sublist) {
                    op.addToAnswer(dataObject, radius);
                    counter++;
                }
            }
            minimalDistanceOfCurList++;
        }
        op.endOperation();
        savedDistanceComputations.add(100 - counter / 10000f);
        return op;
    }

    public KNNQueryOperation knnSearch(KNNQueryOperation op) {
        int counter = 0;
        Sketch querySketch = (Sketch) op.getQueryObject();
        int queryCard = querySketch.getCardinality();
        int minimalDistanceOfCurList = 0;
        float answerDistance = Sketch.MAX_DISTANCE;
        while (minimalDistanceOfCurList < answerDistance) {
            Integer key = queryCard + minimalDistanceOfCurList;
            if (sketches.containsKey(key)) {
                List<Sketch> sublist = sketches.get(key);
                for (Sketch dataObject : sublist) {
                    op.addToAnswer(dataObject);
                    counter++;
                }
            }
            answerDistance = op.getAnswerDistance();
            key = queryCard - minimalDistanceOfCurList;
            if (minimalDistanceOfCurList != 0 && minimalDistanceOfCurList <= answerDistance && sketches.containsKey(key)) {
                List<Sketch> sublist = sketches.get(key);
                for (Sketch dataObject : sublist) {
                    op.addToAnswer(dataObject);
                    counter++;
                }
                answerDistance = op.getAnswerDistance();
            }
            minimalDistanceOfCurList++;
        }
        op.endOperation();
        savedDistanceComputations.add(100 - counter / 10000f);
//        System.err.println("Pocet: " + counter + ", uspora: " + (100 - counter / 10000f) + " %, posledni vzdalenost: " + minimalDistanceOfCurList);
        return op;
    }

}
