package sf.deprecated;

import java.io.IOException;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;
import static main.tools.oldNotUsed.DistanceDistributionToHyperplanesExperiment.floatToByteArrayForPrint;
import main.tools.Tools;
import messif.objects.LocalAbstractObject;
import messif.objects.util.AbstractObjectIterator;
import messif.objects.util.AbstractObjectList;
import sf.deprecated.vykryvaniProstoru.Evaluator;
import sf.deprecated.vykryvaniProstoru.Experiment;
import sf.deprecated.vykryvaniProstoru.ballPartitioning.PivotSelector;
import sf.deprecated.vykryvaniProstoru.ballPartitioning.ballSketches.Experiment03;
import sf.objects.impl.PivotsDistanceList;
import sf.objects.impl.Sketch;


/**
 *
 * @author xmic
 */
public class ExperimentKNNEvaluation {

    public static void evaluate(String[] args) {
        int pivotsCount = 60;
        AbstractObjectList objects = Tools.getObjects(args[1], args[0], Integer.parseInt(args[2]));
        AbstractObjectList queries = Tools.getObjects(args[1], "c:\\Data\\2015_Sketches\\Pokusy\\Queries_random_10.txt", -1);
        AbstractObjectList pivots = Tools.getObjects(args[1], "c:\\Data\\2015_Sketches\\Pokusy\\Jiný range\\03\\Pivots_160.txt", pivotsCount);
        float[] ranges = PivotSelector.countCover(3, objects, pivots);
        String outputFileName = "c:\\Data\\2015_Sketches\\Pokusy\\Jiný range\\03\\Deviations.csv";
        PivotsDistanceList pivotDistanceList = new PivotsDistanceList(pivots, ranges);
        Map<LocalAbstractObject, Sketch> signatureFiles = createSketches(objects.iterator(), pivotDistanceList);
        Evaluator evaluator = new Evaluator(queries, pivotDistanceList, signatureFiles, outputFileName);
        Experiment experiment = new Experiment03();
        evaluator.evaluateExperiment(experiment);
    }

    private static Map<LocalAbstractObject, Sketch> createSketches(AbstractObjectIterator<LocalAbstractObject> objectIterator, PivotsDistanceList pivotDistanceList) {
        List<LocalAbstractObject> pivots = pivotDistanceList.getPivots();
        float[] ranges = pivotDistanceList.getRanges();
        Map<LocalAbstractObject, Sketch> ret = new HashMap();
        while (objectIterator.hasNext()) {
            LocalAbstractObject next = objectIterator.next();
            Sketch sf = new Sketch(pivots.size());
            for (int i = 0; i < pivots.size(); i++) {
                float dist = next.getDistance(pivots.get(i));
                if (dist <= ranges[i]) {
                    sf.setBitValuePositive(i);
                }
            }
            ret.put(next, sf);
        }
        return ret;
    }

    private static void writeDownStats(int type, String fileName, PivotsDistanceList pivotDistanceList, Map<LocalAbstractObject, Sketch> signatureFiles, int pivotsCount) {
        PrintStream os = null;
        try {
            if (fileName == null) {
                os = System.out;
            } else {
                os = new PrintStream(fileName);
            }
            switch (type) {
                case 0: {
                    pivotDistanceList.writeDownPivots(os);
                    break;
                }
                case 1: {
                    pivotDistanceList.writeDownDistances(os);
                    break;
                }
                case 2: {
                    for (Map.Entry<LocalAbstractObject, Sketch> entry : signatureFiles.entrySet()) {
                        Sketch sf = entry.getValue();
                        os.write(entry.getKey().getLocatorURI().getBytes());
                        os.write(';');
                        os.write(Integer.toString(sf.getCardinality()).getBytes());
                        os.write('\n');
                    }
                    break;
                }
                case 3: {
                    int[] numberOfObjectsInBulks = new int[pivotsCount];
                    Arrays.fill(numberOfObjectsInBulks, 0);
                    for (Map.Entry<LocalAbstractObject, Sketch> entry : signatureFiles.entrySet()) {
                        Sketch sf = entry.getValue();
                        for (int i = 0; i < pivotsCount; i++) {
                            if (sf.getBitValue(i)) {
                                numberOfObjectsInBulks[i]++;
                            }
                        }
                    }
                    for (int i = 0; i < pivotsCount; i++) {
                        os.write(pivotDistanceList.getPivot(i).getLocatorURI().getBytes());
                        os.write(';');
                        os.write(floatToByteArrayForPrint(pivotDistanceList.getRanges()[i]));
                        os.write(';');
                        os.write(Integer.toString(numberOfObjectsInBulks[i]).getBytes());
                        os.write('\n');
                    }
                    break;
                }
                case 4: {
                    List<Map.Entry<LocalAbstractObject, Sketch>> entryList = new ArrayList(signatureFiles.entrySet());
                    int[] numberOfSameSF = new int[entryList.size()];
                    Arrays.fill(numberOfSameSF, -1);
                    for (int i = 0; i < entryList.size(); i++) {
                        Sketch sf1 = entryList.get(i).getValue();
                        for (int j = 0; j < entryList.size(); j++) {
                            Sketch sf2 = entryList.get(j).getValue();
                            if (sf1.equals(sf2)) {
                                numberOfSameSF[i]++;
                            }
                        }
                    }
                    for (int i = 0; i < entryList.size(); i++) {
                        os.write(Integer.toString(numberOfSameSF[i]).getBytes());
                        os.write('\n');
                    }
                    break;
                }
            }
        } catch (IOException ex) {
            Logger.getLogger(ExperimentKNNEvaluation.class.getName()).log(Level.SEVERE, null, ex);
        } finally {
            os.flush();
        }
    }

}
