package sf.deprecated.vykryvaniProstoru.ballPartitioning.variableRanges;

import java.util.Iterator;
import java.util.Map;
import java.util.SortedSet;
import java.util.TreeSet;
import messif.objects.LocalAbstractObject;
import sf.objects.impl.PivotsDistanceList;
import sf.objects.impl.Sketch;
import sf.deprecated.vykryvaniProstoru.Experiment;
import sf.deprecated.vykryvaniProstoru.ResultHolder;

/**
 *
 * @author xmic
 */
public class Experiment02 implements Experiment {

    @Override
    public SortedSet<ResultHolder> evaluateQuery(LocalAbstractObject query, Map<LocalAbstractObject, Sketch> signatureFiles, PivotsDistanceList pivotDistanceList) {
        Iterator<Map.Entry<LocalAbstractObject, Sketch>> it = signatureFiles.entrySet().iterator();
        SortedSet<ResultHolder> ret = new TreeSet();
        while (it.hasNext()) {
            Map.Entry<LocalAbstractObject, Sketch> entry = it.next();
            float dist = query.getDistance(entry.getKey());
            Comparable deviation = makeDeviationMinMax(pivotDistanceList, entry.getValue(), query);
            ResultHolder rs = new ResultHolder(query.getLocatorURI(), entry.getKey().getLocatorURI(), dist, deviation);
            ret.add(rs);
        }
        return ret;
    }

    private Comparable makeDeviationMinMax(PivotsDistanceList pivotDistanceList, Sketch sf, LocalAbstractObject query) {
        float maxR = 0f;
        float minR = Float.MAX_VALUE;
        float[] ranges = pivotDistanceList.getRanges();
        for (int i = 0; i < sf.getSize(); i++) {
            if (sf.getBitValue(i)) {
                float dist = query.getDistance(pivotDistanceList.getPivot(i));
                float diff = ranges[i] - dist;
                if (diff < 0) {
                    continue;
                }
                if (maxR < diff) {
                    maxR = diff;
                }
                if (minR > diff) {
                    minR = diff;
                }
            }
        }
        return new Result(maxR, minR);
    }

    private Comparable makeDeviationAverage(PivotsDistanceList pivotDistanceList, Sketch sf, LocalAbstractObject query) {
        float ret = 0f;
        float[] ranges = pivotDistanceList.getRanges();
        int count = 0;
        for (int i = 0; i < sf.getSize(); i++) {
            if (sf.getBitValue(i)) {
                float dist = query.getDistance(pivotDistanceList.getPivot(i));
                float diff = ranges[i] - dist;
                if (diff < 0) {
                    continue;
                }
                ret += diff;
                count++;
            }
        }
        return ret / count;
    }

    public class Result implements Comparable<Result> {

        private final float maxR;
        private final float minR;

        public Result(float maxR, float minR) {
            this.maxR = maxR;
            this.minR = minR;
        }

        @Override
        public int compareTo(Result o) {
            int ret;
//            ret = Float.compare(maxR - minR, o.maxR - o.minR);
//            if (ret != 0) {
//                return ret;
//            }
            ret = Float.compare(maxR, o.maxR);
            if (ret != 0) {
                return -ret;
            }
            ret = Float.compare(minR, o.minR);
            return +ret;
        }

        public float getMaxR() {
            return maxR;
        }

        public float getMinR() {
            return minR;
        }

    }
}
