/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package sf.objects.sketchcreator.impl;

import java.util.Iterator;
import java.util.List;
import messif.objects.LocalAbstractObject;
import sf.objects.impl.Sketch;
import sf.objects.sketchcreator.ObjectToSketchTransformator;
import sf.pivotAnalyzer.CsvPivotInfoFileProcessor;
import sf.pivotAnalyzer.PivotSubsetSelector;
import sf.pivotAnalyzer.csvPivotInfoFileProcessorImpl.CsvFileBBRanges;

/**
 *
 * @author Matej Kvassay <www.matejkvassay.sk>
 */
public class BBPartitioning extends ObjectToSketchTransformator {

    private final float[] bb25;
    private final float[] bb50;
    private final float[] bb75;

    public BBPartitioning(String pivotFile, String objectClass, int pivotCouplesCount, String csvFileBBInfo, boolean uniquePivots) {
        super(pivotFile, objectClass, -1);
        bb25 = new float[pivotCouplesCount];
        bb50 = new float[pivotCouplesCount];
        bb75 = new float[pivotCouplesCount];
        CsvPivotInfoFileProcessor pivotCreator = new CsvFileBBRanges(csvFileBBInfo, pivots, bb25, bb50, bb75);
        pivots = pivotCreator.getPivotsAccordingCsvLocators(pivotCouplesCount * 2);
    }

    @Override
    public Sketch transform(LocalAbstractObject objectToTransform) {
        if (objectToTransform == null) {
            return null;
        }
        Sketch ret = new Sketch(objectToTransform.getLocatorURI(), pivots.length);
        for (int i = 0; i < pivots.length; i += 2) {
            float dist = getMinDistToPivot(objectToTransform, pivots[i], pivots[i + 1]);
            setBitCouple(ret, dist, i);
        }
        return ret;
    }

    private void setBitCouple(Sketch sketch, float dist, int idxP1) {
        if (dist <= bb25[idxP1 / 2]) {
            sketch.setBitValuePositive(idxP1);
            sketch.setBitValuePositive(idxP1 + 1);
        } else if (dist <= bb50[idxP1 / 2]) {
            sketch.setBitValuePositive(idxP1);
        } else if (dist > bb75[idxP1 / 2]) {
            sketch.setBitValuePositive(idxP1 + 1);
        }
    }

    private float getMinDistToPivot(LocalAbstractObject obj, LocalAbstractObject pivot1, LocalAbstractObject pivot2) {
        return Math.min(obj.getDistance(pivot1), obj.getDistance(pivot2));
    }

    @Override
    public void transformToInvertedFile(LocalAbstractObject sampleObj, List<Sketch> invertedSketches, int objectNumber) {
        for (int i = 0; i < pivots.length - 1; i += 2) {
            float dist = getMinDistToPivot(sampleObj, pivots[i], pivots[i + 1]);
            int idx = i / 2;
            if (dist <= bb25[idx]) {
                invertedSketches.get(i).setBitValuePositive(objectNumber);
                invertedSketches.get(i + 1).setBitValuePositive(objectNumber);
            } else if (dist <= bb50[idx]) {
                invertedSketches.get(i).setBitValuePositive(objectNumber);
            } else if (dist > bb75[idx]) {
                invertedSketches.get(i + 1).setBitValuePositive(objectNumber);
            }
        }
    }

    @Override
    public void removeCorrelatedBits(int[] uncorrelatedBits) {
        LocalAbstractObject[] newPivots = new LocalAbstractObject[uncorrelatedBits.length * 2];
        for (int i = 0; i < uncorrelatedBits.length; i += 2) {
            newPivots[i] = pivots[uncorrelatedBits[i]];
            newPivots[i + 1] = pivots[uncorrelatedBits[i + 1]];
        }
        pivots = newPivots;
        PivotSubsetSelector.printAsPairs(pivots, System.out);
    }

    @Override
    public int getInvertedSketchesCount() {
        return pivots.length;
    }

 }
