package sf.pivotAnalyzer.uncorrelatedSelectorImpl;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Random;
import java.util.logging.Level;
import java.util.logging.Logger;
import sf.objects.impl.Sketch;
import sf.pivotAnalyzer.uncorrelatedSelector.InvertedSketchesCorrelationCounter;

/**
 *
 * @author xmic
 */
public class HeuristicCorrelationSubsetSearch extends InvertedSketchesCorrelationCounter {

    public HeuristicCorrelationSubsetSearch(List<Sketch> invertedSketches, int sampleObjectNumber, float correlationToAchieve) {
        super(invertedSketches, sampleObjectNumber, correlationToAchieve);
    }

    @Override
    protected int[] selectLowCorrelatedSubset(int tuplesLength, float[][] sketchesCorrelations, int retSize) {
        Logger.getLogger(InvertedSketchesCorrelationCounter.class.getName()).log(Level.INFO, "Starting selection of low correlated subset.");
        Random rand = new Random(System.currentTimeMillis());
        updateCorrelations(sketchesCorrelations, -correlationToAchieve);
        List<Integer> curIndexes = new ArrayList<>();
        List<Integer> blackList = new ArrayList<>();
        addIndexes(tuplesLength, blackList, curIndexes, retSize, rand, sketchesCorrelations, true);
        List<Integer> bestIndexes = null;
        float lowestFoundCorrelation = Float.MAX_VALUE;
        int iterationsCounter = 0;
        int starts = 0;
        while (starts < 10) {//10
            if (iterationsCounter % 4000 == 0) { // 4000
                starts++;
                curIndexes.clear();
                blackList.clear();
                addIndexes(tuplesLength, blackList, curIndexes, retSize, rand, sketchesCorrelations, false);
                Logger.getLogger(InvertedSketchesCorrelationCounter.class.getName()).log(Level.INFO, "No better solution found for {0} iterations. Lowest max correlation: {1}. Start number {2}", new Object[]{iterationsCounter, lowestFoundCorrelation, starts});
            }
            iterationsCounter++;
            Integer[] curIndexesArray = curIndexes.toArray(new Integer[0]);
            float worstCorrelation = getHighestCorrelation(curIndexesArray, sketchesCorrelations);
            if (worstCorrelation < lowestFoundCorrelation) {
                bestIndexes = new ArrayList<>(curIndexes);
                lowestFoundCorrelation = worstCorrelation;
                Logger.getLogger(InvertedSketchesCorrelationCounter.class.getName()).log(Level.INFO, "Found subset with worst correlation {0}. Started in {1}, iteration {2}", new Object[]{lowestFoundCorrelation, starts, iterationsCounter});
            }
            float diff;
//            if (rand.nextInt(100) < 80) {
            diff = 0.02f;
//            } else {
//                diff = 0.03f;
//            }
            List<Integer> correlatedIndexes = getCorrelatedIndexes(tuplesLength, worstCorrelation, curIndexesArray, sketchesCorrelations, diff);
//            blackList.addAll(correlatedIndexes);
            curIndexes.removeAll(correlatedIndexes);
            boolean bySum = rand.nextInt(100) < 60;
            addIndexes(tuplesLength, blackList, curIndexes, retSize, rand, sketchesCorrelations, bySum);
            while (blackList.size() > retSize * 0.8) {
                blackList.remove(0);
            }
        }
        int[] ret = new int[retSize];
        for (int i = 0; i < ret.length; i++) {
            ret[i] = bestIndexes.get(i);
        }
        updateCorrelations(sketchesCorrelations, correlationToAchieve);
        return ret;
    }

    private void addIndexes(int tuplesLength, Collection<Integer> blackList, Collection<Integer> ret, int retSize, Random rand, float[][] sketchesCorrelations, boolean bySum) {
        int start = rand.nextInt(sketchesCorrelations.length);
        while (ret.size() < retSize) {
            float[] maxCorrelationOfCandWithSet = new float[sketchesCorrelations.length];
            for (int i = 0; i < sketchesCorrelations.length; i++) {
                int idx = (start + i) % sketchesCorrelations.length;
                maxCorrelationOfCandWithSet[idx] = -1;
                if (ret.contains(idx)) {
                    maxCorrelationOfCandWithSet[idx] = 1;
                } else {
                    if (bySum) {
                        float sum = 0;
                        for (int intSet : ret) {
                            sum += sketchesCorrelations[idx][intSet];
                        }
                        maxCorrelationOfCandWithSet[idx] = sum;
                    } else {
                        for (int intSet : ret) {
                            if (sketchesCorrelations[idx][intSet] > maxCorrelationOfCandWithSet[idx]) {
                                maxCorrelationOfCandWithSet[idx] = sketchesCorrelations[idx][intSet];
                            }
                        }
                    }
                }
            }
            float minMaxCorSum = Float.MAX_VALUE;
            int[] addIdx = new int[tuplesLength];
            for (int i = 0; i < maxCorrelationOfCandWithSet.length - tuplesLength + 1; i += tuplesLength) {
                float sum = 0;
                boolean isAllowed = true;
                for (int j = 0; j < tuplesLength; j++) {
                    int curIndx = i + j;
                    if (blackList.contains(curIndx) || ret.contains(curIndx)) {
                        isAllowed = false;
                        break;
                    }
                    sum += maxCorrelationOfCandWithSet[curIndx];
                }
                if (isAllowed && sum < minMaxCorSum) {
                    for (int j = 0; j < tuplesLength; j++) {
                        Integer curIndx = i + j;
                        addIdx[j] = curIndx;
                    }
                    minMaxCorSum = sum;
                }
            }
            for (int j = 0; j < tuplesLength; j++) {
                ret.add(addIdx[j]);
            }
        }
    }

    private List<Integer> getCorrelatedIndexes(int tuplesLength, float worstCorrelation, Integer[] curIndexes, float[][] sketchesCorrelations, float coef) {
        List<Integer> removed = new ArrayList<>();
        float limit = worstCorrelation - coef;
        for (int i = 0; i < curIndexes.length - 1; i++) {
            int idxI = curIndexes[i];
            for (int j = i + 1; j < curIndexes.length; j++) {
                int idxJ = curIndexes[j];
                if (sketchesCorrelations[idxI][idxJ] > limit && !removed.contains(idxI) && !removed.contains(idxJ)) {
                    float sumI = 0;
                    float sumJ = 0;
                    for (int k = 0; k < curIndexes.length; k++) {
                        sumI += sketchesCorrelations[idxI][k];
                        sumJ += sketchesCorrelations[idxJ][k];
                    }
                    int rem = sumI > sumJ ? idxI : idxJ;
                    int startToRemove = rem - rem % tuplesLength;
                    for (int k = 0; k < tuplesLength; k++) {
                        removed.add(startToRemove + k);
                    }
                }
            }
        }
        return removed;
    }

}
