/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package sf.recallEstimation.UnivariateFunctionImpl;

import java.util.logging.Level;
import java.util.logging.Logger;
import org.apache.commons.math4.analysis.UnivariateFunction;
import org.apache.commons.math4.analysis.integration.BaseAbstractUnivariateIntegrator;
import sf.recallEstimation.integrating.OnePeakPositiveFunctionIntegrator;

/**
 *
 * @author xmic
 */
public class MuB {

    private final int bSkLength;
    private final int nDatasetSize;
    private final UnivariateFunction fxDistDistr;
    private final UnivariateFunction pxProbSkBitsDif;
    private final double maxDistance;
    private final Double sketchesIDim;

    public MuB(int nDatasetSize, int bSketchLength, double maxDistance, Double sketchesIDim, UnivariateFunction fxDistDistr, UnivariateFunction pxProbSkBitsDif) {
        this.nDatasetSize = nDatasetSize;
        this.fxDistDistr = fxDistDistr;
        this.pxProbSkBitsDif = pxProbSkBitsDif;
        this.bSkLength = bSketchLength;
        this.maxDistance = maxDistance;
        this.sketchesIDim = sketchesIDim;
    }

    public double evaluateMuB(int bBitIndex) {
        MuBIntegralBody integralBody = new MuBIntegralBody(bBitIndex);
        BaseAbstractUnivariateIntegrator integrator = new OnePeakPositiveFunctionIntegrator(0, maxDistance);
        double ret = nDatasetSize * integrator.integrate(Integer.MAX_VALUE, integralBody, 0, maxDistance);
        Logger.getLogger(MuB.class.getName()).log(Level.INFO, "Mean for {0} bits out of {1} evaluated. Result: {2}", new Object[]{bBitIndex + 1, bSkLength + 1, Double.toString(ret)});
        return ret;
    }

    private class MuBIntegralBody implements UnivariateFunction {

        private final int bBitIndex;

        public MuBIntegralBody(int bBitIndex) {
            this.bBitIndex = bBitIndex;
        }

        @Override
        public double value(double x) {
            Pxb pxb = new Pxb(bSkLength, bBitIndex, sketchesIDim, pxProbSkBitsDif);
            double fx = fxDistDistr.value(x);
            return pxb.value(x) * fx;
        }

    }

}
