/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package sf.recallEstimation.UnivariateFunctionImpl;

import java.util.logging.Level;
import java.util.logging.Logger;
import org.apache.commons.math4.analysis.UnivariateFunction;
import org.apache.commons.math4.analysis.integration.BaseAbstractUnivariateIntegrator;
import sf.recallEstimation.integrating.OnePeakPositiveFunctionIntegrator;

/**
 *
 * @author xmic
 */
public class Sigma2B {

    private final int bSkLength;
    private final int nDatasetSize;
    private final UnivariateFunction fxDistDistr;
    private final UnivariateFunction pxProbSkBitsDif;
    private final double maxDistance;
    private final Double sketchesIDim;

    public Sigma2B(int nDatasetSize, int bSketchLength, double maxDistance, Double sketchesIDim, UnivariateFunction fxDistDistr, UnivariateFunction pxProbSkBitsDif) {
        this.nDatasetSize = nDatasetSize;
        this.fxDistDistr = fxDistDistr;
        this.pxProbSkBitsDif = pxProbSkBitsDif;
        this.bSkLength = bSketchLength;
        this.maxDistance = maxDistance;
        this.sketchesIDim = sketchesIDim;
    }

    public double evaluateSigma2B(int bBitIndex) {
        Sigma2BIntegralBody integralBody = new Sigma2BIntegralBody(bBitIndex);
        BaseAbstractUnivariateIntegrator integrator = new OnePeakPositiveFunctionIntegrator(0, maxDistance);
        double integrate = integrator.integrate(Integer.MAX_VALUE, integralBody, 0, maxDistance);
        double ret = nDatasetSize * integrate;
        if (ret <= 0) {
            Logger.getLogger(Sigma2B.class.getName()).log(Level.WARNING, "Zero variance {0} replaced with minimal value {1}", new Object[]{Double.toString(ret), Double.toString(Double.MIN_NORMAL)});
            return Double.MIN_NORMAL;
        }
        return ret;
    }

    private class Sigma2BIntegralBody implements UnivariateFunction {

        private final int bBitIndex;

        public Sigma2BIntegralBody(int bBitIndex) {
            this.bBitIndex = bBitIndex;
        }

        @Override
        public double value(double x) {
            Pxb pxbFce = new Pxb(bSkLength, bBitIndex, sketchesIDim, pxProbSkBitsDif);
            double pxb = pxbFce.value(x);
            double fx = fxDistDistr.value(x);
            return (pxb - pxb * pxb) * fx;
        }

    }

}
