/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package sf.recallEstimation.hamThresholdEstimation;

import java.util.logging.Level;
import java.util.logging.Logger;
import org.apache.commons.math4.analysis.UnivariateFunction;
import org.apache.commons.math4.analysis.integration.BaseAbstractUnivariateIntegrator;
import sf.recallEstimation.RecallEstimator;
import sf.recallEstimation.UnivariateFunctionImpl.PxbSmallBinAnalogue;
import sf.recallEstimation.integrating.OnePeakPositiveFunctionIntegrator;

/**
 *
 * @author xmic
 */
public class HammingGlobalThresholdEstimatorBasedOnDistanceDensity {

    private final int nDatasetSize;
    private final int k;
    private final double maxDistance;
    private final Double sketchesIDim;
    private final UnivariateFunction fxDistDistr;
    private final UnivariateFunction pxProbSkBitsDif;
    private final int bSkLength;
    private final double x0;

    private double[] px0bCumulative = null;

    public HammingGlobalThresholdEstimatorBasedOnDistanceDensity(int nDatasetSize, int k, double maxDistance, Double sketchesIDim, int bSkLength, UnivariateFunction fxDistDistr, UnivariateFunction pxProbSkBitsDif) {
        this.nDatasetSize = nDatasetSize;
        this.k = k;
        this.maxDistance = maxDistance;
        this.fxDistDistr = fxDistDistr;
        this.bSkLength = bSkLength;
        this.pxProbSkBitsDif = pxProbSkBitsDif;
        this.sketchesIDim = sketchesIDim;
        x0 = estimateX0();
    }

    private double estimateX0() {
        double curX0 = maxDistance / 2;
        double curStep = maxDistance / 4;
        double maxDev = 1E-8;
        double curDev = Double.MAX_VALUE;
        BaseAbstractUnivariateIntegrator integrator = new OnePeakPositiveFunctionIntegrator(0, maxDistance);
        while (Math.abs(curDev) > maxDev) {
            double curK = nDatasetSize * integrator.integrate(Integer.MAX_VALUE, fxDistDistr, 0, curX0);
//            Logger.getLogger(RecallEstimator.class.getName()).log(Level.INFO, "Estimating x0. Current k: {0}, goal: {1}, limit of deviance: {2}, deviance: {3}. X_0: {4}", new Object[]{curK, k, Double.toString(maxDev), Double.toString(curDev), curX0});
            curDev = curK - k;
            if (curDev < 0) {
                curX0 += curStep;
            } else {
                curX0 -= curStep;
            }
            curStep = curStep / 2;
        }
        Logger.getLogger(RecallEstimator.class.getName()).log(Level.INFO, "Finished evaluating x0: {0}", Double.toString(curX0));
        return curX0;
    }

    public int getnDatasetSize() {
        return nDatasetSize;
    }

    public int getK() {
        return k;
    }

    public double getMaxDistance() {
        return maxDistance;
    }

    public Double getSketchesIDim() {
        return sketchesIDim;
    }

    public int getbSkLength() {
        return bSkLength;
    }

    public double getX0() {
        return x0;
    }

    public double estimateProbabilityX0UpToB(int b) {
        if (px0bCumulative != null) {
            return px0bCumulative[b];
        }
        double[] px0b = new double[bSkLength + 1];
        px0bCumulative = new double[bSkLength + 1];

        double pxProbSkBitsDifX = pxProbSkBitsDif.value(x0);
        PxbSmallBinAnalogue pxb = new PxbSmallBinAnalogue(bSkLength, sketchesIDim, pxProbSkBitsDifX);

        double sum = 0;
        for (int i = 0; i <= bSkLength; i++) {
            px0b[i] = pxb.value(i);
            sum += px0b[i];
            px0bCumulative[i] = sum;
        }
        return px0bCumulative[b];
    }

    private double[] filteredObjectsCumulative = null;

    public double estimateNumberOfFilteredObjects(int b) {
        if (filteredObjectsCumulative != null) {
            return filteredObjectsCumulative[b];
        }
        double[] filteredObjects = new double[bSkLength + 1];
        filteredObjectsCumulative = new double[bSkLength + 1];
        PxbSmallBinAnalogue pxb = new PxbSmallBinAnalogue(bSkLength, sketchesIDim, 0.5);
        double sum = 0;
        for (int i = 0; i <= bSkLength; i++) {
            filteredObjects[i] = pxb.value(i);
            sum += filteredObjects[i];
            filteredObjectsCumulative[i] = 1 - sum;
        }
        return filteredObjectsCumulative[b];
    }

}
