/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package sf.recallEstimation.pxEvaluator;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.SortedMap;
import java.util.TreeMap;
import java.util.logging.Level;
import java.util.logging.Logger;
import main.tools.Tools;
import messif.objects.LocalAbstractObject;
import sf.objects.impl.Sketch;
import sf.objects.sketchcreator.ObjectToSketchTransformator;
import sf.objects.sketchcreator.impl.GenHyperplane;

/**
 *
 * @author xmic
 */
public class PxEvaluator {

    private final int minNumberOfExampleForBitToCount;

    private final String className;
    private final String pivotFile;
    private final String csvFileWithPivotsPairs;
    private final int sketchSize;
    private final String queriesFile;
    private final String sampleFile;
    private final double distInterval;
    private final double maxDist;
    private final int objectCount;

    public PxEvaluator(String className, String pivotFile, String csvFileWithPivotsPairs, int sketchSize, String queriesFile, String sampleFile, double distInterval, double maxDist, int objectCount) {
        this(className, pivotFile, csvFileWithPivotsPairs, sketchSize, queriesFile, sampleFile, distInterval, maxDist, objectCount, -1);
    }

    public PxEvaluator(String className, String pivotFile, String csvFileWithPivotsPairs, int sketchSize, String queriesFile, String sampleFile, double distInterval, double maxDist, int objectCount, int minNumberOfExampleForBitToCount) {
        this.className = className;
        this.pivotFile = pivotFile;
        this.csvFileWithPivotsPairs = csvFileWithPivotsPairs;
        this.sketchSize = sketchSize;
        this.queriesFile = queriesFile;
        this.sampleFile = sampleFile;
        this.distInterval = distInterval;
        this.maxDist = maxDist;
        this.objectCount = objectCount;
        this.minNumberOfExampleForBitToCount = minNumberOfExampleForBitToCount;
    }

    private String[] csvFiles = null;

    public PxEvaluator(String className, String pivotFile, String[] csvFileWithPivotsPairs, int sketchSize, String queriesFile, String sampleFile, double distInterval, double maxDist, int objectCount, int minNumberOfExampleForBitToCount) {
        this(className, pivotFile, csvFileWithPivotsPairs[0], sketchSize, queriesFile, sampleFile, distInterval, maxDist, objectCount, minNumberOfExampleForBitToCount);
        this.csvFiles = csvFileWithPivotsPairs;
    }

    public PxEvaluator(String className, String pivotFile, String[] csvFileWithPivotsPairs, int sketchSize, String queriesFile, String sampleFile, double distInterval, double maxDist, int objectCount) {
        this(className, pivotFile, csvFileWithPivotsPairs[0], sketchSize, queriesFile, sampleFile, distInterval, maxDist, -1);
        this.csvFiles = csvFileWithPivotsPairs;
    }

    /**
     * Returns pairs [x,y] for plot
     *
     * @return x .. orig distance, y .. probability
     */
    public SortedMap<Float, Float> evaluateProbabilities() {
        ObjectToSketchTransformator transformator = new GenHyperplane(pivotFile, className, sketchSize, csvFileWithPivotsPairs, false);
//        ObjectToSketchTransformator transformator = new GenHyperplaneSettingAllOnesIfMany(pivotFile, className, sketchSize, csvFileWithPivotsPairs, false);
//        ObjectToSketchTransformator transformator = new MultiVoronoiSketchTransformator(pivotFile, className, sketchSize, csvFiles, false);
        List<LocalAbstractObject> queries = Tools.getObjects(className, queriesFile, -1);
        Iterator<LocalAbstractObject> data = Tools.getIterator(className, sampleFile);
        List<SortedMap<Float, Float>> bitProbabilities = evalBitProbsAndHamDistDistrs(transformator, queries, data);
        return createAverages(bitProbabilities, sketchSize);
    }

    private Map<String, int[]> hamDistDistrForQueries;

    protected float[] evaluateAvgHamDistDistr() {
        float[] ret = null;
        for (Map.Entry<String, int[]> entrySet : hamDistDistrForQueries.entrySet()) {
            int[] queryValues = entrySet.getValue();
            if (ret == null) {
                ret = new float[queryValues.length];
                Arrays.fill(ret, 0f);
            }
            for (int i = 0; i < queryValues.length; i++) {
                ret[i] += queryValues[i];
            }
        }
        for (int i = 0; i < ret.length; i++) {
            ret[i] = ret[i] / hamDistDistrForQueries.size();
        }
        return ret;
    }

    private List<SortedMap<Float, Float>> evalBitProbsAndHamDistDistrs(ObjectToSketchTransformator transformator, List<LocalAbstractObject> queries, Iterator<LocalAbstractObject> data) {
        // probs evaluation
        // for each bit of sketches contains map to create plot [x,y], where x is original distance and y is a probability that coresponding bit differs
        List<SortedMap<Float, Float>> ret = new ArrayList<>();
        for (int i = 0; i < sketchSize; i++) {
            ret.add(new TreeMap<>());
        }
        List<Sketch> querySketches = new ArrayList<>();
        // evaluate distance distribution of sketches for mu and sigmaB estimation as well
        hamDistDistrForQueries = new HashMap<>();
        int sketchLengths = transformator.getSketchLengths() + 1;
        for (int i = 0; i < queries.size(); i++) {
            // probs evaluation
            Sketch querySketch = transformator.transform(queries.get(i));
            querySketches.add(querySketch);
            // mu and sigmaB estimation
            int[] zeros = new int[sketchLengths];
            Arrays.fill(zeros, 0);
            hamDistDistrForQueries.put(querySketch.getLocatorURI(), zeros);
        }
        int dataCounter = 0;
        Map<Float, Integer> denominators = new HashMap<>();
        while (data.hasNext() && dataCounter < objectCount) {
            if (dataCounter % 1000 == 0) {
                Logger.getLogger(TestPxEvaluator.class.getName()).log(Level.INFO, "Processed {0} data objects.", dataCounter);
            }
            dataCounter++;
            LocalAbstractObject dataObject = data.next();
            Sketch dataSketch = transformator.transform(dataObject);
            for (int i = 0; i < queries.size(); i++) {
                // probs evaluation
                LocalAbstractObject query = queries.get(i);
                Sketch querySketch = querySketches.get(i);
                updateProbabilities(query, querySketch, dataObject, dataSketch, ret, denominators, distInterval, sketchSize);
                // mu and sigmaB estimation
                String queryLocator = querySketch.getLocatorURI();
                int[] values = hamDistDistrForQueries.get(queryLocator);
                values[(int) querySketch.getDistance(dataSketch)]++;
                hamDistDistrForQueries.put(queryLocator, values);
            }
        }
        for (int i = 0; i < sketchSize; i++) {
            SortedMap<Float, Float> map = ret.get(i);
            Set<Float> keySet = map.keySet();
            for (Float key : keySet) {
                Float value = map.get(key);
                int denominator = denominators.get(key);
                if (minNumberOfExampleForBitToCount < denominator) {
                    System.err.println("count for dist " + key + ": " + denominator + " value: " + (value / denominator));
                    map.put(key, value / denominator);
                } else {
                    map.put(key, 0f);
                }
            }
            System.err.println("");
        }
        return ret;
    }

    public static void updateProbabilities(LocalAbstractObject query, Sketch querySketch, LocalAbstractObject dataObject, Sketch dataSketch, List<SortedMap<Float, Float>> probabilities, Map<Float, Integer> denominators, double distInterval, int sketchSize) {
        updateProbabilities(query.getDistance(dataObject), querySketch, dataSketch, probabilities, denominators, distInterval, sketchSize);
    }

    public static void updateProbabilities(float dist, Sketch querySketch, Sketch dataSketch, List<SortedMap<Float, Float>> probabilities, Map<Float, Integer> denominators, double distInterval, int sketchSize) {
        float coordX = Tools.round(dist, distInterval, true);
        if (!denominators.containsKey(coordX)) {
            denominators.put(coordX, 0);
        }
        Integer denominator = denominators.get(coordX);
        denominators.put(coordX, denominator + 1);
        boolean[] xoredBits = querySketch.getXoredSketchAsBools(dataSketch);
        for (int i = 0; i < Math.min(sketchSize, xoredBits.length); i++) {
            if (xoredBits[i]) {
                SortedMap<Float, Float> map = probabilities.get(i);
                if (!map.containsKey(coordX)) {
                    map.put(coordX, 0f);
                }
                Float coordY = map.get(coordX);
                map.put(coordX, coordY + 1);
            }
        }
    }

    private static void print(List<SortedMap<Float, Float>> bitProbabilities, boolean isAverage, double distInterval, double maxDist) {
        for (int i = 0; i < bitProbabilities.size(); i++) {
            SortedMap<Float, Float> map = bitProbabilities.get(i);
            if (isAverage) {
//                System.out.println("Average probabilities:");
            } else {
                System.out.println("Bit " + i + " probabilities:");
            }
            float previousKey = 0f;
            for (Map.Entry<Float, Float> entrySet : map.entrySet()) {
                float key = Tools.round(entrySet.getKey(), distInterval, false);
                float value = entrySet.getValue();
                while (key > previousKey) {
                    System.out.println(previousKey + ";" + 0);
                    previousKey += distInterval;
                    previousKey = Tools.round(previousKey, distInterval, false);
                }
                System.out.println(key + ";" + value);
                key += distInterval;
                key = Tools.round(key, distInterval, false);
                previousKey = key;
            }
            while (previousKey <= maxDist) {
                System.out.println(previousKey + ";" + 0);
                previousKey += distInterval;
                previousKey = Tools.round(previousKey, distInterval, false);
            }
        }
    }

    public static SortedMap<Float, Float> createAverages(List<SortedMap<Float, Float>> bitProbabilities, int sketchSize) {
        SortedMap<Float, Float> ret = new TreeMap<>();        
        for (int i = 0; i < bitProbabilities.size(); i++) {
            SortedMap<Float, Float> map = bitProbabilities.get(i);
            for (Map.Entry<Float, Float> entrySet : map.entrySet()) {
                Float key = entrySet.getKey();
                Float value = entrySet.getValue();
                if (!ret.containsKey(key)) {
                    ret.put(key, 0f);
                }
                Float retValue = ret.get(key);
                ret.put(key, retValue + value);
            }
        }
        Set<Float> keySet = ret.keySet();
        for (Float key : keySet) {
            ret.put(key, ret.get(key) / sketchSize);
        }
        return ret;
    }

    public static void printProbabilities(SortedMap<Float, Float> probabilities, double distInterval, double maxDist) {
        List<SortedMap<Float, Float>> singletonList = Collections.singletonList(probabilities);
        print(singletonList, true, distInterval, maxDist);
    }

    private static double[] mub;
    private static double[] sigmab;

    public static double[] getMub() {
        return mub;
    }

    public static double[] getSigmab() {
        return sigmab;
    }

    public static SortedMap<Float, Float> parse(String csvFilePath) {
        SortedMap<Float, Float> ret = new TreeMap<>();
        List<String>[] csv = Tools.parseCsvKeysValues(csvFilePath);
        int count = csv[0].size();
        for (int i = 0; i < count; i++) {
            Float dist = Float.parseFloat(csv[0].get(i));
            Float prob = Float.parseFloat(csv[1].get(i));
            ret.put(dist, prob);
        }
        csv = Tools.parseCsvTriplets(csvFilePath);
        count = csv[0].size();
        mub = new double[count];
        sigmab = new double[count];
        for (int i = 0; i < count; i++) {
            mub[i] = Double.parseDouble(csv[1].get(i));
            sigmab[i] = Double.parseDouble(csv[2].get(i));
        }
        return ret;
    }

    public void printProbsMeansSigmas(SortedMap<Float, Float> probs, float[] avgHamDistDistr) {
//        printProbabilities(probs);
        double sum = 0;
        for (int i = 0; i < avgHamDistDistr.length; i++) {
            double countOfI = avgHamDistDistr[i];
            double mean = sum + countOfI / 2;
            double sigma = countOfI / 4;
            System.out.println(i + ";" + mean + ";" + sigma);
            sum += countOfI;
        }
    }

}
