/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package sf.separabilityAnalytique;

import java.util.logging.Level;
import java.util.logging.Logger;
import org.apache.commons.math4.analysis.BivariateFunction;

/**
 *
 * @author xmic
 */
public class BivariateIntegrator {

    private final BivariateFunction func;
    private final double minDistance;
    private final double maxDistance;
    private final double maxX1;

//    public static final double STEP = 0.01;
    public static final double STEP = 0.1;

    public BivariateIntegrator(BivariateFunction func, double minDistance, double maxDistance) {
        this(func, minDistance, maxDistance, maxDistance);
    }

    public BivariateIntegrator(BivariateFunction func, double minDistance, double maxDistance, double maxX1) {
        this.func = func;
        this.minDistance = minDistance;
        this.maxDistance = maxDistance;
        this.maxX1 = maxX1;
    }

    public double evaluate() {
        checkParams();
        return evaluateWithStep(STEP);
    }

    private void checkParams() {
        boolean ok = maxDistance > minDistance;
        ok = ok && func != null;
        if (!ok) {
            throw new IllegalArgumentException("Bad limits or function func: " + minDistance + ", " + maxDistance + "; " + func);
        }
    }

    private double evaluateWithStep(double step) {
        Logger.getLogger(BivariateIntegrator.class.getName()).log(Level.INFO, "Trying to evaluate with granularity {0}", step);
        System.err.println("Step: " + step);
        double ret = 0;
        double x1 = minDistance;

        int lastPercentage = 0;
        double lastRet = 0;
        while (x1 <= maxX1) {
            double percentageDone = (x1 - minDistance) / (maxX1 - minDistance);
            percentageDone = percentageDone * 100;
            if (percentageDone > lastPercentage + 5) {
                Logger.getLogger(BivariateIntegrator.class.getName()).log(Level.INFO, "Done {0} percent. Current result: {1}. Diff: {2}", new Object[]{percentageDone, ret, (ret - lastRet)});
                lastPercentage = (int) percentageDone;
                lastRet = ret;
            }
            double x2 = x1;
            while (x2 <= maxDistance) {
                double a = func.value(x1, x2);
                double b = func.value(x1 + step, x2);
                double c = func.value(x1, x2 + step);
                double d = func.value(x1 + step, x2 + step);
                double volume = step / 4 * step * (a + b + c + d);
//                System.err.println(volume);
                ret += volume;
                x2 += step;
            }
            x1 += step;
        }
        return ret;
    }

//    private Boolean isSomeDiffBiggerThenThreshold(double... values) {
//        for (int i = 0; i < values.length - 1; i++) {
//            for (int j = i + 1; j < values.length; j++) {
//                if (Math.abs(values[i] - values[j]) > MAX_DIFF_THRESHOLD) {
//                    System.err.println("Diff: " + Math.abs(values[i] - values[j]));
//                    return true;
//                }
//            }
//        }
//        return false;
//    }
}
